% This script contains the commands to generate data files used to
% set initial conditions or friction law parameters.  It can be run
% from the command line by typing:
% matlab < input_heterogeneous.m >& output &
% where output will be created as a text file holding any messages
% from matlab (on event of error, etc.)

% First, we need to set the number of grid points in each
% direction (nx and ny) and generate the x and y coordinate
% vectors, centered on x0 and y0 with grid spacing h:

nx = 256;
ny = 256;

x0 = 0;
y0 = -7.5;
h = 0.3;

x = [0:nx-1]*h;
x = x0+x-0.5*max(x);
y = [0:ny-1]*h;
y = y0+y-0.5*max(y);

% Now let's add variability to the fields, but only inside the
% fault, which is delimited by: xmin,xmax,ymin,ymax

xmin = -15;
xmax = 15;
ymin = -15;
ymax = 0;

% Part I: Initial stress fields.
% There are three initial stress fields, sx0, sy0, and sz0.  Let's give both 
% shear stress fields a complicated form, but leave the normal stress alone:

sx0 = zeros(nx,ny);
sy0 = zeros(nx,ny);
sz0 = zeros(nx,ny);

imin = find(x<=xmin,1,'last');
imax = find(x>=xmax,1,'first');
jmin = find(y<=ymin,1,'last');
jmax = find(y>=ymax,1,'first');

for j=jmin:jmax
  for i=imin:imax
    sx0(i,j) = 3*sin((x(i)-4)/8)+3*sin((x(i)-2.7)/3)*cos((y(j)-0.9)/2)+6*exp(-(x(i)-7)^2/20)*exp(-(y(j)+3)^2/10);
    sy0(i,j) = sin(x(i)/1.2)*sin(y(j)/1.2);
  end
end

% Now we need to save the data to a file.  For illustration, this
% will be a text file called 'heterogeneous.load'.  The order in
% which the fields are output matters!  You also always have to output
% all fields.

file = 'heterogeneous.load';
f = fopen(file,'w');

for j=1:ny
    for i=1:nx
      fprintf(f,'%20.13e ',sx0(i,j));
    end
    fprintf(f,'\n');
end

for j=1:ny
    for i=1:nx
      fprintf(f,'%20.13e ',sy0(i,j));
    end
    fprintf(f,'\n');
end

for j=1:ny
    for i=1:nx
      fprintf(f,'%20.13e ',sz0(i,j));
    end
    fprintf(f,'\n');
end

fclose(f);


% Part II: Slip-weakening fields
% There are three fields: Dc, fd, fs, which correspond to
% slip-weakening distance, dynamic friction, and static friction.
% This time we only make fs variable; the others will remain unperturbed.

Dc = zeros(nx,ny);
fd = zeros(nx,ny);
fs = zeros(nx,ny);

for j=jmin:jmax
  for i=imin:imax
    Dc(i,j) = 0;
    fd(i,j) = 0;
    fs(i,j) = 2*sin(x(i)/0.9)*sin(y(j)/2.4)+3*sin((x(i)-3)/2.4)*sin((y(j)+4)/0.9)-2*sin((x(i)+8)/0.9)*sin((y(j)-13)/1.8); 
  end
end

fs = fs/120;

% Now we need to save the data to a file.  For illustration, this
% will be a binary file called 'heterogeneous.sw'.  You will need
% to make sure the endian of this file is identical to the endian
% of the machine on which you are running the program.  If you are
% running this script on the same machine, set endian='n' (native
% format).  Otherwise, make sure it is set correctly.

endian = 'l';

file = 'heterogeneous.sw';
f = fopen(file,'w',endian);

fwrite(f,[nx*ny*8],'integer*4');
fwrite(f,Dc,'real*8');
fwrite(f,[nx*ny*8],'integer*4');

fwrite(f,[nx*ny*8],'integer*4');
fwrite(f,fd,'real*8');
fwrite(f,[nx*ny*8],'integer*4');

fwrite(f,[nx*ny*8],'integer*4');
fwrite(f,fs,'real*8');
fwrite(f,[nx*ny*8],'integer*4');

fclose(f);


