! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module convolution_routines

  ! CONVOLUTION_ROUTINES contains subroutines to call convolution routines
  ! for identical material or bimaterial problems
  ! 
  ! Modified: 25 January 2007

  implicit none

  
contains

  
  subroutine do_convolution(mdl,krn,cnv,mp,update_history)
    ! DO_CONVOLUTION performs convolution
    ! 
    ! Modified: 25 January 2007
    
    use model, only : model_type
    use kernel, only : kernel_type
    use convolution, only : convolution_type
    use mpi_routines, only : mpi_type,master,scatterv,MPI_CPLX_PR
    use convolution_im, only : do_convolution_im
    use convolution_bm, only : do_convolution_bm
    use mpi

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! KRN = kernel variables
    ! CNV = convolution variables
    ! MP = MPI variables
    ! UPDATE_HISTORY = perform convolution over past history instead of just updating the portion 
    ! of the convolution pertaining to the current time step

    type(model_type),intent(in) :: mdl
    type(kernel_type),intent(in) :: krn
    type(convolution_type),intent(inout) :: cnv
    type(mpi_type),intent(in) :: mp
    logical,intent(in) :: update_history

    ! Internal Parameters:
    ! J = index of ky
    ! IERR = MPI error flag
    ! COUNTS = MPI message sizes
    ! DISPLS = MPI message displacements

    integer :: j,ierr
    integer,dimension(0:mp%np-1) :: counts,displs

    ! prepare for scatter/gather

    call scatterv(cnv%nkx,mp%np,counts,displs)

    ! scatter Fourier transforms of slip and current step's contribution to history from master to slaves

    if (mdl%bm) then
       
       do j = 1,cnv%nky

          call MPI_Scatterv(cnv%Dxp(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dxp(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Dxm(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dxm(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Dyp(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dyp(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Dym(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dym(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Dzp(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dzp(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Dzm(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dzm(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)

       end do
       
    else

       do j = 1,cnv%nky

          call MPI_Scatterv(cnv%Dx(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dx(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Dy(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Dy(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)

       end do
       
    end if
       
    ! perform convolution

    if (mdl%bm) then
       call do_convolution_bm(mdl,krn,cnv,update_history)
    else
       call do_convolution_im(mdl,krn,cnv,update_history)
    end if

    ! gather Fourier transforms of stress transfer

    if (mdl%bm) then

       do j = 1,cnv%nky
          call MPI_Gatherv(cnv%Fxp(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fxp(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Gatherv(cnv%Fxm(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fxm(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Gatherv(cnv%Fyp(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fyp(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Gatherv(cnv%Fym(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fym(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Gatherv(cnv%Fzp(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fzp(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Gatherv(cnv%Fzm(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fzm(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
       end do

    else

       do j = 1,cnv%nky
          call MPI_Gatherv(cnv%Fx(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fx(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Gatherv(cnv%Fy(:,j),counts(mp%myid),MPI_CPLX_PR, &
               cnv%Fy(:,j),counts,displs,MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
       end do

    end if

  end subroutine do_convolution


end module convolution_routines
