! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module dugdale

  ! DUGDALE contains variables and routines related to the
  ! self-similar Dugdale friction law
  ! 
  ! Modified: 26 November 2006

  use constants, only : pr,pin
  use asperity, only : asperity_list

  implicit none

  ! DUGDALE_TYPE is a derived type containing variables related to 
  ! Dugdale friction law
  !
  ! Parameters:
  ! ASPERITY_FILE = add heterogeneity to initial Dugdale fields
  !      T = use asperity array data file
  !      F = do not use asperity array data file
  ! FILENAME = asperity file name
  ! ASCII = asperity file format
  !      T = ascii
  !      F = binary
  ! LIST = singly-linked list of asperities
  ! D0 = weakening distance at each point
  ! FD = dynamic coefficient of friction at each point
  ! FS = static coefficient of friction at each point
  ! VX1 = x-velocity of imposed stress drop region for t<T
  ! VY1 = y-velocity of imposed stress drop region for t<T
  ! VX2 = x-velocity of imposed stress drop region for t>T
  ! VY2 = x-velocity of imposed stress drop region for t>T
  ! RX = semi-axis of imposed stress drop region for static problem, x direction
  ! RY = semi-axis of imposed stress drop region for static problem, y direction
  ! T = time of speed jump
  ! P = power-law exponent of weakening
  ! FORM = form of slip-weakening
  !      linear = linear slip-weakening
  !      exponential = exponential slip-weakening
  !      dugdale = Dugdale slip-weakening
  !      power = power-law slip-weakening

  type dugdale_type
     logical :: asperity_file,ascii
     character(len=64) :: filename,form
     type(asperity_list) :: list
     real(pr),dimension(:,:),allocatable :: d0,fd,fs
     real(pr) :: vx1,vy1,vx2,vy2,T,rx,ry,p
  end type dugdale_type

contains


  subroutine read_dugdale(ninput,dg)
    ! READ_DUGDALE reads in dugdale variables from file
    ! 
    ! Modified: 10 April 2006

    use io, only : error
    use asperity, only : read_asperity_list
    implicit none
    
    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! DG = Dugdale variables

    integer(pin),intent(in) :: ninput
    type(dugdale_type),intent(out) :: dg

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! VX1 = x-velocity of imposed stress drop region for t<T
    ! VY1 = y-velocity of imposed stress drop region for t<T
    ! VX2 = x-velocity of imposed stress drop region for t>T
    ! VY2 = x-velocity of imposed stress drop region for t>T
    ! RX = semi-axis of imposed stress drop region for static problem, x direction
    ! RY = semi-axis of imposed stress drop region for static problem, y direction
    ! T = time of speed jump
    ! FORM = form of slip-weakening
    ! P = power-law exponent
    ! ASPERITY_LIST = add heterogeneity to initial dg fields with list
    ! ASPERITY_FILE = add heterogeneity to initial dg fields from file
    ! FILENAME = asperity file name
    ! ASCII = asperity file format
    ! NDATA = number of field perturbations

    integer(pin) :: stat,ndata
    real(pr) :: vx1,vy1,vx2,vy2,T,rx,ry,p
    logical :: asperity_list,asperity_file,ascii
    character(len=64) :: filename,form

    ! make namelist of user input variables

    namelist /dugdale_list/ vx1,vy1,vx2,vy2,T,rx,ry,p,form,asperity_list,asperity_file,filename,ascii

    ! default values

    vx1 = 0._pr
    vx2 = 0._pr
    vy1 = 0._pr
    vy2 = 0._pr
    rx = 0._pr
    ry = 0._pr
    T = 100._pr
    p = 1._pr
    form = 'dugdale'

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=dugdale_list,iostat=stat)
    if (stat/=0) call error("Error reading namelist 'dugdale_list' in .in file",'read_dugdale')

    ! assign input variables to components of derived type

    dg%vx1 = vx1
    dg%vy1 = vy1
    dg%vx2 = vx2
    dg%vy2 = vy2
    dg%rx = rx
    dg%ry = ry
    dg%T = T
    dg%form = form
    dg%p = p
    dg%asperity_file = asperity_file
    dg%filename = 'problems/' // adjustl(filename)
    dg%ascii = ascii

    ! asperity list, set ndata = 3 (for d0,fd,fs)

    ndata = 3
    if (asperity_list) &
         dg%list = read_asperity_list(ninput,'dugdale',ndata)

  end subroutine read_dugdale


  subroutine init_dugdale(necho,mdl,dg)
    ! INIT_DUGDALE initializes dugdale variables
    ! 
    ! Modified: 12 November 2006

    use io, only : open_array_data,close_array_data,read_array_data,write_matlab
    use model, only : model_type
    use asperity, only : assign_list_data,destroy_asperity_list

    implicit none
    
    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MDL = model variables
    ! DG = Dugdale variables

    integer(pin),intent(in) :: necho
    type(model_type),intent(in) :: mdl
    type(dugdale_type),intent(inout) :: dg

    ! Internal Parameters:
    ! FILEUNIT = asperity file unit

    integer(pin) :: fileunit

    ! allocate arrays

    allocate(dg%d0(mdl%nx,mdl%ny))
    allocate(dg%fs(mdl%nx,mdl%ny))
    allocate(dg%fd(mdl%nx,mdl%ny))

    dg%d0 = 0._pr
    dg%fd = 0._pr
    dg%fs = 0._pr

    ! read asperity arrays from file

    if (dg%asperity_file) then
       call open_array_data(dg%filename,fileunit,dg%ascii)
       call read_array_data(fileunit,dg%d0,dg%ascii,'d0')
       call read_array_data(fileunit,dg%fd,dg%ascii,'fd')
       call read_array_data(fileunit,dg%fs,dg%ascii,'fs')
       call close_array_data(fileunit)
    end if
    
    ! use list data to perturb fields

    call assign_list_data(dg%d0,dg%list,1,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    call assign_list_data(dg%fd,dg%list,2,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    call assign_list_data(dg%fs,dg%list,3,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)

    ! deallocate memory for asperity list

    call destroy_asperity_list(dg%list)

    ! output variable values into matlab file

    call write_matlab(necho,'rx',dg%rx,'dg')
    call write_matlab(necho,'ry',dg%ry,'dg')
    call write_matlab(necho,'vx1',dg%vx1,'dg')
    call write_matlab(necho,'vx2',dg%vx2,'dg')
    call write_matlab(necho,'vy1',dg%vy1,'dg')
    call write_matlab(necho,'vy2',dg%vy2,'dg')
    call write_matlab(necho,'T',dg%T,'dg')
    call write_matlab(necho,'form',dg%form,'dg')
    call write_matlab(necho,'p',dg%p,'dg')

  end subroutine init_dugdale


  subroutine destroy_dugdale(dg)
    ! DESTROY_DUGDALE destroys derived type dg
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! DG = Dugdale variables

    type(dugdale_type),intent(inout) :: dg

    ! deallocate memory assigned to allocatable arrays

    if (allocated(dg%d0)) deallocate(dg%d0)
    if (allocated(dg%fd)) deallocate(dg%fd)
    if (allocated(dg%fs)) deallocate(dg%fs)

  end subroutine destroy_dugdale


  subroutine dugdale_im(u,vx,vy,sx,sy,sx0,sy0,sz,fx,fy,x,y,t,i,j,cv,dg,transverse_slip)
    ! DUGDALE_IM solves Dugdale friction law, identical materials case.
    ! 
    ! Modified: 26 January 2006

    use io, only : error

    implicit none

    ! I/O Parameters:
    ! U = slip
    ! VX = slip velocity in x direction
    ! VY = slip velocity in y direction
    ! SX = stress in x direction
    ! SY = stress in y direction
    ! SX0 = load in x direction
    ! SY0 = load in y direction
    ! SZ = stress in z direction
    ! FX = stress transfer in x direction
    ! FY = stress transfer in y direction
    ! X = distance in x direction
    ! Y = distance in y direction
    ! T = time
    ! CV = coefficient of radiation damping
    ! I = index in x direction at which friction is applied
    ! J = index in y direction at which friction is applied  
    ! DG = Dugdale variables
    ! TRANSVERSE_SLIP = allow rake rotation or not

    real(pr),intent(in) :: u,sx0,sy0,sz,fx,fy,x,y,t,cv
    real(pr),intent(inout) :: vx,vy,sx,sy
    integer(pin),intent(in) :: i,j
    type(dugdale_type),intent(in) :: dg
    logical,intent(in) :: transverse_slip

    ! apply either vector or scalar Dugdale friction law

    if (transverse_slip) then
       call error('Transverse slip not yet implemented in Dugdale model','dugdale_im')
       !call dugdale_vector_im(u,vx,vy,sx,sy,sx0,sy0,sz,fx,fy,x,y,t,i,j,cv,dg)
    else
       call dugdale_scalar_im(u,vx,sx,sx0,sz,fx,x,y,t,i,j,cv,dg)
       vy = 0._pr
       sy = sy0+fy
    end if

  end subroutine dugdale_im


  subroutine dugdale_scalar_im(u,v,s,s0,sz,f,x,y,t,i,j,cv,dg)
    ! DUGDALE_SCALAR_IM solves scalar Dugdale friction law (no rake rotation, 
    ! slip constrained to be in the x direction).  This version is for the
    ! identical materials case.
    ! 
    ! Modified: 2 March 2006

    implicit none

    ! I/O Parameters:
    ! U = slip
    ! V = slip velocity
    ! S = shear stress
    ! S0 = shear load
    ! SZ = normal stress
    ! F = stress transfer
    ! X = distance in x direction
    ! Y = distance in y direction
    ! T = time
    ! CV = coefficient of radiation damping
    ! I = index in x direction at which friction is applied
    ! J = index in y direction at which friction is applied  
    ! DG = Dugdale variables

    real(pr),intent(in) :: u,s0,sz,f,x,y,t,cv
    real(pr),intent(inout) :: v,s
    integer(pin),intent(in) :: i,j
    type(dugdale_type),intent(in) :: dg

    ! Internal Parameters:
    ! V1 = radial velocity before jump
    ! V2 = radial velocity after jump
    ! PHI = azimuthal angle
    ! R = radial distance from origin
    ! R0 = distance to rupture front (really edge of prescribed stress drop region) at t=0
    ! RF = radial distance to rupture front (really edge of prescribed stress drop region)
    ! ST = shear strength
    ! FRIC = coefficient of friction

    real(pr) :: v1,v2,phi,r,r0,rf,st,fric

    ! calculate azimuthal angle and radial velocities at that angle

    phi = atan2(y,x)

    r0 = dg%rx*dg%ry
    if (r0/=0._pr) r0 = r0/sqrt((dg%ry*cos(phi))**2+(dg%rx*sin(phi))**2)
    v1 = dg%vx1*dg%vy1
    if (v1/=0._pr) v1 = v1/sqrt((dg%vy1*cos(phi))**2+(dg%vx1*sin(phi))**2)
    v2 = dg%vx2*dg%vy2
    if (v2/=0._pr) v2 = v2/sqrt((dg%vy2*cos(phi))**2+(dg%vx2*sin(phi))**2)

    ! calculate radial distance of front from origin

    if (t<dg%T) then
       rf = r0+v1*t
    else
       rf = r0+v1*dg%T+v2*(t-dg%T)
    end if

    ! calculate radial distance from origin

    r = sqrt(x**2+y**2)

    ! check if point lies inside prescribed stress drop region

    if (r<rf) then
       ! friction set to dynamic value
       fric = dg%fd(i,j)
       st = fric*(-sz)
       ! stress set to strength, fault slips
       s = st
       v = (-s+s0+f)/cv
    else
       ! calculate coefficient of friction
       fric = coeff_fric(u,i,j,dg)
       st = fric*(-sz)
       ! compare stress to strength
       if (s0+f<st) then
          ! stress less than strength, fault locked
          v = 0._pr
          s = s0+f
       else
          ! stress set to strength, fault slips
          s = st
          v = (-s+s0+f)/cv
       end if
    end if

  end subroutine dugdale_scalar_im


  function coeff_fric(u,i,j,dg) result(fric)
    ! COEFF_FRIC returns the coefficient of friction, given the slip
    ! 
    ! Modified: 11 April 2006

    implicit none

    ! I/O Parameters:
    ! U = cumulative slip (line integral)
    ! I = index in x direction at which friction is applied
    ! J = index in y direction at which friction is applied  
    ! DG = Dugdale variables
    ! FRIC = coefficient of friction

    real(pr),intent(in) :: u
    integer(pin),intent(in) :: i,j
    type(dugdale_type),intent(in) :: dg
    real(pr) :: fric

    select case(dg%form)

    case('power','powerlaw') ! power-law weakening

       fric = max(dg%fs(i,j)-sign(1._pr,u)*(dg%fs(i,j)-dg%fd(i,j))*(dg%p/(dg%p+1._pr)*abs(u)/dg%d0(i,j))**dg%p,dg%fd(i,j))

    case('linear') ! linear weakening

       fric = max(dg%fs(i,j)-(dg%fs(i,j)-dg%fd(i,j))*u/dg%d0(i,j),dg%fd(i,j))

    case('exponential') ! exponential weakening

       fric = dg%fd(i,j)+(dg%fs(i,j)-dg%fd(i,j))*exp(-u/dg%d0(i,j))

    case('dugdale','Dugdale')

       if (u>=dg%d0(i,j)) then
          fric = dg%fd(i,j)
       else
          fric = dg%fs(i,j)
       end if

    end select

  end function coeff_fric


end module dugdale
