! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module fault

  ! FAULT contains arrays of all of the fields (slip, stress, etc.) that 
  ! are defined on the fault.  It also holds the initial values of the fields 
  ! and calls routines associated with asperities in these fields.
  ! 
  ! Modified: 14 February 2007

  use constants, only : pin,pr
  use asperity, only : asperity_input
  use preslip, only : preslip_type

  implicit none

  ! FAULT_TYPE is a derived type containing fault variables
  !
  ! Parameters:
  ! PRESLIPPED = initiate with nonzero initial slip or not
  ! THERMPRES = use thermal pressurization model or not
  ! DYNLOAD = include dynamic (time-dependent) load or not
  ! POROELASTIC = calculate poroelastic response or not
  ! TEMPERATURE = whether or not temperature field will be used by friction law
  ! STATE = whether or not state variable will be used by friction law
  ! RERR = relative error; has three indices, from 1 (new step) to 3 (past step) 
  ! AERR = relative error; has three indices, from 1 (new step) to 3 (past step) 
  ! QM = combination of poroelastic properties, minus side
  ! QP = combination of poroelastic properties, plus side
  ! Q0 = combination of poroelastic properties
  ! UX = slip in x direction
  ! UY = slip in y direction
  ! VX = slip velocity in x direction
  ! VY = slip velocity in y direction
  ! U = cumulative slip (integrated as line integral)
  ! V = magnitude of slip velocity
  ! UXP = displacement in x direction, plus side
  ! UXM = displacement in x direction, minus side
  ! UYP = displacement in y direction, plus side
  ! UYM = displacement in y direction, minus side
  ! UZP = displacement in z direction, plus side
  ! UZM = displacement in z direction, minus side
  ! VXP = particle velocity in x direction, plus side
  ! VXM = particle velocity in x direction, minus side
  ! VYP = particle velocity in y direction, plus side
  ! VYM = particle velocity in y direction, minus side
  ! VZP = particle velocity in z direction, plus side
  ! VZM = particle velocity in z direction, minus side
  ! SX = stress in x direction
  ! SY = stress in y direction
  ! SZ = stress in z direction
  ! SX0 = load in x direction
  ! SY0 = load in y direction
  ! SZ0 = load in z direction, including pore pressure
  ! FXI = stress transfer in x direction (interpolated in time)
  ! FYI = stress transfer in y direction (interpolated in time)
  ! FXPI = stress transfer in x direction, plus side (interpolated in time)
  ! FXMI = stress transfer in x direction, minus side (interpolated in time)
  ! FYPI = stress transfer in y direction, plus side (interpolated in time)
  ! FYMI = stress transfer in y direction, minus side (interpolated in time)
  ! FZPI = stress transfer in z direction, plus side (interpolated in time)
  ! FZMI = stress transfer in z direction, minus side (interpolated in time)
  ! FX = stress transfer in x direction
  ! FY = stress transfer in y direction
  ! FXP = stress transfer in x direction, plus side
  ! FXM = stress transfer in x direction, minus side
  ! FYP = stress transfer in y direction, plus side
  ! FYM = stress transfer in y direction, minus side
  ! FZP = stress transfer in z direction, plus side
  ! FZM = stress transfer in z direction, minus side
  ! TFRONT = time at which the rupture front passes each point
  ! PO = poroelastic pore pressure change
  ! Q = state variable
  ! DQ = rate of change of state variable
  ! T0 = temperature on fault
  ! P0 = pore pressure on fault
  ! T = temperature
  ! DT = rate of change of temperature
  ! P = pore pressure
  ! DP = rate of change of pore pressure
  ! The first two indices of each array correspond to grid points in the
  ! x and y directions (except for temperature and pressure, wheter the first index runs in z direction
  ! and the next two correspond to x and y). Some fields have a final index, which is used to store
  ! values of the field at time locations between time steps (used in higher-order time stepping methods).
  ! SCALE_U = scale for displacement (used for relative error estimates)
  ! SCALE_Q = scale for state (used for relative error estimates)
  ! SCALE_T = scale for temperature (used for relative error estimates)
  ! SCALE_P = scale for pore pressure (used for relative error estimates)
  ! Set above SCALE_XXX to zero if current value of XXX is to be used for typical scale,
  ! but note that this causes problems when value of XXX approaches zero.
  ! PS = preslip parameters
  ! ASPI = asperity input list for fields on fault

  type fault_type
     logical :: preslipped,thermpres,dynload,poroelastic,temperature,state
     real(pr) :: qm,qp,q0,rerr(3),aerr(3),scale_U,scale_Q,scale_T,scale_p
     real(pr),dimension(:,:),allocatable :: fxi,fyi,fxpi,fxmi,fypi,fymi,fzpi,fzmi
     real(pr),dimension(:,:,:),allocatable :: U,V, &
          Ux,Uy,Vx,Vy, & 
          uxp,vxp,uyp,vyp,uzp,vzp, &
          uxm,vxm,uym,vym,uzm,vzm, &
          sx,sy,sz, &
          fx,fy,fxp,fxm,fyp,fym,fzp,fzm, &
          Q,dQ,T0,p0
     real(pr),dimension(:,:,:,:),allocatable :: T,dT,p,dp
     real(pr),dimension(:,:),allocatable :: sx0,sy0,sz0,tfront,po
     type(preslip_type),pointer :: ps=>null()
     type(asperity_input) :: aspi
  end type fault_type


contains


  subroutine read_fault(ninput,mdl,flt)
    ! READ_FAULT reads in fault variables from file
    ! 
    ! Modified: 5 February 2007

    use constants, only : zero
    use asperity, only : read_asperity_list
    use io, only : error
    use model, only : model_type
    use preslip, only : read_preslip

    implicit none

    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! MDL = model variables
    ! FLT = fault variables

    integer(pin),intent(in) :: ninput
    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt

    ! Internal Parameters:
    ! PRESLIPPED = initiate with nonzero initial slip or not
    ! THERMPRES = use thermal pressurization model or not
    ! DYNLOAD = including dynamic (time-dependent) load or not
    ! POROELASTIC = calculate poroelastic response or not
    ! QM = combination of poroelastic properties, minus side
    ! QP = combination of poroelastic properties, plus side
    ! Q0 = combination of poroelastic properties
    ! SCALE_U = scale for displacement (used for relative error estimates)
    ! SCALE_Q = scale for state (used for relative error estimates)
    ! SCALE_T = scale for temperature (used for relative error estimates)
    ! SCALE_P = scale for pore pressure (used for relative error estimates)
    ! STAT = I/O error flag
    ! NDATA = number of fault perturbations
    ! ASPERITY_LIST = add heterogeneity with list
    !      T = use asperity list
    !      F = do not use asperity list
    ! ASPERITY_FILE = add heterogeneity with file
    ! FILENAME = asperity file name
    ! ASCII = asperity file format

    logical :: preslipped,thermpres,dynload,poroelastic,asperity_list,asperity_file,ascii
    real(pr) :: qm,qp,q0,scale_U,scale_Q,scale_T,scale_p
    integer(pin) :: stat,ndata
    character(64) :: filename

    ! make namelist of user input variables

    namelist /fault_list/ preslipped,thermpres,dynload,poroelastic,qm,qp,q0, &
         scale_U,scale_Q,scale_T,scale_p,asperity_list,asperity_file,ascii,filename

    ! set default values

    preslipped = .false.
    dynload = .false.
    thermpres = .false.
    poroelastic = .false.
    qm = zero
    qp = zero
    q0 = zero
    scale_U = zero
    scale_Q = zero
    scale_T = zero
    scale_p = zero
    asperity_list = .false.
    asperity_file = .false.
    ascii = .false.

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=fault_list,iostat=stat)
    if (stat/=0) call error("Error reading namelist 'fault_list' in .in file",'read_fault')

    ! assign input variables to components of derived type
  
    flt%preslipped = preslipped
    flt%dynload = dynload
    flt%thermpres = thermpres
    flt%poroelastic = poroelastic
    flt%qm = qm
    flt%qp = qp
    flt%q0 = q0
    flt%scale_U = scale_U
    flt%scale_Q = scale_Q
    flt%scale_T = scale_T
    flt%scale_p = scale_p
    flt%aspi%asperity_file = asperity_file
    flt%aspi%ascii = ascii
    flt%aspi%filename = 'problems/' // adjustl(filename)

    ! asperity list

    ! set ndata = 3, 4, or 5 (sx0,sy0,sz0, and optionally Q and T)

    ndata = 3
    if (flt%state) ndata = ndata+1
    if (flt%temperature.or.flt%thermpres) ndata = ndata+1

    if (asperity_list) &
         flt%aspi%list = read_asperity_list(ninput,'fault',ndata)

    ! read static crack parameters if needed

    if (flt%preslipped) then
       allocate(flt%ps)
       call read_preslip(ninput,mdl%bm,flt%ps)
    end if

  end subroutine read_fault


  subroutine init_fault(necho,mdl,flt)
    ! INIT_FAULT initializes fault variables
    ! 
    ! Modified: 5 February 2007

    use constants, only : zero,one
    use model, only : model_type
    use asperity, only : assign_list_data,destroy_asperity_list
    use io, only : write_matlab,open_array_data,close_array_data,read_array_data

    implicit none

    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MDL = model variables
    ! FLT = fault variables

    integer(pin),intent(in) :: necho
    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt

    ! Internal Parameters:
    ! FILEUNIT = file unit for asperity file
    ! IFIELD = index of field in asperity list input

    integer(pin) :: fileunit,ifield

    ! allocate memory to fault arrays

    allocate(flt%sx(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%sx0(mdl%nx,mdl%ny))

    allocate(flt%sy(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%sy0(mdl%nx,mdl%ny))

    allocate(flt%sz(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%sz0(mdl%nx,mdl%ny))

    allocate(flt%U(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%V(mdl%nx,mdl%ny,mdl%ns))

    allocate(flt%tfront(mdl%nx,mdl%ny)) ! MAKE TFRONT OPTIONAL

    if (flt%poroelastic) allocate(flt%po(mdl%nx,mdl%ny))

    allocate(flt%Q( mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%dQ(mdl%nx,mdl%ny,mdl%ns))

    if (flt%thermpres.or.flt%temperature) allocate(flt%T0(mdl%nx,mdl%ny,mdl%ns))
    if (flt%thermpres) allocate(flt%p0(mdl%nx,mdl%ny,mdl%ns))

    ! initialize fields

    flt%rerr = zero
    flt%aerr = zero

    flt%sx = zero
    flt%sx0 = zero

    flt%sy = zero
    flt%sy0 = zero

    flt%sz = zero
    flt%sz0 = zero

    if (allocated(flt%tfront)) flt%tfront = zero

    if (allocated(flt%po)) flt%po = zero

    if (allocated(flt%T0)) flt%T0 = zero
    if (allocated(flt%p0)) flt%p0 = zero

    flt%Q  = zero
    flt%dQ = zero

    ! read asperity arrays from file

    if (flt%aspi%asperity_file) then
       call open_array_data(flt%aspi%filename,fileunit,flt%aspi%ascii)
       call read_array_data(fileunit,flt%sx0,flt%aspi%ascii,'sx0')
       call read_array_data(fileunit,flt%sy0,flt%aspi%ascii,'sy0')
       call read_array_data(fileunit,flt%sz0,flt%aspi%ascii,'sz0')
       if (flt%state) &
            call read_array_data(fileunit,flt%Q(:,:,1),flt%aspi%ascii,'Q')
       if (allocated(flt%T0)) &
            call read_array_data(fileunit,flt%T0(:,:,1),flt%aspi%ascii,'T')
       
       call close_array_data(fileunit)
    end if

    ! use list data to perturb fields

    call assign_list_data(flt%sx0,flt%aspi%list,1,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    call assign_list_data(flt%sy0,flt%aspi%list,2,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    call assign_list_data(flt%sz0,flt%aspi%list,3,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    ifield = 4
    if (flt%state) then
       call assign_list_data(flt%Q(:,:,1),flt%aspi%list,ifield,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       flt%Q  = spread(flt%Q(:,:,1),3,mdl%ns)
       ifield = ifield+1
    end if
    if (allocated(flt%T0)) then
       call assign_list_data(flt%T0(:,:,1),flt%aspi%list,ifield,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       flt%T0  = spread(flt%T0(:,:,1),3,mdl%ns)
    end if

    ! deallocate memory for asperity list

    call destroy_asperity_list(flt%aspi%list)

    ! allocate and initialize fields specific to formulation

    if (mdl%bm) then
       call init_fault_bm(mdl,flt)
    else
       call init_fault_im(mdl,flt)
    end if

    ! deallocate memory to preslip type
    
    if (associated(flt%ps)) deallocate(flt%ps)
    flt%ps => null()

    ! output variable values into matlab file

    call write_matlab(necho,'preslipped',flt%preslipped,'flt')
    call write_matlab(necho,'dynload',flt%dynload,'flt')
    call write_matlab(necho,'thermpres',flt%thermpres,'flt')
    call write_matlab(necho,'poroelastic',flt%poroelastic,'flt')
    if (flt%poroelastic) then
       call write_matlab(necho,'qm',flt%qm,'flt')
       call write_matlab(necho,'qp',flt%qp,'flt')
       call write_matlab(necho,'q0',flt%q0,'flt')
    end if
    if (mdl%method=='adaptive'.or.(mdl%method=='substep'.and.mdl%substep_method=='adaptive')) then
       call write_matlab(necho,'scale_U',flt%scale_U,'flt')
       call write_matlab(necho,'scale_Q',flt%scale_Q,'flt')
       call write_matlab(necho,'scale_T',flt%scale_T,'flt')
       call write_matlab(necho,'scale_p',flt%scale_p,'flt')
    end if

  end subroutine init_fault


  subroutine init_fault_im(mdl,flt)
    ! INIT_FAULT_IM initializes fault variables, identical materials case
    ! 
    ! Modified: 7 May 2007

    use constants, only : zero,two
    use model, only : model_type
    use preslip, only : init_preslip_im

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt

    ! allocate memory to fault arrays

    allocate(flt%Ux(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%Uy(mdl%nx,mdl%ny,mdl%ns))

    allocate(flt%Vx(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%Vy(mdl%nx,mdl%ny,mdl%ns))

    allocate(flt%fx(mdl%nx,mdl%ny,mdl%lf:mdl%uf))
    allocate(flt%fy(mdl%nx,mdl%ny,mdl%lf:mdl%uf))

    allocate(flt%fxi(mdl%nx,mdl%ny))
    allocate(flt%fyi(mdl%nx,mdl%ny))

    ! initialize fields

    flt%Ux = zero
    flt%Uy = zero

    flt%Vx = zero
    flt%Vy = zero

    flt%fx = zero
    flt%fy = zero

    flt%fxi = zero
    flt%fyi = zero

    ! set stress to ensure that fx=fy=0 at t=0 (assuming V=0)

    flt%sx = spread(flt%sx0,3,mdl%ns)
    flt%sy = spread(flt%sy0,3,mdl%ns)
    flt%sz = spread(flt%sz0,3,mdl%ns)

    ! calculate preslip fields

    if (flt%preslipped) &
         call init_preslip_im(mdl,flt%ps,flt%Ux(:,:,1),flt%Uy(:,:,1))

    ! spread values of initial slip and slip velocity

    flt%Ux = spread(flt%Ux(:,:,1),3,mdl%ns)
    flt%Uy = spread(flt%Uy(:,:,1),3,mdl%ns)

    flt%Vx = spread(flt%Vx(:,:,1),3,mdl%ns)
    flt%Vy = spread(flt%Vy(:,:,1),3,mdl%ns)

    flt%U = sqrt(flt%Ux**2+flt%Uy**2)
    flt%V = sqrt(flt%Vx**2+flt%Vy**2)

  end subroutine init_fault_im


  subroutine init_fault_bm(mdl,flt)
    ! INIT_FAULT_BM initializes fault variables, bimaterial case
    ! 
    ! Modified: 7 May 2007

    use constants, only : zero
    use model, only : model_type
    use preslip, only : init_preslip_bm

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt

    ! allocate memory to fault arrays

    allocate(flt%uxp(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%uxm(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%uyp(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%uym(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%uzp(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%uzm(mdl%nx,mdl%ny,mdl%ns))

    allocate(flt%vxp(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%vxm(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%vyp(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%vym(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%vzp(mdl%nx,mdl%ny,mdl%ns))
    allocate(flt%vzm(mdl%nx,mdl%ny,mdl%ns))

    allocate(flt%fxp(mdl%nx,mdl%ny,mdl%lf:mdl%uf))
    allocate(flt%fxm(mdl%nx,mdl%ny,mdl%lf:mdl%uf))
    allocate(flt%fyp(mdl%nx,mdl%ny,mdl%lf:mdl%uf))
    allocate(flt%fym(mdl%nx,mdl%ny,mdl%lf:mdl%uf))
    allocate(flt%fzp(mdl%nx,mdl%ny,mdl%lf:mdl%uf))
    allocate(flt%fzm(mdl%nx,mdl%ny,mdl%lf:mdl%uf))

    allocate(flt%fxpi(mdl%nx,mdl%ny))
    allocate(flt%fxmi(mdl%nx,mdl%ny))
    allocate(flt%fypi(mdl%nx,mdl%ny))
    allocate(flt%fymi(mdl%nx,mdl%ny))
    allocate(flt%fzpi(mdl%nx,mdl%ny))
    allocate(flt%fzmi(mdl%nx,mdl%ny))

    ! initialize fields

    flt%uxp = zero
    flt%uxm = zero
    flt%uyp = zero
    flt%uym = zero
    flt%uzp = zero
    flt%uzm = zero

    flt%vxp = zero
    flt%vxm = zero
    flt%vyp = zero
    flt%vym = zero
    flt%vzp = zero
    flt%vzm = zero

    flt%fxp = zero
    flt%fxm = zero
    flt%fyp = zero
    flt%fym = zero
    flt%fzp = zero
    flt%fzm = zero

    flt%fxpi = zero
    flt%fxmi = zero
    flt%fypi = zero
    flt%fymi = zero
    flt%fzpi = zero
    flt%fzmi = zero

    ! set stress to ensure that fx=fy=fz=0 at t=0 (assuming V=0)

    flt%sx = spread(flt%sx0,3,mdl%ns)
    flt%sy = spread(flt%sy0,3,mdl%ns) 
    flt%sz = spread(flt%sz0,3,mdl%ns)

    ! calculate preslip fields

    if (flt%preslipped) &
         call init_preslip_bm(mdl,flt%ps,flt%uxp(:,:,1),flt%uxm(:,:,1),flt%uyp(:,:,1),flt%uym(:,:,1),flt%uzp(:,:,1),flt%uzm(:,:,1))

    ! spread values of initial slip and slip velocity

    flt%uxp = spread(flt%uxp(:,:,1),3,mdl%ns)
    flt%uxm = spread(flt%uxm(:,:,1),3,mdl%ns)
    flt%uyp = spread(flt%uyp(:,:,1),3,mdl%ns)
    flt%uym = spread(flt%uym(:,:,1),3,mdl%ns)
    flt%uzp = spread(flt%uzp(:,:,1),3,mdl%ns)
    flt%uzm = spread(flt%uzm(:,:,1),3,mdl%ns)

    flt%vxp = spread(flt%vxp(:,:,1),3,mdl%ns)
    flt%vxm = spread(flt%vxm(:,:,1),3,mdl%ns)
    flt%vyp = spread(flt%vyp(:,:,1),3,mdl%ns)
    flt%vym = spread(flt%vym(:,:,1),3,mdl%ns)
    flt%vzp = spread(flt%vzp(:,:,1),3,mdl%ns)
    flt%vzm = spread(flt%vzm(:,:,1),3,mdl%ns)

    flt%U = sqrt((flt%uxp-flt%uxm)**2+(flt%uyp-flt%uym)**2+(flt%uzp-flt%uzm)**2)
    flt%V = sqrt((flt%vxp-flt%vxm)**2+(flt%vyp-flt%vym)**2+(flt%vzp-flt%vzm)**2)

  end subroutine init_fault_bm
  

  subroutine destroy_fault(flt)
    ! DESTROY_FAULT destroys derived type flt
    ! 
    ! Modified: 2 January 2007

    implicit none

    ! I/O Parameters:
    ! FLT = fault variables

    type(fault_type),intent(inout) :: flt

    ! deallocate memory assigned to allocatable arrays

    if (allocated(flt%sx)) deallocate(flt%sx)
    if (allocated(flt%sy)) deallocate(flt%sy)
    if (allocated(flt%sz)) deallocate(flt%sz)

    if (allocated(flt%sx0)) deallocate(flt%sx0)
    if (allocated(flt%sy0)) deallocate(flt%sy0)
    if (allocated(flt%sz0)) deallocate(flt%sz0)

    if (allocated(flt%U)) deallocate(flt%U)
    if (allocated(flt%V)) deallocate(flt%V)
       
    if (allocated(flt%Ux)) deallocate(flt%Ux)
    if (allocated(flt%Uy)) deallocate(flt%Uy)

    if (allocated(flt%Vx)) deallocate(flt%Vx)
    if (allocated(flt%Vy)) deallocate(flt%Vy)

    if (allocated(flt%fx)) deallocate(flt%fx)
    if (allocated(flt%fy)) deallocate(flt%fy)

    if (allocated(flt%fxi)) deallocate(flt%fxi)
    if (allocated(flt%fyi)) deallocate(flt%fyi)

    if (allocated(flt%uxp)) deallocate(flt%uxp)
    if (allocated(flt%uxm)) deallocate(flt%uxm)
    if (allocated(flt%uyp)) deallocate(flt%uyp)
    if (allocated(flt%uym)) deallocate(flt%uym)
    if (allocated(flt%uzp)) deallocate(flt%uzp)
    if (allocated(flt%uzm)) deallocate(flt%uzm)

    if (allocated(flt%vxp)) deallocate(flt%vxp)
    if (allocated(flt%vxm)) deallocate(flt%vxm)
    if (allocated(flt%vyp)) deallocate(flt%vyp)
    if (allocated(flt%vym)) deallocate(flt%vym)
    if (allocated(flt%vzp)) deallocate(flt%vzp)
    if (allocated(flt%vzm)) deallocate(flt%vzm)

    if (allocated(flt%fxp)) deallocate(flt%fxp)
    if (allocated(flt%fxm)) deallocate(flt%fxm)
    if (allocated(flt%fyp)) deallocate(flt%fyp)
    if (allocated(flt%fym)) deallocate(flt%fym)
    if (allocated(flt%fzp)) deallocate(flt%fzp)
    if (allocated(flt%fzm)) deallocate(flt%fzm)

    if (allocated(flt%fxpi)) deallocate(flt%fxpi)
    if (allocated(flt%fxmi)) deallocate(flt%fxmi)
    if (allocated(flt%fypi)) deallocate(flt%fypi)
    if (allocated(flt%fymi)) deallocate(flt%fymi)
    if (allocated(flt%fzpi)) deallocate(flt%fzpi)
    if (allocated(flt%fzmi)) deallocate(flt%fzmi)

    if (allocated(flt%tfront)) deallocate(flt%tfront)

    if (allocated(flt%po)) deallocate(flt%po)

    if (allocated(flt%Q)) deallocate(flt%Q)
    if (allocated(flt%dQ)) deallocate(flt%dQ)

    if (allocated(flt%T0)) deallocate(flt%T0)
    if (allocated(flt%p0)) deallocate(flt%p0)

  end subroutine destroy_fault


  subroutine set_stress_transfer(mdl,flt,stage)
    ! SET_STRESS_TRANSFER sets current value of stress transfer
    ! 
    ! Modified: 7 February 2007

    use model, only : model_type

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! STAGE = stage at which f is returned

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    integer(pin),intent(in) :: stage

    if (mdl%bm) then

       flt%fxpi = flt%fxp(:,:,stage)
       flt%fxmi = flt%fxm(:,:,stage)
       flt%fypi = flt%fyp(:,:,stage)
       flt%fymi = flt%fym(:,:,stage)
       flt%fzpi = flt%fzp(:,:,stage)
       flt%fzmi = flt%fzm(:,:,stage)

    else

       flt%fxi = flt%fx(:,:,stage)
       flt%fyi = flt%fy(:,:,stage)

    end if

  end subroutine set_stress_transfer


  subroutine interp_stress_transfer(flt,mdl,t1,t2,t,method)
    ! INTERP_STRESS_TRANSFER interpolates stress transfer over times
    ! within a single elastodynamic time step
    ! 
    ! Modified: 14 February 2007

    use constants, only : half,two
    use io, only : error
    use model, only : model_type

    implicit none

    ! I/O Parameters:
    ! FLT = fault variables
    ! MDL = model variables
    ! T1 = time at beginning of elastodynamic time step
    ! T2 = time at end of elastodynamic time step
    ! T = current time (between t1 and t2)
    ! METHOD = interpolation method

    type(fault_type),intent(inout) :: flt
    type(model_type),intent(in) :: mdl
    real(pr),intent(in) :: t1,t2,t
    character(*),intent(in) :: method

    ! Internal Parameters:
    ! Q = fraction of elastodynamic time step completed

    real(pr) :: q

    q = (t-t1)/(t2-t1)

    ! indices of f are:
    ! ... t-2dt t-dt t t+dt
    ! ...  -2    -1  0   1

    select case(method)
       
    case default

       call error('invalid interpolation method','interp_stress_transfer')

    case('constant') ! constant at f(t)
          
       if (mdl%bm) then
          flt%fxpi = flt%fxp(:,:,0)
          flt%fxmi = flt%fxm(:,:,0)
          flt%fypi = flt%fyp(:,:,0)
          flt%fymi = flt%fym(:,:,0)
          flt%fzpi = flt%fzp(:,:,0)
          flt%fzmi = flt%fzm(:,:,0)
       else
          flt%fxi  = flt%fx( :,:,0)
          flt%fyi  = flt%fy( :,:,0)
       end if
       
    case('linear_backward') ! linear using f(t-dt) and f(t)
          
       if (mdl%bm) then
          flt%fxpi = flt%fxp(:,:,0)+q*(flt%fxp(:,:,0)-flt%fxp(:,:,-1))
          flt%fxmi = flt%fxm(:,:,0)+q*(flt%fxm(:,:,0)-flt%fxm(:,:,-1))
          flt%fypi = flt%fyp(:,:,0)+q*(flt%fyp(:,:,0)-flt%fyp(:,:,-1))
          flt%fymi = flt%fym(:,:,0)+q*(flt%fym(:,:,0)-flt%fym(:,:,-1))
          flt%fzpi = flt%fzp(:,:,0)+q*(flt%fzp(:,:,0)-flt%fzp(:,:,-1))
          flt%fzmi = flt%fzm(:,:,0)+q*(flt%fzm(:,:,0)-flt%fzm(:,:,-1))
       else
          flt%fxi  = flt%fx( :,:,0)+q*(flt%fx( :,:,0)-flt%fx( :,:,-1))
          flt%fyi  = flt%fy( :,:,0)+q*(flt%fy( :,:,0)-flt%fy( :,:,-1))
       end if
       
    case('linear_forward') ! linear using f(t) and f(t+dt)
       
       if (mdl%bm) then
          flt%fxpi = flt%fxp(:,:,0)+q*(flt%fxp(:,:,1)-flt%fxp(:,:,0))
          flt%fxmi = flt%fxm(:,:,0)+q*(flt%fxm(:,:,1)-flt%fxm(:,:,0))
          flt%fypi = flt%fyp(:,:,0)+q*(flt%fyp(:,:,1)-flt%fyp(:,:,0))
          flt%fymi = flt%fym(:,:,0)+q*(flt%fym(:,:,1)-flt%fym(:,:,0))
          flt%fzpi = flt%fzp(:,:,0)+q*(flt%fzp(:,:,1)-flt%fzp(:,:,0))
          flt%fzmi = flt%fzm(:,:,0)+q*(flt%fzm(:,:,1)-flt%fzm(:,:,0))
       else
          flt%fxi  = flt%fx( :,:,0)+q*(flt%fx( :,:,1)-flt%fx( :,:,0))
          flt%fyi  = flt%fy( :,:,0)+q*(flt%fy( :,:,1)-flt%fy( :,:,0))
       end if
       
    case('quadratic') ! quadratic using f(t-dt), f(t), f(t+dt)
       
       if (mdl%bm) then
          flt%fxpi = flt%fxp(:,:,0)+half*q*(flt%fxp(:,:,1)-flt%fxp(:,:,-1))+ &
               half*q**2*(flt%fxp(:,:,-1)-two*flt%fxp(:,:,0)+flt%fxp(:,:,1))
          flt%fxmi = flt%fxm(:,:,0)+half*q*(flt%fxm(:,:,1)-flt%fxm(:,:,-1))+ &
               half*q**2*(flt%fxm(:,:,-1)-two*flt%fxm(:,:,0)+flt%fxm(:,:,1))
          flt%fypi = flt%fyp(:,:,0)+half*q*(flt%fyp(:,:,1)-flt%fyp(:,:,-1))+ &
               half*q**2*(flt%fyp(:,:,-1)-two*flt%fyp(:,:,0)+flt%fyp(:,:,1))
          flt%fymi = flt%fym(:,:,0)+half*q*(flt%fym(:,:,1)-flt%fym(:,:,-1))+ &
               half*q**2*(flt%fym(:,:,-1)-two*flt%fym(:,:,0)+flt%fym(:,:,1))
          flt%fzpi = flt%fzp(:,:,0)+half*q*(flt%fzp(:,:,1)-flt%fzp(:,:,-1))+ &
               half*q**2*(flt%fzp(:,:,-1)-two*flt%fzp(:,:,0)+flt%fzp(:,:,1))
          flt%fzmi = flt%fzm(:,:,0)+half*q*(flt%fzm(:,:,1)-flt%fzm(:,:,-1))+ &
               half*q**2*(flt%fzm(:,:,-1)-two*flt%fzm(:,:,0)+flt%fzm(:,:,1))
       else
          flt%fxi  = flt%fx( :,:,0)+half*q*(flt%fx( :,:,1)-flt%fx( :,:,-1))+ &
               half*q**2*(flt%fx( :,:,-1)-two*flt%fx( :,:,0)+flt%fx( :,:,1))
          flt%fyi  = flt%fy( :,:,0)+half*q*(flt%fy( :,:,1)-flt%fy( :,:,-1))+ &
               half*q**2*(flt%fy( :,:,-1)-two*flt%fy( :,:,0)+flt%fy( :,:,1))
       end if
       
    end select
       
  end subroutine interp_stress_transfer
     

end module fault
    
