! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module fft_routines
  ! FFT_ROUTINES contains variables and routines related to taking
  ! Fourier transforms of fields using the FFTW 3.x library (with 
  ! underscores added to subroutine calls for compatibility with IBM's
  ! compilers)
  ! 
  ! Modified: 4 May 2007

  use constants, only : pr,pin,int8

  implicit none

  ! FFTW constants

  INTEGER FFTW_R2HC
  PARAMETER (FFTW_R2HC=0)
  INTEGER FFTW_HC2R
  PARAMETER (FFTW_HC2R=1)
  INTEGER FFTW_DHT
  PARAMETER (FFTW_DHT=2)
  INTEGER FFTW_REDFT00
  PARAMETER (FFTW_REDFT00=3)
  INTEGER FFTW_REDFT01
  PARAMETER (FFTW_REDFT01=4)
  INTEGER FFTW_REDFT10
  PARAMETER (FFTW_REDFT10=5)
  INTEGER FFTW_REDFT11
  PARAMETER (FFTW_REDFT11=6)
  INTEGER FFTW_RODFT00
  PARAMETER (FFTW_RODFT00=7)
  INTEGER FFTW_RODFT01
  PARAMETER (FFTW_RODFT01=8)
  INTEGER FFTW_RODFT10
  PARAMETER (FFTW_RODFT10=9)
  INTEGER FFTW_RODFT11
  PARAMETER (FFTW_RODFT11=10)
  INTEGER FFTW_FORWARD
  PARAMETER (FFTW_FORWARD=-1)
  INTEGER FFTW_BACKWARD
  PARAMETER (FFTW_BACKWARD=+1)
  INTEGER FFTW_MEASURE
  PARAMETER (FFTW_MEASURE=0)
  INTEGER FFTW_DESTROY_INPUT
  PARAMETER (FFTW_DESTROY_INPUT=1)
  INTEGER FFTW_UNALIGNED
  PARAMETER (FFTW_UNALIGNED=2)
  INTEGER FFTW_CONSERVE_MEMORY
  PARAMETER (FFTW_CONSERVE_MEMORY=4)
  INTEGER FFTW_EXHAUSTIVE
  PARAMETER (FFTW_EXHAUSTIVE=8)
  INTEGER FFTW_PRESERVE_INPUT
  PARAMETER (FFTW_PRESERVE_INPUT=16)
  INTEGER FFTW_PATIENT
  PARAMETER (FFTW_PATIENT=32)
  INTEGER FFTW_ESTIMATE
  PARAMETER (FFTW_ESTIMATE=64)
  INTEGER FFTW_ESTIMATE_PATIENT
  PARAMETER (FFTW_ESTIMATE_PATIENT=128)
  INTEGER FFTW_BELIEVE_PCOST
  PARAMETER (FFTW_BELIEVE_PCOST=256)
  INTEGER FFTW_NO_DFT_R2HC
  PARAMETER (FFTW_NO_DFT_R2HC=512)
  INTEGER FFTW_NO_NONTHREADED
  PARAMETER (FFTW_NO_NONTHREADED=1024)
  INTEGER FFTW_NO_BUFFERING
  PARAMETER (FFTW_NO_BUFFERING=2048)
  INTEGER FFTW_NO_INDIRECT_OP
  PARAMETER (FFTW_NO_INDIRECT_OP=4096)
  INTEGER FFTW_ALLOW_LARGE_GENERIC
  PARAMETER (FFTW_ALLOW_LARGE_GENERIC=8192)
  INTEGER FFTW_NO_RANK_SPLITS
  PARAMETER (FFTW_NO_RANK_SPLITS=16384)
  INTEGER FFTW_NO_VRANK_SPLITS
  PARAMETER (FFTW_NO_VRANK_SPLITS=32768)
  INTEGER FFTW_NO_VRECURSE
  PARAMETER (FFTW_NO_VRECURSE=65536)
  INTEGER FFTW_NO_SIMD
  PARAMETER (FFTW_NO_SIMD=131072)
  INTEGER FFTW_NO_SLOW
  PARAMETER (FFTW_NO_SLOW=262144)
  INTEGER FFTW_NO_FIXED_RADIX_LARGE_N
  PARAMETER (FFTW_NO_FIXED_RADIX_LARGE_N=524288)
  INTEGER FFTW_ALLOW_PRUNING
  PARAMETER (FFTW_ALLOW_PRUNING=1048576)

  ! FFT_TYPE is a derived type containing fourier transform routines variables
  ! 
  ! PLAN_LEVEL = how much overhead spent on optimizing
  ! FFT routines before problem is run
  ! PF_XY = used by FFTW for forward transforms
  ! PI_XY = used by FFTW for inverse transforms
  ! PF_SY_X = used by FFTW for forward transform along x, symmetry in y direction
  ! PI_SY_X = used by FFTW for inverse transform along x, symmetry in y direction
  ! PF_SY_YE = used by FFTW for forward transform along y, even symmetry in y direction
  ! PF_SY_YO = used by FFTW for forward transform along y, odd symmetry in y direction
  ! PI_SY_YE = used by FFTW for inverse transform along y, even symmetry in y direction
  ! PI_SY_YO = used by FFTW for inverse transform along y, odd symmetry in y direction
  ! INV_FFT_FAC = normalization factor for inverse FFT
  ! EXP_FWRD = exponential for forward transform (single Fourier mode)
  ! EXP_BWRD = exponential for backward transform (single Fourier mode)

  type fft_type
     integer(int8) :: plan_level
     integer(int8) :: pf_xy,pi_xy,pf_sy_x,pf_sy_ye,pf_sy_yo,pi_sy_x,pi_sy_ye,pi_sy_yo
     real(pr) :: inv_fft_fac
     complex(pr) :: exp_fwrd,exp_bwrd
  end type fft_type

  ! Global Parameters:
  ! DUMMY_SPACE = temporary array for spatial-domain field
  ! DUMMY_FOURIER = temporary array for Fourier-domain field
  ! DUMMY_SPACE_X = temporary array for 1D spatial-domain field in x direction
  ! DUMMY_SPACE_Y = temporary array for 1D spatial-domain field in y direction
  ! DUMMY_FOURIER_X = temporary array for 1D Fourier-domain field in x direction
  ! DUMMY_FOURIER_Y = temporary array for 1D Fourier-domain field in y direction

  real(pr),dimension(:,:),allocatable :: dummy_space
  complex(pr),dimension(:,:),allocatable :: dummy_fourier
  real(pr),dimension(:),allocatable :: dummy_space_x,dummy_space_y,dummy_fourier_y
  complex(pr),dimension(:),allocatable :: dummy_fourier_x

contains


  subroutine read_fft(ninput,fft)
    ! READ_FFT reads in FFT variables from file
    ! 
    ! Modified: 22 January 2007

    use io, only : error

    implicit none

    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! FFT = FFT variables

    integer(pin),intent(in) :: ninput
    type(fft_type),intent(out) :: fft

    ! Internal Parameters:
    ! PLAN_LEVEL = how much overhead spent on optimizing FFT routines before problem is run
    ! STAT = I/O error flag

    integer(int8) :: plan_level
    integer(pin) :: stat

    ! make namelist of user input variables

    namelist /fft_list/ plan_level

    ! defaults

    plan_level = 2

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=fft_list,iostat=stat)
    if (stat>0) call error("Error reading namelist 'fft_list' in .in file",'read_fft')
    
    ! assign input variables to components of derived type

    fft%plan_level = plan_level

  end subroutine read_fft


  subroutine init_fft(mdl,fft)
    ! INIT_FFT initializes fft_routines variables
    ! 
    ! Modified: 7 February 2007

    use constants, only : one,two,img
    use model, only : model_type

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FFT = FFT variables

    type(model_type),intent(in) :: mdl
    type(fft_type),intent(inout) :: fft

    ! allocate memory for temporary arrays

    select case(mdl%dim)
    case(1,2)
       allocate(dummy_space(mdl%nx,mdl%ny))
       allocate(dummy_fourier(mdl%nx/2+1,mdl%ny))
    case(3)
       allocate(dummy_space(mdl%nx,mdl%ny))
       allocate(dummy_fourier(mdl%nx/2+1,mdl%ny))
       if (mdl%sym_y) then
          allocate(dummy_space_x(mdl%nx))
          allocate(dummy_fourier_x(mdl%nx/2+1))
          allocate(dummy_space_y(mdl%ny))
          allocate(dummy_fourier_y(mdl%ny))
       end if
    end select

    ! initialize FFT variables

    if (mdl%dim==1) then

       fft%exp_fwrd = exp(-img*mdl%k*mdl%x(1))
       fft%exp_bwrd = exp( img*mdl%k*mdl%x(1))

    else

       ! assign the factor needed to get the proper inverse transform
       
       if (mdl%sym_y) then 
          fft%inv_fft_fac = one/(real(mdl%nx,pr)*two*real(mdl%ny,pr))
       else
          fft%inv_fft_fac = one/(real(mdl%nx,pr)*real(mdl%ny,pr))
       end if
       
       ! plan fourier transforms (required by FFTW)
       
       call plan_fft(mdl,fft,mdl%sym_y)

    end if

  end subroutine init_fft


  subroutine destroy_fft(fft)
    ! DESTROY_FFT deallocates temporary arrays
    ! 
    ! Modified: 22 January 2007

    implicit none

    ! I/O Parameters:
    ! FFT = FFT variables

    type(fft_type),intent(inout) :: fft

    ! deallocate temporary arrays

    if (allocated(dummy_space)) deallocate(dummy_space)
    if (allocated(dummy_fourier)) deallocate(dummy_fourier) 

    if (allocated(dummy_space_x)) deallocate(dummy_space_x)
    if (allocated(dummy_space_y)) deallocate(dummy_space_y)
    if (allocated(dummy_fourier_x)) deallocate(dummy_fourier_x)  
    if (allocated(dummy_fourier_y)) deallocate(dummy_fourier_y)  

  end subroutine destroy_fft

  
  subroutine plan_fft(mdl,fft,sym_y)
    ! PLAN_FFT runs several test problems to choose the
    ! optimum FFT method
    ! 
    ! Modified: 7 February 2007

    use constants, only : real4,real8
    use model, only : model_type
    use io, only : error

    implicit none    

    ! I/O Parameters:
    ! MDL = model variables
    ! FFT = FFT variables
    ! SYM_Y = symmetry in y direction

    type(model_type),intent(in) :: mdl
    type(fft_type),intent(inout) :: fft
    logical,intent(in) :: sym_y

    ! FFTW requires that you 'plan' the FFTs before executing them.
    ! It often runs a few tests to pick the fastest routine given
    ! the array dimensions.  There is obviously a trade-off between
    ! the time spent planning and the time saved later, which is controlled
    ! by the following variables.  From least to most planning are:
    ! fftw_estimate, fftw_measure, fftw_patient, fftw_exhaustive

    ! assign FFTW parameter based on user input value between 1 and 4

    select case(fft%plan_level)
    case(1_int8)
       fft%plan_level = fftw_estimate
    case(2_int8)
       fft%plan_level = fftw_measure
    case(3_int8)
       fft%plan_level = fftw_patient
    case(4_int8)
       fft%plan_level = fftw_exhaustive
    end select

    ! single precision options - this requires installing the single-precision
    ! version of FFTW in addition to the standard double-precision version.
    ! This section is included if you wish to do this.  It should be as simple
    ! as copying what was done below for double precision, but replacing
    ! any instances of dfftw_ with dfftwf_.  This also requires linking to
    ! the single precision library in the Makefile.

    select case(pr)
    case(real4) ! single-precision transforms
       call error('Single precision FFTs not implemented','plan_fft')
       select case(mdl%dim)
       case(1)
       case(2)
       case(3)
       end select

    case(real8) ! double-precision transforms
       select case(mdl%dim)
       case(1)
       case(2)
          call dfftw_plan_dft_r2c_1d_(fft%pf_xy,mdl%nx,dummy_space(:,1),dummy_fourier(:,1),fft%plan_level)
          call dfftw_plan_dft_c2r_1d_(fft%pi_xy,mdl%nx,dummy_fourier(:,1),dummy_space(:,1),fft%plan_level)
       case(3)
          if (sym_y) then
             call dfftw_plan_r2r_1d_(fft%pf_sy_ye,mdl%ny,dummy_space_y,dummy_fourier_y,FFTW_REDFT10,fft%plan_level)
             call dfftw_plan_r2r_1d_(fft%pi_sy_ye,mdl%ny,dummy_fourier_y,dummy_space_y,FFTW_REDFT01,fft%plan_level)
             call dfftw_plan_r2r_1d_(fft%pf_sy_yo,mdl%ny,dummy_space_y,dummy_fourier_y,FFTW_RODFT10,fft%plan_level)
             call dfftw_plan_r2r_1d_(fft%pi_sy_yo,mdl%ny,dummy_fourier_y,dummy_space_y,FFTW_RODFT01,fft%plan_level)
             call dfftw_plan_dft_r2c_1d_(fft%pf_sy_x,mdl%nx,dummy_space_x,dummy_fourier_x,fft%plan_level)
             call dfftw_plan_dft_c2r_1d_(fft%pi_sy_x,mdl%nx,dummy_fourier_x,dummy_space_x,fft%plan_level)
          else
             call dfftw_plan_dft_r2c_2d_(fft%pf_xy,mdl%nx,mdl%ny,dummy_space,dummy_fourier,fft%plan_level)
             call dfftw_plan_dft_c2r_2d_(fft%pi_xy,mdl%nx,mdl%ny,dummy_fourier,dummy_space,fft%plan_level) 
          end if
       end select
    end select

  end subroutine plan_fft


  subroutine forward_fft(field_space,field_fourier,fft,sym_y,sym) 
    ! FORWARD_FFT forward FFTs a spatial field to Fourier domain
    ! 
    ! Modified: 22 January 2007

    use constants, only : img,onec

    implicit none
    
    ! I/O Parameters:
    ! FIELD_SPACE = field in space
    ! FIELD_FOURIER = field in Fourier domain
    ! FFT = FFT variables
    ! SYM_Y = symmetry in y direction
    ! SYM = even or odd symmetry

    real(pr),dimension(:,:),intent(in) :: field_space
    complex(pr),dimension(:,:),intent(out) :: field_fourier
    type(fft_type),intent(in) :: fft
    logical,intent(in) :: sym_y
    character(1),intent(in),optional :: sym

    ! Internal Parameters:
    ! PY = plan for transform in y direction
    ! I = index in x direction
    ! J = index in y direction
    ! TEMP = temporary array
    ! FAC = factor used for DCT and DST

    integer(int8) :: py
    integer(pin) :: i,j
    real(pr),dimension(size(field_space,1),size(field_space,2)) :: temp
    complex(pr) :: fac

    ! special treatment of single Fourier mode

    if (size(field_space,1)==1.and.size(field_space,2)==1) then

       field_fourier = fft%exp_fwrd*field_space

       return

    end if

    ! use dummy arrays as temporary holders for the fields to be transformed

    if (sym_y) then
       select case(sym)
       case('e')
          py = fft%pf_sy_ye
          fac = onec ! since exp(-ikx)=cos(kx)-i*sin(kx) [and taking coefficient of cosine]
       case('o')
          py = fft%pf_sy_yo
          fac = -img ! since exp(-ikx)=cos(kx)-i*sin(kx) [and taking coefficient of sine]
       end select
       do i = 1,size(field_space,1)
          dummy_space_y = field_space(i,:)
          call dfftw_execute_(py) ! dummy_space_y -> dummy_fourier_y
          temp(i,:) = dummy_fourier_y
       end do
       do j = 1,size(field_space,2)
          dummy_space_x = temp(:,j)
          call dfftw_execute_(fft%pf_sy_x) ! dummy_space_x -> dummy_fourier_x
          field_fourier(:,j) = dummy_fourier_x*fac
       end do
    else
       dummy_space = field_space
       call dfftw_execute_(fft%pf_xy)
       field_fourier = dummy_fourier
    end if

  end subroutine forward_fft


  subroutine inverse_fft(field_fourier,field_space,fft,sym_y,sym)
    ! INVERSE_FFT inverse FFTs a Fourier-domain field to space
    ! 
    ! Modified: 7 February 2007

    use constants, only : img,onec,zero

    implicit none

    ! I/O Parameters:
    ! FIELD_FOURIER = field in Fourier domain
    ! FIELD_SPACE = field in space
    ! FFT = FFT variables
    ! SYM_Y = symmetry in y direction
    ! SYM = even or odd symmetry

    complex(pr),dimension(:,:),intent(in) :: field_fourier
    real(pr),dimension(:,:),intent(inout) :: field_space
    type(fft_type),intent(in) :: fft
    logical,intent(in) :: sym_y
    character(1),intent(in),optional :: sym

    ! Internal Parameters:
    ! PY = plan for transform in y direction
    ! I = index in x direction
    ! J = index in y direction
    ! TEMP = temporary array
    ! FAC = factor used for DCT and DST

    integer(int8) :: py
    integer(pin) :: i,j
    real(pr),dimension(size(field_space,1),size(field_space,2)) :: temp
    complex(pr) :: fac

    if (size(field_space,1)==1.and.size(field_space,2)==1) then

       ! special treatment of single Fourier mode

       dummy_space = real(fft%exp_bwrd*field_fourier,pr)

    else

       ! use dummy arrays as temporary holders for the fields to be transformed, 
       ! multiply by the proper factor to recover the inverse transformed field amplitude
       
       if (sym_y) then

          select case(sym)
          case('e')
             py = fft%pi_sy_ye
             fac = onec
          case('o')
             py = fft%pi_sy_yo
             fac = img
          end select
          do j = 1,size(field_space,2)
             dummy_fourier_x = field_fourier(:,j)*fac
             call dfftw_execute_(fft%pi_sy_x) ! dummy_fourier_x -> dummy_space_x
             temp(:,j) = dummy_space_x
          end do
          do i = 1,size(field_space,1)
             dummy_fourier_y = temp(i,:)
             call dfftw_execute_(py) ! dummy_fourier_y -> dummy_space_y
             dummy_space(i,:) = dummy_space_y*fft%inv_fft_fac
          end do

       else

          dummy_fourier = field_fourier
          call dfftw_execute_(fft%pi_xy)
          dummy_space = dummy_space*fft%inv_fft_fac

       end if
       
    end if

    ! dummy_space now holds new value inverse transformed field and field_space 
    ! holds old value, so overwrite old value with new one

    field_space = dummy_space

  end subroutine inverse_fft


end module fft_routines
