! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module fourier

  ! FOURIER contains routines to manipulate fields in Fourier domain (to store history
  ! for use in convolutions, to calculate strains, etc.)
  ! 
  ! Modified: 26 January 2007

  use constants, only : pr,pin

  implicit none

contains


  subroutine fft_history(mdl,flt,cnv,fft,stage)
    ! FFT_HISTORY Fourier transforms displacement/slip or velocity
    ! for use in performing convolution, only history portion here
    ! 
    ! Modified: 24 January 2007

    use model, only : model_type
    use fault, only : fault_type
    use convolution, only : convolution_type
    use fft_routines, only : fft_type

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! CNV = convolution variables
    ! FFT = FFT variables
    ! STAGE = integration stage

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(in) :: flt
    type(convolution_type),intent(inout) :: cnv
    type(fft_type),intent(in) :: fft
    integer(pin),intent(in) :: stage

    ! return if history isn't stored

    if (cnv%formulation=='static'.or.cnv%formulation=='quasidynamic') return

    ! forward FFT displacement/slip U or velocity V and store in history vector, 
    ! shifting the contents to make room for the new value (if needed)
    
    if (mdl%bm) then
       
       select case(cnv%formulation)
       case('displacement')
          call fft_history_bm( &
               flt%uxp(:,:,stage),flt%uxm(:,:,stage), &
               flt%uyp(:,:,stage),flt%uym(:,:,stage), &
               flt%uzp(:,:,stage),flt%uzm(:,:,stage), &
               mdl,cnv,fft)
       case('velocity')
          call fft_history_bm( &
               flt%vxp(:,:,stage),flt%vxm(:,:,stage), &
               flt%vyp(:,:,stage),flt%vym(:,:,stage), &
               flt%vzp(:,:,stage),flt%vzm(:,:,stage), &
               mdl,cnv,fft)
       end select
       
    else
       
       select case(cnv%formulation)
       case('displacement')
          call fft_history_im( &
               flt%Ux(:,:,stage),flt%Uy(:,:,stage), &
               mdl,cnv,fft)
       case('velocity')
          call fft_history_im( &
               flt%Vx(:,:,stage),flt%Vy(:,:,stage), &
               mdl,cnv,fft)
       end select
       
    end if

  end subroutine fft_history


  subroutine fft_history_im(field_x,field_y,mdl,cnv,fft)
    ! FFT_HISTORY_IM Fourier transforms spatial field for later storage of its Fourier coefficients
    ! in history vector.  This version is for the identical materials case.
    ! 
    ! Modified: 24 January 2007

    use constants, only : zeroc,zero
    use model, only : model_type
    use convolution, only : convolution_type
    use fft_routines, only : fft_type,forward_fft

    implicit none
    
    ! I/O Parameters:
    ! FIELD_X = component of spatial field in x direction
    ! FIELD_Y = component of spatial field in y direction
    ! MDL = model variables
    ! CNV = convolution variables
    ! FFT = FFT variables

    real(pr),dimension(:,:),intent(in) :: field_x,field_y
    type(model_type),intent(in) :: mdl
    type(convolution_type),intent(inout) :: cnv
    type(fft_type),intent(in) :: fft

    select case(mdl%dim)

    case(1)

       cnv%Hx(1,1) = cmplx(field_x(1,1),zero,pr)
       if (mdl%transverse_slip) &
            cnv%Hy(1,1) = cmplx(field_y(1,1),zero,pr)

    case(2,3)

       if (mdl%sym_y) then
          cnv%Hx(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
          call forward_fft(field_x,cnv%Hx(:,1:cnv%nky-1),fft,mdl%sym_y,'e') ! even about y=y0
       else
          call forward_fft(field_x,cnv%Hx               ,fft,mdl%sym_y)
       end if

       if (mdl%transverse_slip) then

          if (mdl%sym_y) then
             cnv%Hy(:,1) = zeroc ! zero wavenumber has no contribution for odd symmetry
             call forward_fft(field_y,cnv%Hy(:,2:cnv%nky),fft,mdl%sym_y,'o') ! odd about y=y0
          else
             call forward_fft(field_y,cnv%Hy             ,fft,mdl%sym_y)
          end if

       end if

    end select

  end subroutine fft_history_im


  subroutine fft_history_bm(field_xp,field_xm,field_yp,field_ym,field_zp,field_zm,mdl,cnv,fft)
    ! FFT_HISTORY_BM Fourier transforms spatial field for later storage of its Fourier coefficients
    ! in history vector.  This version is for the bimaterial case.
    ! 
    ! Modified: 24 January 2007

    use constants, only : zeroc,zero
    use model, only : model_type
    use convolution, only : convolution_type
    use fft_routines, only : fft_type,forward_fft

    implicit none
    
    ! I/O Parameters:
    ! FIELD_XP = component of spatial field in x direction, plus direction
    ! FIELD_XM = component of spatial field in x direction, minus direction
    ! FIELD_YP = component of spatial field in y direction, plus direction
    ! FIELD_YM = component of spatial field in y direction, minus direction
    ! FIELD_ZP = component of spatial field in z direction, plus direction
    ! FIELD_ZM = component of spatial field in z direction, minus direction
    ! MDL = model variables
    ! CNV = convolution variables
    ! FFT = FFT variables

    real(pr),dimension(:,:),intent(in) :: field_xp,field_xm,field_yp,field_ym,field_zp,field_zm
    type(model_type),intent(in) :: mdl
    type(convolution_type),intent(inout) :: cnv
    type(fft_type),intent(in) :: fft

    select case(mdl%dim)

    case(1)

       cnv%Hxp(1,1) = cmplx(field_xp(1,1),zero,pr)
       cnv%Hxm(1,1) = cmplx(field_xm(1,1),zero,pr)

       if (mdl%transverse_slip) then
          cnv%Hyp(1,1) = cmplx(field_yp(1,1),zero,pr)
          cnv%Hym(1,1) = cmplx(field_ym(1,1),zero,pr)
       else
          cnv%Hyp(1,1) = cmplx(field_yp(1,1),zero,pr)
       end if

       if (mdl%opening) then
          cnv%Hzp(1,1) = cmplx(field_zp(1,1),zero,pr)
          cnv%Hzm(1,1) = cmplx(field_zm(1,1),zero,pr)
       else
          cnv%Hzp(1,1) = cmplx(field_zp(1,1),zero,pr)
       end if

    case(2,3)

       if (mdl%sym_y) then
          cnv%Hxp(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
          cnv%Hxm(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
          call forward_fft(field_xp,cnv%Hxp(:,1:cnv%nky-1),fft,mdl%sym_y,'e') ! even about y=y0
          call forward_fft(field_xm,cnv%Hxm(:,1:cnv%nky-1),fft,mdl%sym_y,'e') ! even about y=y0
       else
          call forward_fft(field_xp,cnv%Hxp,fft,mdl%sym_y)
          call forward_fft(field_xm,cnv%Hxm,fft,mdl%sym_y)
       end if

       if (mdl%sym_y) then
          cnv%Hyp(:,1) = zeroc ! zero wavenumber has no contribution for odd symmetry
          call forward_fft(field_yp,cnv%Hyp(:,2:cnv%nky),fft,mdl%sym_y,'o') ! odd about y=y0
       else
          call forward_fft(field_yp,cnv%Hyp             ,fft,mdl%sym_y)
       end if

       if (mdl%transverse_slip) then
          if (mdl%sym_y) then
             cnv%Hym(:,1) = zeroc ! zero wavenumber has no contribution for odd symmetry
             call forward_fft(field_ym,cnv%Hym(:,2:cnv%nky),fft,mdl%sym_y,'o') ! odd about y=y0
          else
             call forward_fft(field_ym,cnv%Hym             ,fft,mdl%sym_y)
          end if
       end if

       if (mdl%sym_y) then
          cnv%Hzp(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
          call forward_fft(field_zp,cnv%Hzp(:,1:cnv%nky-1),fft,mdl%sym_y,'e') ! even about y=y0
       else
          call forward_fft(field_zp,cnv%Hzp               ,fft,mdl%sym_y)
       end if

       if (mdl%opening) then
          if (mdl%sym_y) then
             cnv%Hzm(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
             call forward_fft(field_zm,cnv%Hzm(:,1:cnv%nky-1),fft,mdl%sym_y,'e') ! even about y=y0
          else
             call forward_fft(field_zm,cnv%Hzm               ,fft,mdl%sym_y)
          end if
       end if

    end select

  end subroutine fft_history_bm


  subroutine store_history(mdl,cnv,mp,update_history)
    ! STORE_HISTORY stores fields at current time step in history vectors
    ! 
    ! Modified: 25 January 2007

    use model, only : model_type
    use convolution, only : convolution_type
    use mpi_routines, only : mpi_type,master,scatterv,MPI_CPLX_PR
    use mpi

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! CNV = convolution variables
    ! MP = MPI variables
    ! UPDATE_HISTORY = overwrite history or shift it

    type(model_type),intent(in) :: mdl
    type(convolution_type),intent(inout) :: cnv
    type(mpi_type),intent(in) :: mp
    logical,intent(in) :: update_history

    ! Internal Parameters:
    ! J = index of ky
    ! IERR = MPI error flag
    ! COUNTS = MPI message sizes
    ! DISPLS = MPI message displacements

    integer :: j,ierr
    integer,dimension(0:mp%np-1) :: counts,displs

    ! return if history isn't stored

    if (cnv%formulation=='static'.or.cnv%formulation=='quasidynamic') return

    ! prepare for scatter/gather

    call scatterv(cnv%nkx,mp%np,counts,displs)

    ! scatter Fourier transforms of current step's contribution to history from master to slaves

    if (mdl%bm) then
       
       do j = 1,cnv%nky

          call MPI_Scatterv(cnv%Hxp(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hxp(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Hxm(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hxm(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Hyp(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hyp(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Hym(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hym(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Hzp(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hzp(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Hzm(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hzm(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)

       end do
       
    else

       do j = 1,cnv%nky

          call MPI_Scatterv(cnv%Hx(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hx(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)
          call MPI_Scatterv(cnv%Hy(:,j),counts,displs,MPI_CPLX_PR, &
               cnv%Hy(:,j),counts(mp%myid),MPI_CPLX_PR,master,MPI_COMM_WORLD,ierr)

       end do
       
    end if
       
    ! store result in history vector, either shifting or overwriting with new value

    if (mdl%bm) then
       call store_history_bm(mdl,cnv,update_history)
    else
       call store_history_im(mdl,cnv,update_history)
    end if

  end subroutine store_history


  subroutine store_history_im(mdl,cnv,update)
    ! STORE_HISTORY_IM 
    !
    ! Modified: 25 January 2007

    use model, only : model_type
    use convolution, only : convolution_type

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! CNV = convolution variables
    ! UPDATE = whether or not to update/shift the history vector
    !      T = shift contents of history vector, adding new Fourier
    !          coefficients as first element and moving all the other
    !          elements backward in the vector
    !      F = do not shift elements, instead just overwrite the first
    !          element corresponding to the current time step

    type(model_type),intent(in) :: mdl
    type(convolution_type),intent(inout) :: cnv
    logical,intent(in) :: update

    ! Internal Parameters:
    ! I = index in kx direction
    ! J = index in ky direction

    integer(pin) :: i,j

    ! For each mode, store new Fourier coefficients in history vector.  If the Fourier coefficient for this
    ! time step is new, then shift the contents of the history vector by one element and place the new value
    ! as the first element.  This is accomplished using the intrinsic eoshift (end-off shift). Otherwise,
    ! overwrite the first element of the history vector with the new value.

    do j = 1,cnv%nky
       do i = cnv%li,cnv%ui

          if (cnv%H(i,j)%n==0) cycle
          
          if (update) then

             cnv%H(i,j)%Hx = eoshift(cnv%H(i,j)%Hx,-1,cnv%Hx(i,j))
             if (mdl%transverse_slip) &
                  cnv%H(i,j)%Hy = eoshift(cnv%H(i,j)%Hy,-1,cnv%Hy(i,j))

          else

             cnv%H(i,j)%Hx(0) = cnv%Hx(i,j)
             if (mdl%transverse_slip) &
                  cnv%H(i,j)%Hy(0) = cnv%Hy(i,j)

          end if

       end do
    end do

  end subroutine store_history_im


  subroutine store_history_bm(mdl,cnv,update)
    ! STORE_HISTORY_BM
    ! 
    ! Modified: 24 January 2007

    use model, only : model_type
    use convolution, only : convolution_type

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! CNV = convolution variables
    ! UPDATE = whether or not to update/shift the history vector
    !      T = shift contents of history vector, adding new Fourier coefficients as first element 
    !          and moving all the other elements backward in the vector
    !      F = do not shift elements, instead just overwrite the first element corresponding to the 
    !          current time step

    type(model_type),intent(in) :: mdl
    type(convolution_type),intent(inout) :: cnv
    logical,intent(in) :: update

    ! Internal Parameters:
    ! I = index in kx direction
    ! J = index in ky direction

    integer(pin) :: i,j

    ! For each mode, store new Fourier coefficients in history vector.  If the Fourier coefficient for this
    ! time step is new, then shift the contents of the history vector by one element and place the new value
    ! as the first element.  This is accomplished using the intrinsic eoshift (end-off shift). Otherwise,
    ! overwrite the first element of the history vector with the new value.

    do j = 1,cnv%nky
       do i = cnv%li,cnv%ui
          
          if (cnv%H(i,j)%n==0) cycle
          
          if (update) then

             cnv%H(i,j)%Hxp = eoshift(cnv%H(i,j)%Hxp,-1,cnv%Hxp(i,j))
             cnv%H(i,j)%Hxm = eoshift(cnv%H(i,j)%Hxm,-1,cnv%Hxm(i,j))

             if (mdl%transverse_slip) then
                cnv%H(i,j)%Hyp = eoshift(cnv%H(i,j)%Hyp,-1,cnv%Hyp(i,j))
                cnv%H(i,j)%Hym = eoshift(cnv%H(i,j)%Hym,-1,cnv%Hym(i,j))
             else
                cnv%H(i,j)%Hyp = eoshift(cnv%H(i,j)%Hyp,-1,cnv%Hyp(i,j))
             end if

             if (mdl%opening) then
                cnv%H(i,j)%Hzp = eoshift(cnv%H(i,j)%Hzp,-1,cnv%Hzp(i,j))
                cnv%H(i,j)%Hzm = eoshift(cnv%H(i,j)%Hzm,-1,cnv%Hzm(i,j))
             else
                cnv%H(i,j)%Hzp = eoshift(cnv%H(i,j)%Hzp,-1,cnv%Hzp(i,j))
             end if

          else

             cnv%H(i,j)%Hxp(0) = cnv%Hxp(i,j)
             cnv%H(i,j)%Hxm(0) = cnv%Hxm(i,j)

             if (mdl%transverse_slip) then
                cnv%H(i,j)%Hyp(0) = cnv%Hyp(i,j)
                cnv%H(i,j)%Hym(0) = cnv%Hym(i,j)
             else
                cnv%H(i,j)%Hyp(0) = cnv%Hyp(i,j)
             end if

             if (mdl%opening) then
                cnv%H(i,j)%Hzp(0) = cnv%Hzp(i,j)
                cnv%H(i,j)%Hzm(0) = cnv%Hzm(i,j)
             else
                cnv%H(i,j)%Hzp(0) = cnv%Hzp(i,j)
             end if

          end if
          
       end do
    end do
  
  end subroutine store_history_bm
        

  subroutine fft_current_U(mdl,flt,cnv,fft,stage)
    ! FFT_CURRENT_U Fourier transforms displacement/slip
    ! 
    ! Modified: 23 January 2007

    use constants, only : zero,zeroc
    use model, only : model_type
    use fault, only : fault_type
    use convolution, only : convolution_type
    use fft_routines, only : fft_type,forward_fft

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! CNV = convolution variables
    ! FFT = FFT variables
    ! STAGE = integration stage

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(in) :: flt
    type(convolution_type),intent(inout) :: cnv
    type(fft_type),intent(in) :: fft
    integer(pin),intent(in) :: stage

    if (mdl%bm) then

       ! bimaterial version

       select case(mdl%dim)
       case(1)
          cnv%Dxp(1,1) = cmplx(flt%uxp(1,1,stage),zero,pr)
          cnv%Dxm(1,1) = cmplx(flt%uxm(1,1,stage),zero,pr)
          cnv%Dyp(1,1) = cmplx(flt%uyp(1,1,stage),zero,pr)
          cnv%Dym(1,1) = cmplx(flt%uym(1,1,stage),zero,pr)
          cnv%Dzp(1,1) = cmplx(flt%uzp(1,1,stage),zero,pr)
          cnv%Dzm(1,1) = cmplx(flt%uzm(1,1,stage),zero,pr)
       case(2,3)
          if (mdl%sym_y) then
             cnv%Dxp(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
             cnv%Dxm(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
             cnv%Dyp(:,1      ) = zeroc ! zero wavenumber has no contribution for odd symmetry
             cnv%Dym(:,1      ) = zeroc ! zero wavenumber has no contribution for odd symmetry
             cnv%Dzp(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
             cnv%Dzm(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
             call forward_fft(flt%uxp(:,:,stage),cnv%Dxp(:,1:cnv%nky-1),fft,mdl%sym_y,sym='e')
             call forward_fft(flt%uxm(:,:,stage),cnv%Dxm(:,1:cnv%nky-1),fft,mdl%sym_y,sym='e')
             call forward_fft(flt%uyp(:,:,stage),cnv%Dyp(:,2:cnv%nky  ),fft,mdl%sym_y,sym='o')
             call forward_fft(flt%uym(:,:,stage),cnv%Dym(:,2:cnv%nky  ),fft,mdl%sym_y,sym='o')
             call forward_fft(flt%uzp(:,:,stage),cnv%Dzp(:,1:cnv%nky-1),fft,mdl%sym_y,sym='e')
             call forward_fft(flt%uzm(:,:,stage),cnv%Dzm(:,1:cnv%nky-1),fft,mdl%sym_y,sym='e')
          else
             call forward_fft(flt%uxp(:,:,stage),cnv%Dxp,fft,mdl%sym_y)
             call forward_fft(flt%uxm(:,:,stage),cnv%Dxm,fft,mdl%sym_y)
             call forward_fft(flt%uyp(:,:,stage),cnv%Dyp,fft,mdl%sym_y)
             call forward_fft(flt%uym(:,:,stage),cnv%Dym,fft,mdl%sym_y)
             call forward_fft(flt%uzp(:,:,stage),cnv%Dzp,fft,mdl%sym_y)
             call forward_fft(flt%uzm(:,:,stage),cnv%Dzm,fft,mdl%sym_y)
          end if
       end select

    else

       ! identical materials version

       select case(mdl%dim)
       case(1)
          if (mdl%transverse_slip) then
             cnv%Dx(1,1) = cmplx(flt%Ux(1,1,stage),zero,pr)
             cnv%Dy(1,1) = cmplx(flt%Uy(1,1,stage),zero,pr)
          else
             cnv%Dx(1,1) = cmplx(flt%Ux(1,1,stage),zero,pr)
          end if
       case(2,3)
          if (mdl%transverse_slip) then
             if (mdl%sym_y) then
                cnv%Dx(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
                call forward_fft(flt%Ux(:,:,stage),cnv%Dx(:,1:cnv%nky-1),fft,mdl%sym_y,sym='e') ! even about y=y0
                cnv%Dy(:,1) = zeroc ! zero wavenumber has no contribution for odd symmetry
                call forward_fft(flt%Uy(:,:,stage),cnv%Dy(:,2:cnv%nky),fft,mdl%sym_y,sym='o') ! odd about y=y0
             else
                call forward_fft(flt%Ux(:,:,stage),cnv%Dx,fft,mdl%sym_y)
                call forward_fft(flt%Uy(:,:,stage),cnv%Dy,fft,mdl%sym_y)
             end if
          else
             if (mdl%sym_y) then
                cnv%Dx(:,cnv%nky) = zeroc ! Nyquist wavenumber has no contribution for even symmetry
                call forward_fft(flt%Ux(:,:,stage),cnv%Dx(:,1:cnv%nky-1),fft,mdl%sym_y,sym='e') ! even about y=y0
             else
                call forward_fft(flt%Ux(:,:,stage),cnv%Dx,fft,mdl%sym_y)
             end if
          end if
       end select

    end if

  end subroutine fft_current_U


  subroutine inverse_fft_F(mdl,flt,cnv,fft,stage)
    ! INVERSE_FFT_F inverse Fourier transforms stress transfer functional
    ! 
    ! Modified: 7 February 2007

    use model, only : model_type
    use fault, only : fault_type
    use convolution, only : convolution_type
    use fft_routines, only : fft_type,inverse_fft

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! CNV = convolution variables   
    ! FFT = FFT variables
    ! STAGE = stage at which f is returned

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(convolution_type),intent(inout) :: cnv
    type(fft_type),intent(in) :: fft
    integer(pin),intent(in) :: stage

    ! invert Fourier transform

    if (mdl%bm) then

       select case(mdl%dim)
       case(1,2)
          call inverse_fft(cnv%Fxp,flt%fxp(:,:,stage),fft,mdl%sym_y)
          call inverse_fft(cnv%Fxm,flt%fxm(:,:,stage),fft,mdl%sym_y)
          call inverse_fft(cnv%Fyp,flt%fyp(:,:,stage),fft,mdl%sym_y)
          call inverse_fft(cnv%Fym,flt%fym(:,:,stage),fft,mdl%sym_y)
          call inverse_fft(cnv%Fzp,flt%fzp(:,:,stage),fft,mdl%sym_y)
          call inverse_fft(cnv%Fzm,flt%fzm(:,:,stage),fft,mdl%sym_y)
       case(3)
          if (mdl%sym_y) then
             call inverse_fft(cnv%Fxp(:,1:cnv%nky-1),flt%fxp(:,:,stage),fft,mdl%sym_y,sym='e')
             call inverse_fft(cnv%Fxm(:,1:cnv%nky-1),flt%fxm(:,:,stage),fft,mdl%sym_y,sym='e')
             call inverse_fft(cnv%Fyp(:,2:cnv%nky  ),flt%fyp(:,:,stage),fft,mdl%sym_y,sym='o')
             call inverse_fft(cnv%Fym(:,2:cnv%nky  ),flt%fym(:,:,stage),fft,mdl%sym_y,sym='o')
             call inverse_fft(cnv%Fzp(:,1:cnv%nky-1),flt%fzp(:,:,stage),fft,mdl%sym_y,sym='e')
             call inverse_fft(cnv%Fzm(:,1:cnv%nky-1),flt%fzm(:,:,stage),fft,mdl%sym_y,sym='e')
          else
             call inverse_fft(cnv%Fxp,flt%fxp(:,:,stage),fft,mdl%sym_y)
             call inverse_fft(cnv%Fxm,flt%fxm(:,:,stage),fft,mdl%sym_y)
             call inverse_fft(cnv%Fyp,flt%fyp(:,:,stage),fft,mdl%sym_y)
             call inverse_fft(cnv%Fym,flt%fym(:,:,stage),fft,mdl%sym_y)
             call inverse_fft(cnv%Fzp,flt%fzp(:,:,stage),fft,mdl%sym_y)
             call inverse_fft(cnv%Fzm,flt%fzm(:,:,stage),fft,mdl%sym_y)
          end if
       end select

    else
       
       select case(mdl%dim)
       case(1,2)
          call inverse_fft(cnv%Fx,flt%fx(:,:,stage),fft,mdl%sym_y)
          call inverse_fft(cnv%Fy,flt%fy(:,:,stage),fft,mdl%sym_y)
       case(3)
          if (mdl%sym_y) then
             call inverse_fft(cnv%Fx(:,1:cnv%nky-1),flt%fx(:,:,stage),fft,mdl%sym_y,sym='e')
             call inverse_fft(cnv%Fy(:,2:cnv%nky  ),flt%fy(:,:,stage),fft,mdl%sym_y,sym='o')
          else
             call inverse_fft(cnv%Fx,flt%fx(:,:,stage),fft,mdl%sym_y)
             call inverse_fft(cnv%Fy,flt%fy(:,:,stage),fft,mdl%sym_y)
          end if
       end select

    end if

  end subroutine inverse_fft_F


  function strain(mdl,cnv,fft,mp) result(e)
    ! STRAIN calculates fault-parallel strains on one side of fault using
    ! spectral methods to take spatial derivatives
    ! 
    ! Modified: 23 January 2007

    use constants, only : zeroc,img,zero,half,one,two
    use model, only : model_type
    use convolution, only : convolution_type,wavenumber
    use fft_routines, only : fft_type,inverse_fft

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! CNV = convolution variables
    ! FFT = FFT variables
    ! MP = side of fault
    ! E = sum of fault-parallel strains

    type(model_type),intent(in) :: mdl
    type(convolution_type),intent(in) :: cnv
    type(fft_type),intent(in) :: fft
    integer(pin),intent(in) :: mp
    real(pr),dimension(mdl%nx,mdl%ny) :: e

    ! Internal Parameters:
    ! I = index in the kx direction
    ! J = index in the ky direction
    ! K = wavenumber amplitude (k**2=kx**2+ky**2)
    ! KM = wavenumber product array 
    !      km(1,1) = kx**2
    !      km(1,2) = km(2,1) = kx*ky
    !      km(2,2) = ky**2
    ! KX = wavenumber in x direction
    ! KY = wavenumber in y direction
    ! EXX = xx component of strain
    ! EYY = yy component of strain

    integer(pin) :: i,j
    real(pr) :: k,km(2,2),kx,ky
    real(pr),dimension(mdl%nx,mdl%ny) :: exx,eyy
    complex(pr),dimension(cnv%nkx,cnv%nky) :: Ex,Ey

    if (mdl%bm) then

       ! bimaterial version

       do j = 1,cnv%nky
          do i = 1,cnv%nkx
             
             ! calculate wavenumber array
             
             call wavenumber(mdl,cnv,i,j,k,km,kx,ky)

             ! calculate Fourier transforms of strains, Ex and Ey
             
             select case(mp)
             case(-1)
                Ex(i,j) = kx*cnv%Dxm(i,j)
                if (mdl%transverse_slip) then
                   Ey(i,j) = ky*cnv%Dym(i,j)
                else
                   Ey(i,j) = zeroc
                end if
             case(1)
                Ex(i,j) = kx*cnv%Dxp(i,j)
                if (mdl%transverse_slip) then
                   Ey(i,j) = ky*cnv%Dyp(i,j)
                else
                   Ey(i,j) = zeroc
                end if
             end select

          end do
       end do

       Ex = img*Ex
       Ey = img*Ey
   
    else

       ! identical materials version

       do j = 1,cnv%nky
          do i = 1,cnv%nkx
             
             ! calculate wavenumber array
             
             call wavenumber(mdl,cnv,i,j,k,km,kx,ky)

             ! calculate Fourier transforms of strains, Ex and Ey
             
             Ex(i,j) = kx*cnv%Dx(i,j)
             Ey(i,j) = ky*cnv%Dy(i,j)

          end do
       end do

       ! multiply by necessary factor, including 1/2 to go from slip to displacement
    
       select case(mp)
       case(-1) ! minus side of fault (gives extra '-' sign)
          Ex = -half*img*Ex
          Ey = -half*img*Ey
       case(1) ! plus side of fault
          Ex = half*img*Ex
          Ey = half*img*Ey
       end select
    
    end if

    ! invert Fourier transforms to obtain strains
    
    select case(mdl%dim)
    case(1)
       exx(1,1) = real(Ex(1,1),pr)
       eyy(1,1) = real(Ey(1,1),pr)
    case(2)
       if (mdl%mode==0)  then
          call inverse_fft(Ex,exx,fft,mdl%sym_y)
          call inverse_fft(Ey,eyy,fft,mdl%sym_y)
       else
          call inverse_fft(Ex,exx,fft,mdl%sym_y)
          eyy = zero
       end if
    case(3)
       if (mdl%sym_y) then
          call inverse_fft(Ex(:,1:cnv%nky-1),exx,fft,mdl%sym_y,sym='e')
          call inverse_fft(Ey(:,2:cnv%nky  ),eyy,fft,mdl%sym_y,sym='o')
       else
          call inverse_fft(Ex,exx,fft,mdl%sym_y)
          call inverse_fft(Ey,eyy,fft,mdl%sym_y)
       end if
    end select
    
    ! sum fault-parallel strains

    e = exx+eyy

  end function strain


  subroutine set_normal_displacement(mdl,krn,flt,cnv,fft,stage)
    ! SET_NORMAL_DISPLACEMENT sets normal displacement in bimaterial problem
    ! using no opening constraint (in Fourier domain)
    ! 
    ! Modified: 23 January 2007

    use constants, only : zeroc,img,two
    use io, only : error
    use model, only : model_type
    use kernel, only : kernel_type
    use fault, only : fault_type
    use convolution, only : convolution_type
    use fft_routines, only : fft_type,inverse_fft

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! KRN = kernel variables
    ! FLT = fault variables
    ! CNV = convolution variables
    ! FFT = FFT variables
    ! STAGE = integration stage

    type(model_type),intent(in) :: mdl
    type(kernel_type),intent(in) :: krn
    type(fault_type),intent(inout) :: flt
    type(convolution_type),intent(inout) :: cnv
    type(fft_type),intent(in) :: fft
    integer(pin),intent(in) :: stage

    ! return if identical materials

    if (.not.mdl%bm) return

    ! return if opening is permitted

    if (mdl%opening) return

    ! solve for fault-normal displacement in Fourier domain

    select case(mdl%mode)
    case(0)
       call error('setting normal displacement not yet implemented in mixed mode or 3D','normal_displacement')
    case(1,2)
       cnv%Dzp = -img* &
            (mdl%mup*(krn%C12S+two-mdl%etap)*cnv%Dxp-mdl%mum*(krn%C12S+two-mdl%etam)*cnv%Dxm)/ &
            ((mdl%mup+mdl%mum)*krn%C22S)
       cnv%Dzm = cnv%Dzp
    case(3)
       cnv%Dzp = zeroc
       cnv%Dzm = zeroc
    end select

    ! invert transforms

    select case(mdl%dim)
    case(1,2)
       call inverse_fft(cnv%Dzp,flt%uzp(:,:,stage),fft,mdl%sym_y)
       call inverse_fft(cnv%Dzm,flt%uzm(:,:,stage),fft,mdl%sym_y)
    case(3)
       call error('setting normal displacement not yet implemented in 3D','normal_displacement')
    end select
    
  end subroutine set_normal_displacement


end module fourier
