! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module friction

  ! FRICTION contains variables and routines related to applying
  ! friction laws on the fault surface, coupling slip, slip velocity,
  ! state variable, and stress fields
  ! 
  ! Modified: 31 January 2007

  use constants, only : pr,pin
  use slipweak, only : slipweak_type
  use dugdale, only : dugdale_type
  use ratestate, only : ratestate_type

  implicit none

  ! FRICTION_TYPE is a derived type containing variables related to friction law
  !
  ! Parameters:
  ! FRICTION = type of friction to be applied
  ! SW = slip-weakening variables
  ! RS = rate-and-state variables
  ! DG = Dugdale variables
  ! (for these derived types, memory is only allocated if needed)
  ! SLIP_X = extent of slipping region in x direction
  ! SLIP_X0 = center of slipping region in x direction
  ! The following variables are used for temporary storage when using Newton's method:
  ! ST0 = velocity-independent strength
  ! Q = state variable
  ! SZ = stress in z-direction
  ! CV = radiation damping coefficient
  ! SFX = sum of load and stress transfer in x direction (sfx = sx0+fx)
  ! SFY = sum of load and stress transfer in y direction (sfy = sy0+fy)
  ! I = index in x direction
  ! J = index in y direction
  
  type friction_type
     character(64) :: friction
     type(slipweak_type),pointer :: sw=>null()
     type(dugdale_type),pointer :: dg=>null()
     type(ratestate_type),pointer :: rs=>null()
     real(pr) :: slip_x,slip_x0,st0,Q,sz,cV,sfx,sfy
     integer(pin) :: i,j
  end type friction_type

contains


  subroutine read_friction(ninput,flt,fri)
    ! READ_FRICTION reads in friction variables from file
    ! 
    ! Modified: 31 January 2007

    use constants, only : zero,one
    use io, only : error
    use fault, only : fault_type
    use slipweak, only : read_slipweak
    use dugdale, only : read_dugdale
    use ratestate, only : read_ratestate

    implicit none

    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! FLT  fault variables
    ! FRI = friction variables

    integer(pin),intent(in) :: ninput
    type(fault_type),intent(inout) :: flt
    type(friction_type),intent(out) :: fri

    ! Internal Parameters:
    ! FRICTION = type of friction to be applied
    ! SLIP_X = extent of slipping region in x direction
    ! SLIP_X0 = center of slipping region in x direction
    ! STAT = I/O error flag

    character(64) :: friction
    real(pr) :: slip_x,slip_x0
    integer(pin) :: stat

    ! make namelist of user input variables

    namelist /friction_list/ friction,slip_x,slip_x0

    ! set default values

    friction = 'slipweak'
    slip_x = huge(one)
    slip_x0 = zero

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=friction_list,iostat=stat)
    if (stat/=0) call error("Error reading namelist 'friction_list' in .in file",'read_friction')

    ! assign input variables to components of derived type

    fri%friction = friction
    fri%slip_x = slip_x
    fri%slip_x0 = slip_x0

    ! read specific friction law, after checking if this is a valid law;
    ! also set whether or not state and temperature fields are required by friction law

    select case(fri%friction)
    case default
       call error("Invalid parameter 'friction' in namelist 'friction_list' in .in file",'read_friction')
    case('none','constfV')
       flt%state = .false.
       flt%temperature = .false.
    case('slipweak')
       allocate(fri%sw)
       call read_slipweak(ninput,fri%sw)
       flt%state = .false.
       flt%temperature = .false.
    case('ratestate')
       allocate(fri%rs)
       call read_ratestate(ninput,fri%rs,flt%temperature)
       flt%state = .true.
    case('dugdale')
       allocate(fri%dg)
       call read_dugdale(ninput,fri%dg)
       flt%state = .false.
       flt%temperature = .false.
    end select

  end subroutine read_friction


  subroutine init_friction(necho,mdl,fri,flt)
    ! INIT_FRICTION initializes friction variables
    ! 
    ! Modified: 21 November 2006

    use model, only : model_type
    use fault, only : fault_type
    use io, only : write_matlab
    use slipweak, only : init_slipweak,set_strength_sw
    use dugdale, only : init_dugdale
    use ratestate, only : init_ratestate

    implicit none

    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MDL = model variables
    ! FRI = friction variables

    integer(pin),intent(in) :: necho
    type(model_type),intent(inout) :: mdl
    type(friction_type),intent(inout) :: fri
    type(fault_type),intent(inout) :: flt

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction
    ! STAGE = integration stage

    integer(pin) :: i,j,stage

    ! initialize specific friction law

    select case(fri%friction)

    case('slipweak')

       call init_slipweak(necho,mdl,fri%sw)

       ! set fault strength as 'state' variable for use in regularized friction law

       do j = 1,mdl%ny
          do i = 1,mdl%nx
             do stage = 1,mdl%ns
                flt%Q(i,j,stage) = set_strength_sw(flt%U(i,j,stage),flt%sz(i,j,stage),i,j,mdl%x(i),fri%sw)
             end do
          end do
       end do

    case('ratestate')

       call init_ratestate(necho,mdl,fri%rs)

    case('dugdale')

       call init_dugdale(necho,mdl,fri%dg)

    end select

    ! output variable values into matlab file

    call write_matlab(necho,'friction',fri%friction,'fri')
    call write_matlab(necho,'slip_x',fri%slip_x,'fri')

  end subroutine init_friction


  subroutine destroy_friction(fri)
    ! DESTROY_FRICTION destroys derived type fri
    ! 
    ! Modified: 17 July 2006

    use slipweak, only : destroy_slipweak
    use dugdale, only : destroy_dugdale
    use ratestate, only : destroy_ratestate

    implicit none
    
    ! I/O Parameters:
    ! FRI = friction variables    
    
    type(friction_type),intent(inout) :: fri
    
    ! destroy derived types for specific friction law and
    ! deallocate memory associated with them

    select case(fri%friction)
    case('slipweak')
       call destroy_slipweak(fri%sw)
       if (associated(fri%sw)) deallocate(fri%sw)
       fri%sw => null()
    case('ratestate')
       call destroy_ratestate(fri%rs)
       if (associated(fri%rs)) deallocate(fri%rs)
       fri%rs => null()
    case('dugdale')
       call destroy_dugdale(fri%dg)
       if (associated(fri%dg)) deallocate(fri%dg)
       fri%dg => null()
    end select
    
  end subroutine destroy_friction


  subroutine strengthV(fri,V,stV,dstdV,err)
    ! STRENGTHV returns the velocity-dependent strength and derivative
    ! of strength with respect to slip velocity at a point
    ! 
    ! Modified: 22 September 2006

    use constants, only : zero
    use ratestate, only : strength_rs

    implicit none

    ! I/O Parameters:
    ! FRI = friction variables
    ! V = slip velocity
    ! STV = velocity-dependent strength
    ! DSTDV = derivative of strength with respect to slip velocity
    ! ERR = error flag (non-zero if strength isn't defined at particular V)

    type(friction_type),intent(in) :: fri
    real(pr),intent(in) :: V
    real(pr),intent(out) :: stV,dstdV
    integer(pin),intent(out) :: err

    select case(fri%friction)
    case default
       stV = zero
       dstdV = zero
       err = 0
    case('ratestate')
       call strength_rs(fri%Q,fri%sz,V,fri%i,fri%j,stV,dstdV,err,fri%rs)
    end select

  end subroutine strengthV


end module friction
