! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module friction_routines

  ! FRICTION_ROUTINES contains subroutines to call friction-law
  ! solvers for identical material or bimaterial problems
  ! 
  ! Modified: 20 April 2007

  use constants, only : pr,pin

  implicit none

contains


  subroutine normal_friction(mdl,flt,cnv,fft,stage)
    ! NORMAL_FRICTION calculates effective normal stress by adding contributions from
    ! 1. normal stress loads (in absence of slip)
    ! 2. normal stress changes due to elastic mismatch, assuming fault walls remain in contact
    ! 3. pore pressure changes due to poroelastic response
    ! 4. pore pressure changes due to thermal pressurization.  
    ! and also fault-normal velocities
    ! 
    ! Modified: 22 February 2007

    use constants, only : zero
    use model, only : model_type
    use fault, only : fault_type
    use convolution, only : convolution_type
    use fourier, only : strain
    use fft_routines, only : fft_type
    !use io, only : warning

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! CNV = convolution variables
    ! FFT = FFT variables
    ! STAGE = integration substep

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(convolution_type),intent(in) :: cnv
    type(fft_type),intent(in) :: fft
    integer(pin),intent(in) :: stage

    ! Internal Parameters:
    ! EM = fault-parallel strains, minus side
    ! EP = fault-parallel strains, plus side
    ! MUM = (far-field) shear modulus, minus side
    ! MUP = (far-field) shear modulus, plus side

    real(pr),dimension(mdl%nx,mdl%ny) :: em,ep
    real(pr) :: mum,mup

    ! fault-normal velocity, assuming fault walls remain in contact

    if (mdl%bm) then
       flt%vzp(:,:,stage) = (flt%fzpi-flt%fzmi)/(mdl%cvnp+mdl%cvnm)
       flt%vzm(:,:,stage) = flt%vzp(:,:,stage)
    end if
       
    ! normal stress changes from slip, assuming fault walls remain in contact

    if (mdl%bm) then
       flt%sz(:,:,stage) = flt%fzpi-mdl%cvnp*flt%vzp(:,:,stage)
    else
       flt%sz(:,:,stage) = zero
    end if

    ! poroelastic fault zone response

    if (flt%poroelastic) then
    
       ! shear moduli of both sides
    
       if (mdl%bm) then
          mum = mdl%mum
          mup = mdl%mup
       else
          mum = mdl%mu
          mup = mdl%mu
       end if
       
       ! calculate fault-parallel strains on each side
       
       em = strain(mdl,cnv,fft,mp=-1)
       ep = strain(mdl,cnv,fft,mp= 1)
       
       ! calculate poro pressure change from fault-parallel strains and normal stress

       flt%po = -flt%q0*flt%sz(:,:,stage)-flt%qp*mup*ep-flt%qm*mum*em

       ! add this to effective normal stress
    
       flt%sz(:,:,stage) = flt%sz(:,:,stage)+flt%po

    end if

    ! add pore pressure changes from thermal pressurization

    if (allocated(flt%p0)) &
         flt%sz(:,:,stage) = flt%sz(:,:,stage)+flt%p0(:,:,stage)

    ! add normal load

    flt%sz(:,:,stage) = flt%sz(:,:,stage)+flt%sz0

    ! check if effect normal stress is tensile

    !if (any(flt%sz(:,:,stage)>zero)) &
    !     call warning('effective normal stress turned tensile, fault opening','normal_stress')

    ! switch boundary conditions to free surface

    if (mdl%bm) then
       where (flt%sz(:,:,stage)>zero)
          flt%sz(:,:,stage) = zero
          flt%vzp(:,:,stage) =  (flt%sz0+flt%fzpi)/mdl%cvnp
          flt%vzm(:,:,stage) = -(flt%sz0+flt%fzmi)/mdl%cvnm
       end where
    end if
  
  end subroutine normal_friction
 

  subroutine solve_friction(flt,mdl,fri,stage,accept)
    ! SOLVE_FRICTION solves the friction law coupled with elasticity
    ! for unknown velocity and stress
    ! 
    ! Modified: 20 April 2007

    use constants, only : zero,half,one
    use model, only : model_type
    use fault, only : fault_type
    use friction, only : friction_type
    use friction_solver, only : solve_elasticity_im,solve_elasticity_bm
    use ratestate, only : check_state_rs
    use slipweak, only : strength_sw

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! FRI = friction variables
    ! STAGE = integration sub-step
    ! ACCEPT = accept solution

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(friction_type),intent(inout) :: fri
    integer(pin),intent(in) :: stage
    logical,intent(out) :: accept

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction
    ! ST0 = velocity-independent portion of strength

    integer(pin) :: i,j
    real(pr) :: st0,V

    ! loop over fault

    do j = 1,mdl%ny
       do i = 1,mdl%nx

          ! check if value of state is acceptable, otherwise reject step

          select case(fri%friction)
          case default
             accept = .true.
          case('ratestate')
             accept = check_state_rs(flt%Q(i,j,stage),fri%rs)
          end select

          ! if unacceptable, reject this step and exit routine
          ! (usually to try again with smaller step size)

          if (.not.accept) return

          ! calculate strength, but only if velocity-independent

          select case(fri%friction)
          case('slipweak')
             st0 = strength_sw(flt%U(i,j,stage),flt%Q(i,j,stage),flt%sz(i,j,stage), &
                  i,j,mdl%x(i),mdl%t,fri%sw)
          case('constfV')
             ! constant f,V: set V=1 and calculate new shear resistance
             ! (assuming f=1 and no resistance in y direction)
             if (mdl%bm) then
                flt%vxp(i,j,stage) =  half
                flt%vxm(i,j,stage) = -half
                flt%vyp(i,j,stage) = zero
                flt%vym(i,j,stage) = zero
             else
                flt%Vx(i,j,stage) = one
                flt%Vy(i,j,stage) = zero
             end if
             flt%sx(i,j,stage) = -flt%sz(i,j,stage)
             flt%sy(i,j,stage) = zero
             cycle
          case default
             st0 = zero
          end select

          ! imposed slip velocity region

          !if (abs(mdl%x(i))<1.5_pr) then
          !   V = 0.5_pr
          !   if (mdl%bm) then
          !      flt%vxp(i,j,stage) = (flt%fxpi(i,j)-flt%fxmi(i,j))/(mdl%cvsp+mdl%cvsm)+ &
          !           mdl%cvsm/(mdl%cvsp+mdl%cvsm)*V
          !      flt%vxm(i,j,stage) = (flt%fxpi(i,j)-flt%fxmi(i,j))/(mdl%cvsp+mdl%cvsm)- &
          !           mdl%cvsp/(mdl%cvsp+mdl%cvsm)*V
          !      flt%vyp(i,j,stage) = 0._pr
          !      flt%vym(i,j,stage) = 0._pr
          !      flt%sx(i,j,stage) = flt%sx0(i,j)+flt%fxpi(i,j)-mdl%cvsp*flt%vxp(i,j,stage)
          !      flt%sy(i,j,stage) = flt%sy0(i,j)+flt%fypi(i,j)-mdl%cvsp*flt%vyp(i,j,stage)
          !   else
          !      flt%Vx(i,j,stage) = V
          !      flt%Vy(i,j,stage) = 0._pr
          !      flt%sx(i,j,stage) = flt%sx0(i,j)+flt%fxi(i,j)-mdl%cV*flt%Vx(i,j,stage)
          !      flt%sy(i,j,stage) = flt%sy0(i,j)+flt%fyi(i,j)-mdl%cV*flt%Vy(i,j,stage)
          !   end if
          !   cycle
          !end if

          ! solve elasticity coupled with possibly velocity-dependent friction law

          if (mdl%bm) then
             call solve_elasticity_bm(fri,i,j, &
                  st0,flt%Q(i,j,stage),flt%sz(i,j,stage), &
                  flt%vxp(i,j,stage),flt%vxm(i,j,stage), &
                  flt%vyp(i,j,stage),flt%vym(i,j,stage), &
                  flt%sx(i,j,stage),flt%sy(i,j,stage), &
                  flt%sx0(i,j),flt%sy0(i,j), &
                  flt%fxpi(i,j),flt%fxmi(i,j), &
                  flt%fypi(i,j),flt%fymi(i,j), &
                  mdl%cvsp,mdl%cvsm,mdl%transverse_slip)
          else
             call solve_elasticity_im(fri,i,j, &
                  st0,flt%Q(i,j,stage),flt%sz(i,j,stage), &
                  flt%Vx(i,j,stage),flt%Vy(i,j,stage), &
                  flt%sx(i,j,stage),flt%sy(i,j,stage), &
                  flt%sx0(i,j),flt%sy0(i,j), &
                  flt%fxi(i,j),flt%fyi(i,j), &
                  mdl%cV,mdl%transverse_slip)
          end if
          
       end do
    end do

    ! calculate magnitude of slip velocity

    if (mdl%bm) then
       flt%V(:,:,stage) = sqrt( &
            (flt%vxp(:,:,stage)-flt%vxm(:,:,stage))**2+ &
            (flt%vyp(:,:,stage)-flt%vym(:,:,stage))**2+ &
            (flt%vzp(:,:,stage)-flt%vzm(:,:,stage))**2)
    else
       flt%V(:,:,stage) = sqrt(flt%Vx(:,:,stage)**2+flt%Vy(:,:,stage)**2)
    end if

  end subroutine solve_friction


  subroutine state_rate(flt,mdl,fri,stage)
    ! STATE_RATE solves for the state-rate
    ! 
    ! Modified: 1 February 2007

    use model, only : model_type
    use fault, only : fault_type
    use friction, only : friction_type
    use slipweak, only : evolution_sw
    use ratestate, only : evolution_rs

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! FRI = friction variables
    ! STAGE = integration sub-step

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(friction_type),intent(in) :: fri
    integer(pin),intent(in) :: stage

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction

    integer(pin) :: i,j

    ! loop over fault

    do j = 1,mdl%ny
       do i = 1,mdl%nx

          ! solve friction law at each point

          select case(fri%friction)
             
          case('slipweak') ! slip-weakening ('state' being strength)

             flt%dQ(i,j,stage) = evolution_sw(flt%U(i,j,stage),flt%V(i,j,stage),flt%Q(i,j,stage), &
                  flt%sz(i,j,stage),i,j,mdl%x(i),mdl%t,fri%sw)

          case('ratestate') ! rate-and-state
          
             if (flt%temperature) then
                flt%dQ(i,j,stage) = evolution_rs(flt%V(i,j,stage),flt%Q(i,j,stage),i,j,fri%rs,flt%T0(i,j,stage))
             else
                flt%dQ(i,j,stage) = evolution_rs(flt%V(i,j,stage),flt%Q(i,j,stage),i,j,fri%rs)
             end if

          end select

       end do
    end do

  end subroutine state_rate


  subroutine static_friction(flt,mdl,fri,stage,g,m,p)
    ! STATIC_FRICTION evaluates the misfit between friction law and elasticity
    ! 
    ! Modified: 2 January 2007

    use constants, only : zero
    use model, only : model_type
    use fault, only : fault_type
    use friction, only : friction_type
    use slipweak, only : strength_sw
    use friction_static, only : static_elasticity_im,static_elasticity_bm

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! FRI = friction variables
    ! STAGE = integration sub-step
    ! G = misfit
    ! M = minimum index of slipping zone in x direction
    ! P = maximum index of slipping zone in x direction

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(friction_type),intent(inout) :: fri
    integer(pin),intent(in) :: stage
    real(pr),dimension(:),intent(out) :: g
    integer(pin),intent(in) :: m,p

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction
    ! ST = strength
    ! MISFIT = misfit function (in array format)
    ! N = size of misfit

    integer(pin) :: i,j,n
    real(pr) :: st
    real(pr),dimension(:,:,:),allocatable :: misfit

    ! allocate memory to misfit array

    if (mdl%bm) then
       allocate(misfit(m:p,mdl%ny,4))
    else
       allocate(misfit(m:p,mdl%ny,2))
    end if

    ! set stresses

    if (mdl%bm) then
       flt%sx(:,:,stage) = flt%sx0+flt%fxpi
       flt%sy(:,:,stage) = flt%sy0+flt%fypi
    else
       flt%sx(:,:,stage) = flt%sx0+flt%fxi
       flt%sy(:,:,stage) = flt%sy0+flt%fyi
    end if

    ! set vector slip

    if (mdl%bm) then
       flt%U(:,:,stage) = sqrt( &
            (flt%uxp(:,:,stage)-flt%uxm(:,:,stage))**2+ &
            (flt%uyp(:,:,stage)-flt%uym(:,:,stage))**2+ &
            (flt%uzp(:,:,stage)-flt%uzm(:,:,stage))**2 )
    else
       flt%U(:,:,stage) = sqrt(flt%Ux(:,:,stage)**2+flt%Uy(:,:,stage)**2)
    end if

    ! loop over slipping portion of fault and calculate misfit

    do j = 1,mdl%ny
       do i = m,p

          ! calculate strength

          select case(fri%friction)
          case('slipweak')
             st = strength_sw(flt%U(i,j,stage),flt%Q(i,j,stage),flt%sz(i,j,stage), &
                  i,j,mdl%x(i),mdl%t,fri%sw)
          case default
             st = zero
          end select
          
          ! evaluate misfit

          if (mdl%bm) then
             call static_elasticity_bm(st, &
                  flt%sx(i,j,stage),flt%sy(i,j,stage), &
                  flt%fxpi(i,j),flt%fxmi(i,j), &
                  flt%fypi(i,j),flt%fymi(i,j), &
                  flt%uxp(i,j,stage),flt%uxm(i,j,stage), &
                  flt%uyp(i,j,stage),flt%uym(i,j,stage), &
                  mdl%transverse_slip,misfit(i,j,:))
          else
             call static_elasticity_im(st, &
                  flt%sx(i,j,stage),flt%sy(i,j,stage), &
                  flt%Ux(i,j,stage),flt%Uy(i,j,stage), &
                  mdl%transverse_slip,misfit(i,j,:))
          end if
          
       end do
    end do

    ! pack misfit array into vector

    n = (p-m+1)*mdl%ny

    if (mdl%bm) then
       if (mdl%transverse_slip) then
          g(    1:  n) = reshape(misfit(:,:,1),shape(g(1:n)))
          g(  n+1:2*n) = reshape(misfit(:,:,2),shape(g(1:n)))
          g(2*n+1:3*n) = reshape(misfit(:,:,3),shape(g(1:n)))
          g(3*n+1:4*n) = reshape(misfit(:,:,4),shape(g(1:n)))
       else
          g(    1:  n) = reshape(misfit(:,:,1),shape(g(1:n)))
          g(  n+1:2*n) = reshape(misfit(:,:,2),shape(g(1:n)))
       end if
    else
       if (mdl%transverse_slip) then
          g(  1:  n) = reshape(misfit(:,:,1),shape(g(1:n)))
          g(n+1:2*n) = reshape(misfit(:,:,2),shape(g(1:n)))
       else
          g(1:n) = reshape(misfit(:,:,1),shape(g(1:n)))
       end if
    end if
    
    ! deallocate memory for misfit array
    
    deallocate(misfit)

  end subroutine static_friction


end module friction_routines
