! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module friction_solver

  ! FRICTION_SOLVER contains routines to solve friction laws
  ! 
  ! Modified: 8 May 2007

  use constants, only : pr,pin

  implicit none


contains


  subroutine solve_elasticity_im(fri,i,j,st0,Q,sz,Vx,Vy,sx,sy,sx0,sy0,fx,fy,cV,transverse_slip)
    ! SOLVE_ELASTICITY_IM solves elasticity equation with stress equal to strength
    ! (when slipping).  This version is for identical materials.
    ! 
    ! Modified: 8 May 2007

    use constants, only : zero,one,two,three
    use friction, only : friction_type
    use utilities, only : newton

    implicit none

    ! I/O Parameters:
    ! FRI = friction variables
    ! I = index in x direction
    ! J = index in y direction
    ! ST0 = portion of shear strength independent of slip velocity
    ! Q = state variable
    ! SZ = stress in z direction
    ! VX = slip velocity in x direction
    ! VY = slip velocity in y direction
    ! SX = stress in x direction
    ! SY = stress in y direction
    ! SX0 = load in x direction
    ! SY0 = load in y direction
    ! FX = stress transfer in x direction
    ! FY = stress transfer in y direction
    ! CV = coefficient of radiation damping
    ! TRANSVERSE_SLIP = scalar or vector slip

    type(friction_type),intent(inout) :: fri
    integer(pin),intent(in) :: i,j
    real(pr),intent(in) :: Q,st0,sz,sx0,sy0,fx,fy,cV
    real(pr),intent(inout) :: Vx,Vy,sx,sy
    logical,intent(in) :: transverse_slip
    
    ! Internal Parameters:
    ! SL = stress if fault is locked
    ! NEWTON_X = independent variables for Newton's method
    ! NEWTON_XSCALE = scale of independent variables for convergence of Newton's method
    ! NEWTON_FSCALE = scale of function for convergence of Newton's method
    ! PARAM = parameters required for function that is solved using Newton's method
    ! NEWTON_NMAX = maximum number of interations allowed to solve equations with Newton's method
    ! NEWTON_TOLX = maximum tolerated error in independent variables
    ! NEWTON_TOLF = maximum tolerated error in dependent variables

    real(pr) :: sl
    real(pr),dimension(:),allocatable :: newton_x,newton_xscale,newton_fscale,param
    integer(pin) :: newton_nmax
    real(pr) :: newton_tolx,newton_tolf
   
    ! set iteration limit and tolerances

    newton_nmax = 1000
    newton_tolx = epsilon(one)
    newton_tolf = epsilon(one)**(two/three)
 
    if (transverse_slip) then

       ! assume fault is locked, calculate stress
       
       sl = sqrt((sx0+fx)**2+(sy0+fy)**2)
    
       ! compare strength to elastic shear stress if fault were locked,
       ! allow failure if stress exceeds strength
          
       if (fri%friction/='ratestate'.and.sl<st0) then
          
          ! stress less than strength, fault locked
          
          Vx = zero
          sx = sx0+fx
          Vy = zero
          sy = sy0+fy
          
       else
          
          ! stress set to strength, fault slips;
          ! solution via Newton's method--see the subroutine elastic for further description
          
          allocate(newton_x(4))

          ! addition of eps to Vx avoids V=0 problem in some rate-and-state laws

          newton_x(1) = Vx+epsilon(zero) 
          newton_x(2) = sx
          newton_x(3) = Vy
          newton_x(4) = sy
          
          allocate(newton_xscale(4))

          newton_xscale(1) = max(abs(Vx),abs(Vy))
          newton_xscale(2) = max(abs(sx),abs(sy))
          newton_xscale(3) = max(abs(Vx),abs(Vy))
          newton_xscale(4) = max(abs(sx),abs(sy))

          allocate(newton_fscale(4))

          newton_fscale(2) = max(abs(fx),abs(cV*Vx))
          newton_fscale(3) = max(abs(fy),abs(cV*Vy))
          newton_fscale(1) = newton_fscale(2)**2+newton_fscale(3)**2
          newton_fscale(4) = abs(newton_fscale(2)*Vy)+abs(newton_fscale(3)*Vx)

          fri%i = i
          fri%j = j
          fri%st0 = st0
          fri%Q = Q
          fri%sz = sz
          fri%cV = cV
          fri%sfx = sx0+fx
          fri%sfy = sy0+fy
          
          allocate(param(size(transfer(fri,param))))
          param = transfer(fri,param)

          call newton(newton_x,newton_nmax,newton_tolx,newton_tolf,param,elastic,newton_xscale,newton_fscale)
          
          Vx = newton_x(1)
          sx = newton_x(2)
          Vy = newton_x(3)
          sy = newton_x(4)
          
          deallocate(newton_x)
          deallocate(newton_xscale)
          deallocate(newton_fscale)
          deallocate(param)

       end if

    else

       ! lock fault in y-direction
       
       Vy = zero
       sy = sy0+fy
       
       ! assume fault is locked, calculate stress
       
       sl = abs(sx0+fx)
       
       ! compare strength to elastic shear stress if fault were locked,
       ! allow failure if stress exceeds strength
       
       if (fri%friction/='ratestate'.and.sl<st0) then
          
          ! stress less than strength, fault locked
          
          Vx = zero
          sx = sx0+fx
          
       else
          
          ! stress set to strength, fault slips;
          ! solution via Newton's method--see the subroutine elastic for further description
          
          allocate(newton_x(2))

          ! addition of eps to Vx avoids V=0 problem in some rate-and-state laws

          newton_x(1) = Vx+epsilon(zero)
          newton_x(2) = sx
          
          allocate(newton_xscale(2))

          newton_xscale(1) = abs(Vx)
          newton_xscale(2) = abs(sx)

          allocate(newton_fscale(2))

          newton_fscale(1) = max(abs(fx),abs(cV*Vx))
          newton_fscale(2) = newton_fscale(1)

          fri%i = i
          fri%j = j
          fri%st0 = st0
          fri%Q = Q
          fri%sz = sz
          fri%cV = cV
          fri%sfx = sx0+fx
          
          allocate(param(size(transfer(fri,param))))
          param = transfer(fri,param)

          call newton(newton_x,newton_nmax,newton_tolx,newton_tolf,param,elastic,newton_xscale,newton_fscale)
          
          Vx = newton_x(1)
          sx = newton_x(2)
          
          deallocate(newton_x)
          deallocate(newton_xscale)
          deallocate(newton_fscale)
          deallocate(param)

       end if
       
    end if

  end subroutine solve_elasticity_im


  subroutine solve_elasticity_bm(fri,i,j,st0,Q,sz,vxp,vxm,vyp,vym, &
       sx,sy,sx0,sy0,fxp,fxm,fyp,fym,cvsp,cvsm,transverse_slip)
    ! SOLVE_ELASTICITY_BM solves elasticity equation with stress equal to strength
    ! (when slipping).  This version is for the bimaterial case.
    ! 
    ! Modified: 8 May 2007

    use constants, only : zero,one,two,three
    use friction, only : friction_type
    use utilities, only : newton

    implicit none

    ! I/O Parameters:
    ! FRI = friction variables
    ! I = index in x direction
    ! J = index in y direction
    ! ST0 = portion of shear strength independent of slip velocity
    ! SZ = stress in z direction
    ! Q = state variable
    ! VXP = particle velocity in x direction, plus side
    ! VXM = particle velocity in x direction, minus side
    ! VYP = particle velocity in y direction, plus side
    ! VYM = particle velocity in y direction, minus side
    ! SX = stress in x direction
    ! SY = stress in y direction
    ! SX0 = load in x direction
    ! SY0 = load in y direction
    ! FXP = stress transfer in x direction, plus side
    ! FXM = stress transfer in x direction, minus side
    ! FYP = stress transfer in y direction, plus side
    ! FYM = stress transfer in y direction, minus side
    ! CVSP = coefficient of shear radiation damping, plus side
    ! CVSM = coefficient of shear radiation damping, minus side
    ! TRANSVERSE_SLIP = scalar or vector slip

    type(friction_type),intent(inout) :: fri
    integer(pin),intent(in) :: i,j
    real(pr),intent(in) :: st0,Q,sz,sx0,sy0,fxp,fxm,fyp,fym,cvsp,cvsm
    real(pr),intent(inout) :: vxp,vxm,vyp,vym,sx,sy
    logical,intent(in) :: transverse_slip

    ! Internal Parameters:
    ! VXPL = particle velocity in x direction if fault is locked, plus side
    ! VXML = particle velocity in x direction if fault is locked, minus side
    ! VYPL = particle velocity in y direction if fault is locked, plus side
    ! VYML = particle velocity in y direction if fault is locked, minus side
    ! SXL = stress in x direction if fault is locked
    ! SYL = stress in y direction if fault is locked
    ! SL = magnitude of shear stress vector if fault is locked
    ! NEWTON_X = independent variables for Newton's method
    ! NEWTON_XSCALE = scale of independent variables for convergence of Newton's method
    ! NEWTON_FSCALE = scale of function for convergence of Newton's method
    ! PARAM = parameters required for function that is solved using Newton's method
    ! NEWTON_NMAX = maximum number of interations allowed to solve equations with Newton's method
    ! NEWTON_TOLX = maximum tolerated error in independent variables
    ! NEWTON_TOLF = maximum tolerated error in dependent variables

    real(pr) :: vxpl,vxml,vypl,vyml,sxl,syl,sl
    real(pr),dimension(:),allocatable :: newton_x,newton_xscale,newton_fscale,param
    integer(pin) :: newton_nmax
    real(pr) :: newton_tolx,newton_tolf
    
    ! set iteration limit and tolerances

    newton_nmax = 1000
    newton_tolx = epsilon(one)
    newton_tolf = epsilon(one)**(two/three)

    if (transverse_slip) then

       ! assume fault is locked, calculate velocities and stresses
       
       vxpl = (fxp-fxm)/(cvsp+cvsm)
       vxml = vxpl
       sxl = sx0+fxp-cvsp*vxpl
       vypl = (fyp-fym)/(cvsp+cvsm)
       vyml = vypl
       syl = sy0+fyp-cvsp*vypl
       sl = sqrt(sxl**2+syl**2)
       
       ! compare strength to elastic shear stress if fault were locked,
       ! allow failure if stress exceeds strength
       
       if (fri%friction/='ratestate'.and.sl<st0) then
          
          ! stress less than strength, fault locked
          
          vxp = vxpl
          vxm = vxml
          vyp = vypl
          vym = vyml
          sx = sxl
          sy = syl
          
       else
          
          ! stress set to strength, fault slips;
          ! solution via Newton's method--see the subroutine elastic for further description
          
          allocate(newton_x(4))

          ! addition of eps to Vx avoids V=0 problem in some rate-and-state laws

          newton_x(1) = vxp-vxm+epsilon(zero)
          newton_x(2) = sx
          newton_x(3) = vyp-vym
          newton_x(4) = sy
          
          allocate(newton_xscale(4))

          newton_xscale(1) = max(abs(vxp-vxm),abs(vyp-vym))
          newton_xscale(2) = max(abs(sx),abs(sy))
          newton_xscale(3) = max(abs(vxp-vxm),abs(vyp-vym))
          newton_xscale(4) = max(abs(sx),abs(sy))

          allocate(newton_fscale(4))

          newton_fscale(2) = max(abs(fxp),abs(fxm),abs(cvsp*vxp),abs(cvsm*vxm))
          newton_fscale(3) = max(abs(fyp),abs(fym),abs(cvsp*vyp),abs(cvsm*vym))
          newton_fscale(1) = newton_fscale(2)**2+newton_fscale(3)**2
          newton_fscale(4) = abs(newton_fscale(2)*(vyp-vym))+abs(newton_fscale(3)*(vxp-vxm))

          fri%i = i
          fri%j = j
          fri%st0 = st0
          fri%Q = Q
          fri%sz = sz
          fri%cV = cvsp*cvsm/(cvsp+cvsm)
          fri%sfx = (sx0+fxp)/(one+cvsp/cvsm)+(sx0+fxm)/(one+cvsm/cvsp)
          fri%sfy = (sy0+fyp)/(one+cvsp/cvsm)+(sy0+fym)/(one+cvsm/cvsp)

          allocate(param(size(transfer(fri,param))))
          param = transfer(fri,param)

          call newton(newton_x,newton_nmax,newton_tolx,newton_tolf,param,elastic,newton_xscale,newton_fscale)
          
          vxp =  (-newton_x(2)+sx0+fxp)/cvsp
          vxm = -(-newton_x(2)+sx0+fxm)/cvsm
          sx = newton_x(2)
          vyp =  (-newton_x(4)+sy0+fyp)/cvsp
          vym = -(-newton_x(4)+sy0+fym)/cvsm
          sy = newton_x(4)
          
          deallocate(newton_x)
          deallocate(newton_xscale)
          deallocate(newton_fscale)
          deallocate(param)

       end if

    else

       ! lock fault in y-direction
       
       vyp = (fyp-fym)/(cvsp+cvsm)
       vym = vyp
       sy = sy0+fyp-cvsp*vyp
       
       ! assume fault is locked, calculate velocities and stresses
       
       vxpl = (fxp-fxm)/(cvsp+cvsm)
       vxml = vxpl
       sxl = sx0+fxp-cvsp*vxpl
       
       sl = abs(sxl)

       ! compare strength to elastic shear stress if fault were locked,
       ! allow failure if stress exceeds strength
       
       if (fri%friction/='ratestate'.and.sl<st0) then
          
          ! stress less than strength, fault locked
          
          vxp = vxpl
          vxm = vxml
          sx = sxl
          
       else
          
          ! stress set to strength, fault slips;
          ! solution via Newton's method--see the subroutine elastic for further description
          
          allocate(newton_x(2))

          ! addition of eps to Vx avoids V=0 problem in some rate-and-state laws

          newton_x(1) = vxp-vxm+epsilon(zero)
          newton_x(2) = sx
          
          allocate(newton_xscale(2))

          newton_xscale(1) = abs(vxp-vxm)
          newton_xscale(2) = abs(sx)

          allocate(newton_fscale(2))

          newton_fscale(1) = max(abs(fxp),abs(fxm),abs(cvsp*vxp),abs(cvsm*vxm))
          newton_fscale(2) = newton_fscale(1)

          fri%i = i
          fri%j = j
          fri%st0 = st0
          fri%Q = Q
          fri%sz = sz
          fri%cV = cvsp*cvsm/(cvsp+cvsm)
          fri%sfx = (sx0+fxp)/(one+cvsp/cvsm)+(sx0+fxm)/(one+cvsm/cvsp)
          
          allocate(param(size(transfer(fri,param))))
          param = transfer(fri,param)

          call newton(newton_x,newton_nmax,newton_tolx,newton_tolf,param,elastic,newton_xscale,newton_fscale)
          
          vxp =  (-newton_x(2)+sx0+fxp)/cvsp
          vxm = -(-newton_x(2)+sx0+fxm)/cvsm
          sx = newton_x(2)
          
          deallocate(newton_x)
          deallocate(newton_xscale)
          deallocate(newton_fscale)
          deallocate(param)

       end if

    end if

  end subroutine solve_elasticity_bm


  subroutine elastic(x,f,param,err,jac)
    ! ELASTIC solves elasticity equations for shear stress and slip velocity given
    ! shear strength.  This version applies to both scalar and vector slip (in which 
    ! shear stress opposes the slip velocity in direction).  The equations are:
    ! (1) st(V) = abs(sx) OR st(V) = sqrt(sx**2+sy**2) 
    ! (2) sx = sx0+fx-cV*Vx
    ! (3) sy = sy0+fy-cV*Vy
    ! (4) Vx*sy = Vy*sx
    ! where (2) and (3) are the elastodynamic equations, (1) says that the magnitude of 
    ! the shear stress is equal to the strength, and (4) says that shear stress 
    ! opposes slip velocity.  There are either two (Vx,sx) or four unknowns (Vx,Vy,sx,sy), 
    ! which are solved for using Newton's method.  This is for the identical materials case.
    ! 
    ! Modified: 1 November 2006

    use constants, only : zero,one,two
    use friction, only : friction_type,strengthV

    implicit none
    
    ! I/O Parameters:
    ! X = independent variable for Newton's method
    ! F = dependent variable for Newton's method
    ! PARAM = variables required for evaluating elastodynamic equation and strength
    ! ERR = error flag (nonzero if error)
    ! JAC = Jacobian df/dx

    real(pr),dimension(:),intent(in) :: x
    real(pr),intent(out) :: f(size(x))
    real(pr),dimension(:),intent(in) :: param
    integer(pin),intent(out) :: err
    real(pr),intent(out),optional :: jac(size(x),size(x))

    ! Internal Parameters:
    ! VX = slip velocity in x direction
    ! VY = slip velocity in y direction
    ! V = magnitude of slip velocity
    ! SX = stress in x direction
    ! SY = stress in y direction
    ! ST = shear strength
    ! STV = velocity-dependent strength
    ! DSTDV = derivative of shear strength with respect to slip velocity
    ! FRI = friction variables

    real(pr) :: Vx,Vy,V,sx,sy,st,stV,dstdV
    type(friction_type) :: fri

    fri = transfer(param,fri)

    ! determine if scalar or vector slip, solve appropriate equations

    select case(size(x))

    case(2) ! scalar

       Vx = x(1)
       sx = x(2)

       call strengthV(fri,Vx,stV,dstdV,err)
       st = fri%st0+stV

       f(1) = -st+sx
       f(2) = -sx+fri%sfx-fri%cV*Vx

       if (present(jac)) then
          
          jac(1,1) = -dstdV
          jac(1,2) = one
          
          jac(2,1) = -fri%cV
          jac(2,2) = -one
          
       end if
    
    case(4) ! vector
       
       Vx = x(1)
       sx = x(2)
       Vy = x(3)
       sy = x(4)

       V = sqrt(Vx**2+Vy**2)

       call strengthV(fri,V,stV,dstdV,err)
       st = fri%st0+stV

       f(1) = -st**2+sx**2+sy**2
       f(2) = -sx+fri%sfx-fri%cV*Vx
       f(3) = -sy+fri%sfy-fri%cV*Vy
       f(4) = -Vy*sx+Vx*sy
       
       if (present(jac)) then
          
          if (V==zero) then
             jac(1,1) = zero
             jac(1,3) = zero
          else
             jac(1,1) = -two*st*dstdV*Vx/V
             jac(1,3) = -two*st*dstdV*Vy/V
          end if
          jac(1,2) = two*sx
          jac(1,4) = two*sy
          
          jac(2,1) = -fri%cV
          jac(2,2) = -one
          jac(2,3) = zero
          jac(2,4) = zero
          
          jac(3,1) = zero
          jac(3,2) = zero
          jac(3,3) = -fri%cV
          jac(3,4) = -one
          
          jac(4,1) = sy
          jac(4,2) = -Vy
          jac(4,3) = -sx
          jac(4,4) = Vx
          
       end if

    end select

  end subroutine elastic


end module friction_solver
