! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module front

  ! FRONT contains variables and routines for outputting rupture front history
  ! 
  ! Modified: 26 November 2006

  use constants, only : pr,pin

  implicit none

  ! FRONT_TYPE is a derived type containing rupture front variables
  !
  ! Parameters:
  ! FRONT = output rupture front history
  !      T = output rupture front history
  !      F = do not output rupture front history
  ! FIELD (character = field component to be contoured
  ! VAL = minimum velocity a point must exceed to denote 
  ! the passage of the rupture front
  ! XMIN = minimum value of x
  ! XMAX = maximum value of x
  ! YMIN = minimum value of y
  ! YMAX = maximum value of y
  ! OTS_XMIN = whether or not to include the grid point 
  ! just outside of bound set by xmin
  ! OTS_XMAX = same as above for xmax
  ! OTS_YMIN = same as above for ymin
  ! OTS_YMAX = same as above for ymax
  ! UNIT = unit number for output file
  ! MX = index corresponding to xmin
  ! PX = index corresponding to xmax
  ! MY = index corresponding to ymin
  ! PY = index corresponding to ymax
  ! NX = number of grid points in x direction
  ! NY = number of grid points in y direction
  ! DX = spacing in x direction
  ! DY = spacing in y direction
  ! STRIDE_X = stride in x direction
  ! STRIDE_Y = stride in y direction
  ! T = array containing the time at which the rupture
  ! front passes each point

  type front_type
     logical :: front,ots_xmin,ots_xmax,ots_ymin,ots_ymax
     character(4) :: field
     integer(pin) :: unit,mx,px,my,py,nx,ny,stride_x,stride_y
     real(pr) :: val,xmin,xmax,ymin,ymax,dx,dy
     real(pr),dimension(:,:),allocatable :: t
  end type front_type

contains
  

  subroutine read_front(ninput,frt)
    ! READ_FRONT reads in front variables from file
    ! 
    ! Modified: 24 November 2006

    use io, only : error
    implicit none

    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! FRT = rupture front variables

    integer(pin),intent(in) :: ninput
    type(front_type),intent(out) :: frt
    
    ! Internal Parameters:
    ! FRONT = output rupture front history
    ! FIELD = field component to be contoured
    ! XMIN = minimum value of x
    ! XMAX = maximum value of x
    ! YMIN = minimum value of y
    ! YMAX = maximum value of y
    ! OTS_XMIN = whether or not to include the grid point 
    ! just outside of bound set by xmin
    ! OTS_XMAX = same as above for xmax
    ! OTS_YMIN = same as above for ymin
    ! OTS_YMAX = same as above for ymax
    ! STRIDE_X = stride in x direction
    ! STRIDE_Y = stride in y direction
    ! VAL = minimum value a point must exceed to denote 
    ! the passage of the rupture front
    ! STAT = I/O error flag

    logical :: front,ots_xmin,ots_xmax,ots_ymin,ots_ymax
    character(4) :: field
    integer(pin) :: stride_x,stride_y
    real(pr) :: val,xmin,xmax,ymin,ymax
    integer(pin) :: stat

    ! make namelist of user input variables

    namelist /front_list/ front,field,val,xmin,xmax,ymin,ymax, &
         ots_xmin,ots_xmax,ots_ymin,ots_ymax,stride_x,stride_y

    ! defaults

    front = .false.
    field = 'Vx'
    val = 1.d-3
    xmin = -15._pr
    xmax = 15._pr
    ymin = -15._pr
    ymax = 0._pr
    ots_xmin = .false.
    ots_xmax = .false.
    ots_ymin = .false.
    ots_ymax = .false.
    stride_x = 1
    stride_y = 1

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=front_list,iostat=stat)
    if (stat>0) call error("Error reading namelist 'front_list' in .in file",'read_front')

    ! assign input variables to components of derived type

    frt%front = front
    frt%field = field
    frt%val = val
    frt%xmin = xmin
    frt%xmax = xmax
    frt%ymin = ymin
    frt%ymax = ymax
    frt%ots_xmin = ots_xmin
    frt%ots_xmax = ots_xmax
    frt%ots_ymin = ots_ymin
    frt%ots_ymax = ots_ymax
    frt%stride_x = stride_x
    frt%stride_y = stride_y

  end subroutine read_front

  
  subroutine init_front(necho,mdl,frt)
    ! INIT_FRONT initializes front variables
    ! 
    ! Modified: 12 November 2006

    use model, only : model_type,get_bound_range
    use io, only : write_matlab
    implicit none

    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MDL = model variables
    ! FRT = rupture front variables

    integer(pin),intent(in) :: necho
    type(model_type),intent(in) :: mdl
    type(front_type),intent(inout) :: frt
    
    ! return if not using this output method

    if (.not.frt%front) return

    ! get integer bounds and update min,max, shifting to nearest grid points
    ! x direction
    call get_bound_range(frt%field,'x',frt%xmin,frt%xmax,frt%ots_xmin,frt%ots_xmax,mdl,frt%mx,frt%px)
    frt%nx = 1+(frt%px-frt%mx)/frt%stride_x
    frt%dx = real(frt%stride_x,pr)*mdl%dx
    ! y direction
    call get_bound_range(frt%field,'y',frt%ymin,frt%ymax,frt%ots_ymin,frt%ots_ymax,mdl,frt%my,frt%py)
    frt%ny = 1+(frt%py-frt%my)/frt%stride_y
    frt%dy = real(frt%stride_y,pr)*mdl%dy

    ! allocate memory for rupture front time array,
    ! initialize array with t=0

    allocate(frt%t(frt%nx,frt%ny))
    frt%t = 0._pr

    ! output variable values into matlab file

    call write_matlab(necho,'field',frt%field,'frt')
    call write_matlab(necho,'val',frt%val,'frt')
    call write_matlab(necho,'xmin',frt%xmin,'frt')
    call write_matlab(necho,'xmax',frt%xmax,'frt')
    call write_matlab(necho,'ymin',frt%ymin,'frt')
    call write_matlab(necho,'ymax',frt%ymax,'frt')
    call write_matlab(necho,'nx',frt%nx,'frt')
    call write_matlab(necho,'ny',frt%ny,'frt')
    call write_matlab(necho,'dx',frt%dx,'frt')
    call write_matlab(necho,'dy',frt%dy,'frt')
    
  end subroutine init_front

 
  subroutine destroy_front(frt)
    ! DESTROY_FRONT destroys derived type frt
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! FRT = rupture front variables

    type(front_type),intent(inout) :: frt

    ! return if not using this output method

    if (.not.frt%front) return

    ! close output file

    close(frt%unit)

    ! deallocate memory assigned to allocatable arrays

    if (allocated(frt%t)) deallocate(frt%t)

  end subroutine destroy_front


  subroutine open_front(pb_name,frt)
    ! OPEN_FRONT opens output file and writes header information
    ! 
    ! Modified: 27 July 2005

    use io, only : new_io_unit,error
    implicit none

    ! I/O Parameters:
    ! PB_NAME = name of problem
    ! FRT = rupture front variables

    character(*),intent(in) :: pb_name
    type(front_type),intent(inout) :: frt

    ! Internal Parameters:
    ! NAME = file name
    ! STAT = I/O error flag

    character(64) :: name
    integer(pin) :: stat

    ! open file unit for output

    frt%unit = new_io_unit()
    name = 'data/' // trim(adjustl(pb_name)) // '_front.dat'
    open(frt%unit,file=name,form='formatted',iostat=stat)
    if (stat/=0) call error("Error opening file: '" // trim(adjustl(name)) // "'",'open_front_file')

    ! write header to file

    write(frt%unit, '("# field component = ",a)' ,err=100) trim(adjustl(frt%field))
    write(frt%unit, '("# minimum value of x = ",f14.6," km")' ,err=100) frt%xmin
    write(frt%unit, '("# maximum value of x = ",f14.6," km")' ,err=100) frt%xmax
    write(frt%unit, '("# number of grid points in x direction = ",i4)' ,err=100) frt%nx
    write(frt%unit, '("# element size in x = ",f14.6," km")' ,err=100) frt%dx
    write(frt%unit, '("# minimum value of y = ",f14.6," km")' ,err=100) frt%ymin
    write(frt%unit, '("# maximum value of y = ",f14.6," km")' ,err=100) frt%ymax
    write(frt%unit, '("# number of grid points in y direction = ",i4)' ,err=100) frt%ny
    write(frt%unit, '("# element size in y = ",f14.6," km")' ,err=100) frt%dy
    write(frt%unit, '("# critical value that defines rupture front = ",f14.6," m/s")' ,err=100) frt%val
    write(frt%unit, '("# Time of passage of rupture front in E14.6:")',err=100)

    return

100 call error("Error writing header to file: '" // trim(adjustl(name)) // "'",'open_front_file')

  end subroutine open_front


  subroutine output_front(pb_name,frt)
    ! OUTPUT_FRONT writes rupture front data to output file
    ! 
    ! Modified: 18 January 2006

    use io, only : error,flush_io

    implicit none

    ! I/O Parameters:
    ! PB_NAME = name of problem
    ! FRT = rupture front variables

    character(*),intent(in) :: pb_name
    type(front_type),intent(inout) :: frt

    ! Internal Parameters:
    ! J = index in y direction
    ! NX = number of grid points in x direction
    ! NY = number of grid points in y direction
    ! STAT = I/O error flag
    ! FMT = output format

    integer(pin) :: j,nx,ny,stat
    character(64) :: fmt

    ! return if not using this output method

    if (.not.frt%front) return

    ! open output file

    call open_front(pb_name,frt)

    ! assign pyropriate dimensions of array

    nx = size(frt%t,1)
    ny = size(frt%t,2)

    ! make format specifier to include nx copies of a real variable

    write(fmt,'(i12)') nx
    fmt = '(' // trim(adjustl(fmt)) // 'E16.7)'

    ! loop over y locations, at each value of y output the time value at 
    ! the various x locations in a single row, then move on to next row 
    ! for next y location

    do j=1,ny
       write(frt%unit,trim(fmt),iostat=stat) frt%t(:,j)
       if (stat/=0) call error('Error writing front data to file','output_front')            
    end do

    ! flush the file unit

    call flush_io(frt%unit)
    
    ! close output file

    close(frt%unit)

  end subroutine output_front


  subroutine set_front(frt,fld,t,dt)
    ! SET_FRONT marks progress of rupture front, assigns new
    ! value of time if front has passed a particular point
    ! 
    ! Modified: 28 July 2005

    use fields, only : fields_type,get_field,get_output_time
    implicit none

    ! I/O Parameters:
    ! FRT = rupture front variables
    ! FLD = fields variables
    ! T = time of current time step 
    ! DT = time step

    type(fields_type),intent(in) :: fld
    type(front_type),intent(inout) :: frt
    real(pr),intent(in) :: t,dt

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction
    ! TIM = time corresponding to particular field
    ! DATA = strided section of field array
    
    integer(pin) :: i,j
    real(pr) :: tim
    real(pr),dimension(:,:),allocatable :: data

    ! return if not using this output method

    if (.not.frt%front) return

    ! search everywhere that has not already been set
    ! if rupture front condition is met, then store current time
    ! in the pyropriate spot in the array

    allocate(data(frt%nx,frt%ny))

    call get_field(data,frt%field,fld,frt%nx,frt%ny,1, &
         frt%mx,frt%px,frt%my,frt%py,0,0,frt%stride_x,frt%stride_y,1)

    tim = get_output_time(frt%field,t,dt)

    do j=1,frt%ny
       do i=1,frt%nx
          if (frt%t(i,j)/=0._pr) cycle
          if (data(i,j)>=frt%val) frt%t(i,j) = tim
       end do
    end do

    deallocate(data)

  end subroutine set_front


end module front

