! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module init

  ! INIT contains routines to initialize a problem
  ! 
  ! Modified: 5 February 2007

  use constants, only : pin

  implicit none

contains


  subroutine initialize(pb)
    ! INITIALIZE initializes a problem
    ! 
    ! Modified: 5 February 2007
    
    use io, only : new_io_unit,error
    use problem, only : problem_type
    use model, only : read_model,init_model,distribute_model
    use convolution, only : read_convolution,init_convolution_master, &
         distribute_convolution,init_convolution_slave
    use kernel, only : read_kernel,init_kernel,distribute_kernel
    use fields, only : read_fields,init_fields_master,init_fields_slave
    use fft_routines, only : read_fft,init_fft
    use friction, only : read_friction,init_friction
    use mesh, only : read_mesh,init_mesh
    use front, only : read_front,init_front
    use mpi_routines, only : is_master

    implicit none
    
    ! I/O Parameters:
    ! PB = problem
    
    type(problem_type),intent(inout) :: pb
    
    ! Internal Parameters:
    ! NAME_INPUT = name of input file
    ! NAME_ECHO = name of file in which to echo all input variables
    ! (this is set up for use with matlab)
    ! NINPUT = unit for input file
    ! NECHO = unit for echo (matlab) file
    ! STAT = I/O error flag

    character(64) :: name_input,name_echo
    integer(pin) :: ninput,necho,stat
    
    if (is_master(pb%mp)) then

       ! open file unit for reading input file

       ninput = new_io_unit()
       name_input = 'problems/' // trim(adjustl(pb%name)) // '.in'
       open(ninput,file=name_input,iostat=stat)
       if (stat/=0) call error("Error opening file: '" // &
            trim(adjustl(name_input)) // "'",'initialize')
       
       ! open file unit for echoing input variables
       
       necho = new_io_unit()
       name_echo = 'data/' // trim(adjustl(pb%name)) // '.m'
       open(necho,file=name_echo,iostat=stat)
       if (stat/=0) call error("Error opening file: '" // &
            trim(adjustl(name_echo)) // "'",'initialize')
       
       ! read in configuration variables set in *.in input file,
       ! allocate and initialize main data types
       ! echo input variables to a single matlab file stored in data directory
       
       call read_model(ninput,pb%mdl)
       call read_kernel(ninput,pb%krn)
       call read_convolution(ninput,pb%cnv)
       call read_fft(ninput,pb%fft)
       call read_friction(ninput,pb%fld%flt,pb%fri)
       call read_fields(ninput,pb%mdl,pb%fld)
       call read_mesh(ninput,pb%msh)
       call read_front(ninput,pb%frt)
       
       ! close input file
       
       close(ninput,iostat=stat)
       if (stat/=0) call error("Error closing file: '" // &
            trim(adjustl(name_input)) // "'",'initialize')

       ! initialize master

       call init_model(necho,pb%mdl)
       call init_fields_master(necho,pb%mdl,pb%fld)
       call init_convolution_master(necho,pb%mdl,pb%cnv)
       call init_kernel(necho,pb%mdl,pb%krn,pb%cnv%formulation)
       call init_fft(pb%mdl,pb%fft)
       call init_friction(necho,pb%mdl,pb%fri,pb%fld%flt)

    end if

    ! distribute parameters from master to slave

    call distribute_model(pb%mdl,pb%mp)

    call distribute_convolution(pb%cnv)

    call distribute_kernel(pb%mdl,pb%krn,pb%cnv%formulation)

    ! initialize slaves (all processes)

    call init_convolution_slave(pb%mdl,pb%krn,pb%cnv,pb%mp)

    call init_fields_slave(pb%mdl,pb%fld,pb%mp)

    ! initialize output routines

    if (is_master(pb%mp)) then
       
       call init_mesh(necho,pb%name,pb%msh,pb%mdl,pb%fld)
       call init_front(necho,pb%mdl,pb%frt)
       
       ! close echo file
       
       close(necho,iostat=stat)
       if (stat/=0) call error("Error closing file: '" // &
            trim(adjustl(name_echo)) // "'",'initialize')

    end if

  end subroutine initialize
  

end module init
