! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module load

  ! LOAD contains routines to manipulate time-dependent loads
  ! 
  ! Modified: 20 April 2007

  use constants, only : pin,pr
  use asperity, only : asperity_list

  implicit none

  ! LOAD_TYPE is a derived type containing load variables
  !
  ! Parameters:
  ! ASPERITY_FILE = add heterogeneity to initial slip-weakening fields
  !      T = use asperity array data file
  !      F = do not use asperity array data file
  ! FILENAME = asperity file name
  ! ASCII = asperity file format
  !      T = ascii
  !      F = binary
  ! FORM = form of time-dependent load
  ! T = loading time parameter (time beyond which load is held fixed, period of loading, etc.)
  ! SX0 = load in x direction
  ! SY0 = load in y direction
  ! SZ0 = load in z direction
  ! AX = loading parameter (loading rate, perturbation amplitude, etc.) in x direction 
  ! AY = loading parameter in y direction
  ! AZ = loading parameter in z direction

  type load_type
     logical :: asperity_file,ascii
     character(64) :: filename,form
     type(asperity_list) :: list
     real(pr) :: T
     real(pr),dimension(:,:),allocatable :: sx0,sy0,sz0,Ax,Ay,Az
  end type load_type

contains


  subroutine read_load(ninput,ld)
    ! READ_LOAD reads in load variables from file
    ! 
    ! Modified: 20 April 2007

    use constants, only : one
    use io, only : error
    use asperity, only : read_asperity_list

    implicit none

    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! LD = load variables

    integer(pin),intent(in) :: ninput
    type(load_type),intent(out) :: ld

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! ASPERITY_LIST = add heterogeneity to initial ld fields with list
    ! ASPERITY_FILE = add heterogeneity to initial ld fields from file
    ! FILENAME = asperity file name
    ! ASCII = asperity file format
    ! NDATA = number of field perturbations
    ! T = loading time parameter
    ! FORM = form of time-dependent load

    integer(pin) :: stat,ndata
    logical :: asperity_list,asperity_file,ascii
    character(64) :: filename,form
    real(pr) :: T

    ! make namelist of user input variables

    namelist /load_list/ asperity_list,asperity_file,filename,ascii,T,form

    ! default values

    asperity_list = .false.
    asperity_file = .false.
    ascii = .false.
    T = huge(one)
    form = 'ramp'

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=load_list,iostat=stat)
    if (stat/=0) call error("Error reading namelist 'load_list' in .in file",'read_load')

    ! assign input variables to components of derived type
  
    ld%asperity_file = asperity_file
    ld%filename = 'problems/' // adjustl(filename)
    ld%ascii = ascii
    ld%T = T
    ld%form = form

    ! asperity list, set ndata = 3 (for Ax,Ay,Az)

    ndata = 3
    if (asperity_list) &
         ld%list = read_asperity_list(ninput,'load',ndata)

  end subroutine read_load


  subroutine init_load(mdl,flt,ld)
    ! INIT_LOAD initializes load variables
    ! 
    ! Modified: 20 April 2007

    use constants, only : zero
    use io, only : open_array_data,close_array_data,read_array_data
    use model, only : model_type
    use asperity, only : assign_list_data,destroy_asperity_list
    use fault, only : fault_type

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! LD = load variables

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(in) :: flt
    type(load_type),intent(inout) :: ld

    ! Internal Parameters:
    ! FILEUNIT = asperity file unit

    integer(pin) :: fileunit

    ! return if not including time-dependent load

    if (.not.flt%dynload) return

    ! allocate arrays

    allocate(ld%sx0(mdl%nx,mdl%ny))
    allocate(ld%sy0(mdl%nx,mdl%ny))
    allocate(ld%sz0(mdl%nx,mdl%ny))
    allocate(ld%Ax(mdl%nx,mdl%ny))
    allocate(ld%Ay(mdl%nx,mdl%ny))
    allocate(ld%Az(mdl%nx,mdl%ny))

    ld%sx0 = zero
    ld%sy0 = zero
    ld%sz0 = zero
    ld%Ax = zero
    ld%Ay = zero
    ld%Az = zero

    ! read asperity arrays from file

    if (ld%asperity_file) then
       call open_array_data(ld%filename,fileunit,ld%ascii)
       call read_array_data(fileunit,ld%Ax,ld%ascii,'Ax')
       call read_array_data(fileunit,ld%Ay,ld%ascii,'Ay')
       call read_array_data(fileunit,ld%Az,ld%ascii,'Az')
       call close_array_data(fileunit)
    end if
    
    ! use list data to perturb fields

    call assign_list_data(ld%Ax,ld%list,1,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    call assign_list_data(ld%Ay,ld%list,2,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
    call assign_list_data(ld%Az,ld%list,3,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)

    ! deallocate memory for asperity list

    call destroy_asperity_list(ld%list)

  end subroutine init_load


  subroutine destroy_load(ld)
    ! DESTROY_LOAD destroys derived type ld
    ! 
    ! Modified: 20 April 2007

    implicit none

    ! I/O Parameters:
    ! LD = load variables

    type(load_type),intent(inout) :: ld

    ! deallocate memory assigned to allocatable arrays

    if (allocated(ld%sx0)) deallocate(ld%sx0)
    if (allocated(ld%sy0)) deallocate(ld%sy0)
    if (allocated(ld%sz0)) deallocate(ld%sz0)
    if (allocated(ld%Ax)) deallocate(ld%Ax)
    if (allocated(ld%Ay)) deallocate(ld%Ay)
    if (allocated(ld%Az)) deallocate(ld%Az)

  end subroutine destroy_load


  subroutine set_load(mdl,ld,flt)
    ! SET_LOAD sets values of the load
    ! 
    ! Modified: 20 April 2007

    use constants, only : half,one,pi
    use model, only : model_type
    use fault, only : fault_type
    use io, only : error

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! LD = load variables
    ! FLT = fault variables

    type(model_type),intent(in) :: mdl
    type(load_type),intent(inout) :: ld
    type(fault_type),intent(inout) :: flt

    ! Internal Parameters:
    ! F = fraction of load to be applied

    real(pr) :: f

    ! return if not including time-dependent load

    if (.not.flt%dynload) return

    ! subtract old time-dependent load from load
    
    flt%sx0 = flt%sx0-ld%sx0
    flt%sy0 = flt%sy0-ld%sy0
    flt%sz0 = flt%sz0-ld%sz0

    ! calculate current time-dependent load

       if (mdl%t<ld%T) then

          select case(ld%form)
          case default
             call error('Invalid form of time-dependent load','set_load')
          case('step')
             f = one
          case('ramp')
             f = min(mdl%t,ld%T)
          case('smooth')
             f = half*(one-cos(pi*mdl%t/ld%T))
          end select

       else

          f = one

       end if

    ld%sx0 = ld%Ax*f
    ld%sy0 = ld%Ay*f
    ld%sz0 = ld%Az*f

    ! add time-dependent load to load

    flt%sx0 = flt%sx0+ld%sx0
    flt%sy0 = flt%sy0+ld%sy0
    flt%sz0 = flt%sz0+ld%sz0

  end subroutine set_load


end module load
