! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
program main

  ! MAIN is a program to solve several problems as defined in user-
  ! configured input files
  ! 
  ! Modified: 17 May 2007

  use constants, only : pin
  use problem, only : problem_type,init_problem,destroy_problem
  use problem_routines, only : solve_problem
  use io, only : new_io_unit,open_status_file,close_status_file,error,write_status
  use test_utilities, only : test_newton,test_interp,test_update,test_gauss
  use mpi_routines, only : start_mpi,finish_mpi,is_master,master
  use mpi

  implicit none

  ! Parameters:
  ! PB = problem variables
  ! DONE = flag indicating if all problems are solved
  ! NPB = file unit for problem list file
  ! NAME = problem name
  ! LISTFILE = name of file containing problem list
  ! STATUSFILE = name of status file
  ! TEMP = dummy string
  ! STAT = I/O error flag
  ! IERR = MPI error flag

  logical :: done
  type(problem_type) :: pb
  integer(pin) :: npb,stat,ierr
  character(64) :: name,temp,listfile,statusfile

  namelist /input_list/ listfile,statusfile

  ! start MPI

  call start_mpi(pb%mp)

  !---BEGIN INITIALIZATION---

  if (is_master(pb%mp)) then
     
     npb = new_io_unit()
     
     ! defaults

     listfile = 'list'
     statusfile = 'status'

     ! open input file
     
     open(npb,file='input.in',status='old',iostat=stat)
     
     ! if input file found, read in file names; otherwise, use defaults

     if (stat==0) then

        read(npb,nml=input_list) 
        close(npb)

     else

        write(6,'(a)') 'Problem opening input.in'
        write(6,'(a)') 'Attempting to use default problem list (list.in) and status file (status)'

     end if

     ! initialize status file
     
     call open_status_file(statusfile)
     call write_status(time=.true.,date=.true.)
     
     ! write MPI status
     
     if (pb%mp%np/=1) then
        write(temp,'(a,i3,a,i3,a)') 'Start MPI: number of processes = ',pb%mp%np
        call write_status(temp,time=.true.)
     end if
     
     ! open file containing problem list
     
     listfile = 'problems/' // trim(adjustl(listfile)) // '.in'
     open(npb,file=trim(adjustl(listfile)),status='old',iostat=stat)
     
     if (stat/=0) &
          call error("Error opening file '" // trim(adjustl(listfile)) // "'",'main')
     
     ! move to start problem list
     
     do
        read(npb,*,end=100,iostat=stat) temp
        if (stat/=0) call error("Error finding problem name list in file '" // &
             trim(adjustl(listfile)) // "'",'main')
        if (temp=='!---begin:problem') exit
     end do
     
     ! give update in status file
     
     call write_status("Begin problem list from file '" // &
          trim(adjustl(listfile)) // "'",time=.true.)
     
  end if
  
  !---END INITIALIZATION---

  ! solve the problems
  
  done = .false.

  do

     call MPI_Barrier(MPI_COMM_WORLD,ierr)

     if (is_master(pb%mp)) then

        ! read in problem name
        
        read(npb,*,end=150,iostat=stat) name
        if (stat/=0) call error("Error reading problem name in file '" // &
             trim(adjustl(listfile)) // "'",'main')
        if (name=='!---end:problem') then
           
           ! all problems solved

           done = .true.

        else        

           ! give update in status file
           
           call write_status('Start problem: ' // trim(adjustl(name)),time=.true.)

        end if

     end if

     ! distribute termination status

     call MPI_Bcast(done,1,MPI_LOGICAL,master,MPI_COMM_WORLD,ierr)

     ! exit if done

     if (done) exit

     ! distribute problem name and termination status

     call MPI_Bcast(name,64,MPI_CHARACTER,master,MPI_COMM_WORLD,ierr)

     ! solve problem (in parallel)

     select case(name)

     case default

        ! create derived type pb
        
        call init_problem(pb,name)
        
        ! solve the problem
        
        call solve_problem(pb)
        
        ! destroy the problem (deallocate memory in pointers, etc.)
        
        call destroy_problem(pb)

     case ('test_newton')

        ! test Newton's method routines

        call test_newton

     case ('test_interp')

        ! test interpolation routines

        call test_interp

     case ('test_update')

        ! test integration update routines

        call test_update

     case ('test_gauss')

        ! test solving linear system with Gaussian elimination

        call test_gauss

     case('stop')

        stop

     end select

     if (is_master(pb%mp)) then

        ! give update in status file
        
        call write_status('End problem:   ' // trim(adjustl(name)),time=.true.)

     end if

  end do
     
  if (is_master(pb%mp)) then
     
     close(npb)

     ! give update in status file
     
     call write_status('Finish problem list',time=.true.)
     call write_status(time=.true.,date=.true.)
     
     ! close status file
     
     call close_status_file()
     
  end if

  ! finish MPI
  
  !print *, 'goodbye from process ',pb%mp%myid
  
  call finish_mpi(pb%mp)
  
  stop

  ! I/O errors

100 call error("Cannot find start of list 'problem_list' in file '" // &
         trim(adjustl(listfile)) // "'",'main')
150 call error("Did not reach end of list 'problem_list' in file '" // &
         trim(adjustl(listfile)) // "'",'main')

end program main
