! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module mesh

  ! MESH contains variables and routines pertaining to outputting data on a mesh
  ! 
  ! Modified: 26 November 2006

  use constants, only : pr,pin

  implicit none

  interface output_data
     ! OUTPUT_DATA overloads data output routines for arrays of different ranks
     module procedure output_data0d,output_data1d,output_data2d,output_data3d
  end interface

  ! MESH_ITEM is a derived type containing mesh variables
  !
  ! Parameters:
  ! NAME = name of item
  ! FIELD = field component to output
  ! UNIT = file unit for output
  ! SX = stride in x direction 
  ! (e.g., 2 means output every other point)
  ! SY = stride in y direction
  ! SZ = stride in z direction
  ! ST = stride in time
  ! N = number of time steps within time output window
  ! MX = minimum index in x direction
  ! PX = maximum index in x direction
  ! MY = minimum index in y direction
  ! PY = maximum index in y direction
  ! MZ = minimum index in z direction
  ! PZ = maximum index in z direction
  ! NX = number of strided grid points to be output in x direction
  ! NY = number of strided grid points to be output in y direction
  ! NZ = number of strided grid points to be output in z direction
  ! IX = interpolate in x direction
  ! IY = interpolate in y direction
  ! IZ = interpolate in z direction
  ! XMIN = minimum value of x
  ! XMAX = maximum value of x
  ! YMIN = minimum value of y
  ! YMAX = maximum value of y
  ! ZMIN = minimum value of z
  ! ZMAX = maximum value of z
  ! TMIN = minimum value of t
  ! TMAX = maximum value of t
  ! DX = spatial step in x direction
  ! DY = spatial step in y direction
  ! DZ0 = spatial step in z direction
  ! FX = normalized postion of output in x direction
  ! (used in interpolation)
  ! FY = normalized position of output in y direction
  ! FZ = normalized position of output in z direction
  ! OTS_XMIN = whether or not to include the grid point 
  ! just outside of bound set by xmin
  ! OTS_XMAX = same as above for xmax
  ! OTS_YMIN = same as above for ymin
  ! OTS_YMAX = same as above for ymax
  ! OTS_ZMIN = same as above for zmin
  ! OTS_ZMAX = same as above for zmax

  type :: mesh_item
     character(64) :: name
     character(5) :: field
     logical :: ots_xmin,ots_xmax,ots_ymin,ots_ymax,ots_zmin,ots_zmax, &
          ix,iy,iz
     integer(pin) :: unit,n,sx,sy,sz,st, &
          mx,px,my,py,mz,pz,nx,ny,nz
     real(pr) :: xmin,xmax,dx,ymin,ymax,dy,zmin,zmax,dz0,tmin,tmax, &
          fx,fy,fz
  end type mesh_item

  ! MESH_NODE is a derived type for basic node in singly-linked list 
  ! (single mesh item variables and pointer to next node)
  !
  ! Parameters:
  ! MS = mesh variables
  ! NEXT = pointer to next node

  type :: mesh_node
     private
     type(mesh_item) :: ms
     type(mesh_node),pointer :: next=>null()
  end type mesh_node

  ! MESH_LIST is a derived type containing singly-linked list of mesh items
  !
  ! Parameters:
  ! HEAD = pointer to first node in list

  type :: mesh_list
     private
     type(mesh_node),pointer :: head=>null()
  end type mesh_list

  ! MESH_TYPE is a derived type containing all mesh variables
  !
  ! Parameters:
  ! LIST = singly-linked list of mesh items
  ! UNIT = file unit for time output

  type :: mesh_type
     type(mesh_list) :: list
     integer(pin) :: unit
  end type mesh_type

contains


  subroutine read_mesh(ninput,msh)
    ! READ_MESH reads in mesh variables from file
    ! 
    ! Modified: 28 August 2006

    use io, only : error

    implicit none

    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! MSH = mesh variables

    integer(pin),intent(in) :: ninput
    type(mesh_type),intent(out) :: msh

    ! Internal Parameters:
    ! TEMP = temporary character string
    ! MS = mesh item

    character(64) :: temp
    type(mesh_item) :: ms
       
    ! move to start of mesh list
       
    rewind(ninput)
    do
       read(ninput,*,end=100) temp
       if (temp=='!---begin:mesh_list---') exit
    end do
       
    ! read mesh data and add nodes to list

    do
       ! read item data
       read(ninput,*,end=100) ms%name
       if (ms%name=='!---end:mesh_list---') goto 200
       read(ninput,*) ms%field
       read(ninput,*) ms%xmin,ms%xmax,ms%sx,ms%ots_xmin,ms%ots_xmax
       read(ninput,*) ms%ymin,ms%ymax,ms%sy,ms%ots_ymin,ms%ots_ymax
       read(ninput,*) ms%zmin,ms%zmax,ms%sz,ms%ots_zmin,ms%ots_zmax
       read(ninput,*) ms%tmin,ms%tmax,ms%st
       ! store as node in list
       call create_mesh_node(msh%list,ms)
    end do
    
100 call error("Error reading list 'mesh' in .in file",'read_mesh')
    
    ! come here when end of file is reached
200 continue
    
  end subroutine read_mesh


  subroutine init_mesh(necho,pb_name,msh,mdl,fld)
    ! INIT_MESH initializes mesh variables and creates mesh items
    ! 
    ! Modified: 28 August 2006

    use model, only : model_type
    use fields, only : fields_type

    implicit none
    
    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! PB_NAME = name of problem
    ! MSH = mesh variables
    ! MDL = model variables
    ! FLD = fields variables

    integer(pin),intent(in) :: necho
    character(*),intent(in) :: pb_name
    type(mesh_type),intent(inout) :: msh
    type(model_type),intent(in) :: mdl
    type(fields_type),intent(in) :: fld

    ! Internal Parameters:
    ! I_MS = index of mesh item
    ! N_MS = total number of mesh items
    ! WRITE_ML = whether or not to write matlab data
    ! MS = mesh item

    integer(pin) :: i_ms,n_ms
    logical :: write_ml
    type(mesh_item) :: ms

    ! do output matlab data
    
    write_ml = .true.

    ! get number of mesh nodes

    n_ms = get_num_mesh_nodes(msh%list)

    ! loop over each node

    do i_ms = 1,n_ms

       ! get mesh item

       call get_mesh_node(msh%list,i_ms,ms)

       ! initialize item, write matlab info

       call init_mesh_item(necho,ms,mdl,write_ml,fld)

       ! open file unit for output

       call open_mesh_file(ms%unit,ms%name,pb_name)

       ! overwrite old mesh item with new one

       call overwrite_mesh_node(msh%list,i_ms,ms)

    end do

    ! open file unit for time output
    
    call open_mesh_file(msh%unit,'time',pb_name)

 end subroutine init_mesh
   

  subroutine init_mesh_item(necho,ms,mdl,write_ml,fld)
    ! INIT_MESH_ITEM creates and initializes a mesh item
    ! 
    ! Modified: 12 November 2006

    use model, only : model_type,get_bound_point,get_bound_range
    use io, only : write_matlab
    use fields, only : fields_type,check_field

    implicit none
    
    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MS = mesh item
    ! MDL = model variables
    ! WRITE_ML = whether or not to write matlab data
    ! FLD = fields variables

    integer(pin),intent(in) :: necho
    type(mesh_item),intent(inout) :: ms
    type(model_type),intent(in) :: mdl
    logical,intent(in) :: write_ml
    type(fields_type),intent(in) :: fld

    ! check if memory is allocated to field

    call check_field(ms%field,fld)

    ! by default, no interpolation

    ms%ix = .false.
    ms%iy = .false.
    ms%iz = .false.

    ! get integer bounds and update min,max when not interpolating

    if (ms%xmin==ms%xmax) then
       call get_bound_point(ms%field,'x',ms%xmin,ms%ots_xmin,ms%ots_xmax,mdl,ms%mx,ms%px,ms%fx)
       ms%xmax = ms%xmin
       ms%nx = 1
       ms%dx = 0._pr
       if (ms%mx/=ms%px) ms%ix = .true.
    else
       call get_bound_range(ms%field,'x',ms%xmin,ms%xmax,ms%ots_xmin,ms%ots_xmax,mdl,ms%mx,ms%px)
       ms%nx = 1+(ms%px-ms%mx)/ms%sx
       ms%dx = real(ms%sx,pr)*mdl%dx
       ms%xmax = ms%xmin+ms%dx*real(ms%nx-1,pr)
       ms%fx = 0._pr
    end if

    if (ms%ymin==ms%ymax) then
       call get_bound_point(ms%field,'y',ms%ymin,ms%ots_ymin,ms%ots_ymax,mdl,ms%my,ms%py,ms%fy)
       ms%ymax = ms%ymin
       ms%ny = 1
       ms%dy = 0._pr
       if (ms%my/=ms%py) ms%iy = .true.
   else
       call get_bound_range(ms%field,'y',ms%ymin,ms%ymax,ms%ots_ymin,ms%ots_ymax,mdl,ms%my,ms%py)
       ms%ny = 1+(ms%py-ms%my)/ms%sy
       ms%dy = real(ms%sy,pr)*mdl%dy
       ms%ymax = ms%ymin+ms%dy*real(ms%ny-1,pr)
       ms%fy = 0._pr
    end if

    if (ms%zmin==ms%zmax) then
       call get_bound_point(ms%field,'z',ms%zmin,ms%ots_zmin,ms%ots_zmax,mdl,ms%mz,ms%pz,ms%fz)
       ms%zmax = ms%zmin
       ms%nz = 1
       ms%dz0 = 0._pr
       if (ms%mz/=ms%pz) ms%iz = .true.
    else
       call get_bound_range(ms%field,'z',ms%zmin,ms%zmax,ms%ots_zmin,ms%ots_zmax,mdl,ms%mz,ms%pz)
       ms%nz = 1+(ms%pz-ms%mz)/ms%sz
       ms%dz0 = real(ms%sz,pr)*mdl%dz0
       ms%zmax = ms%zmin+ms%dz0*real(ms%nz-1,pr)
       ms%fz = 0._pr
    end if

    ! zero number of elapsed time steps within output window

    ms%n = 0

    if (write_ml) then
       call write_matlab(necho,'name',ms%name,ms%name)
       call write_matlab(necho,'field',ms%field,ms%name)
       call write_matlab(necho,'nx',ms%nx,ms%name)
       call write_matlab(necho,'ny',ms%ny,ms%name)
       call write_matlab(necho,'nz',ms%nz,ms%name)
       call write_matlab(necho,'xmin',ms%xmin,ms%name)
       call write_matlab(necho,'xmax',ms%xmax,ms%name)
       call write_matlab(necho,'dx',ms%dx,ms%name)
       call write_matlab(necho,'ymin',ms%ymin,ms%name)
       call write_matlab(necho,'ymax',ms%ymax,ms%name)
       call write_matlab(necho,'dy',ms%dy,ms%name)
       call write_matlab(necho,'zmin',ms%zmin,ms%name)
       call write_matlab(necho,'zmax',ms%zmax,ms%name)
       call write_matlab(necho,'dz0',ms%dz0,ms%name)
       call write_matlab(necho,'tmin',ms%tmin,ms%name)
       call write_matlab(necho,'tmax',ms%tmax,ms%name)
       call write_matlab(necho,'stride_x',ms%sx,ms%name)
       call write_matlab(necho,'stride_y',ms%sy,ms%name)
       call write_matlab(necho,'stride_z',ms%sz,ms%name)
       call write_matlab(necho,'stride_t',ms%st,ms%name)
    end if

  end subroutine init_mesh_item


  subroutine open_mesh_file(unit,file_name,pb_name)
    use io, only : new_io_unit,error
    implicit none

    integer(pin),intent(out) :: unit
    character(*),intent(in) :: file_name,pb_name
    
    character(64) :: name
    integer(pin) :: stat

    ! open file

    unit = new_io_unit()
    name = 'data/' // trim(adjustl(pb_name)) // '_' // trim(adjustl(file_name)) // '.dat'
    open(unit,file=name,form='unformatted',status='replace',iostat=stat)
    if (stat/=0) call error("Error opening file: '" // trim(adjustl(name)) // "'",'open_mesh_file')

  end subroutine open_mesh_file


  subroutine output_mesh(fld,msh,t)
    ! OUTPUT_MESH obtains array of field values, appropriately strided, and calls
    ! a routine to write them to a data file
    !
    ! Modified: 24 October 2006

    use fields, only : fields_type,get_field

    implicit none

    ! I/O Parameters:
    ! FLD = fields variables
    ! MSH = mesh variables
    ! T = current time

    type(fields_type),intent(in) :: fld
    type(mesh_type),intent(inout) :: msh
    real(pr),intent(in) :: t

    ! Internal Parameters:
    ! I_MS = index of mesh item
    ! N_MS = number of mesh items
    ! NX = number of grid points in x direction (either strided or pre-interpolation)
    ! NY = number of grid points in y direction (either strided or pre-interpolation)
    ! NZ = number of grid points in z direction (either strided or pre-interpolation)
    ! DATA = field data

    integer(pin) :: i_ms,n_ms,nx,ny,nz
    type(mesh_item) :: ms
    real(pr),dimension(:,:,:),allocatable :: data

    ! output time

    call output_data0d(msh%unit,t)

    ! output mesh items

    n_ms = get_num_mesh_nodes(msh%list)

    do i_ms = 1,n_ms

       ! get mesh data

       call get_mesh_node(msh%list,i_ms,ms)

       ! cycle if time step is not appropriate

       if (t<ms%tmin.or.ms%tmax<t) cycle

       ! if tmin<=t<=tmax, then increment potential time output index

       ms%n = ms%n+1
       call overwrite_mesh_node(msh%list,i_ms,ms)

       ! check stride of potential time output index

       if (mod(ms%n-1,ms%st)/=0) cycle

       ! set bounds

       nx = ms%nx
       ny = ms%ny
       nz = ms%nz

       if (ms%ix) nx = 1+ms%px-ms%mx ! should be 2
       if (ms%iy) ny = 1+ms%py-ms%my ! should be 2
       if (ms%iz) nz = 1+ms%pz-ms%mz ! should be 2

       ! allocate memory for field data

       allocate(data(nx,ny,nz))

       ! get field values within bounds

       call get_field(data,ms%field,fld,nx,ny,nz, &
            ms%mx,ms%px,ms%my,ms%py,ms%mz,ms%pz,ms%sx,ms%sy,ms%sz)

       ! output data

       call output_mesh_data(ms%unit,data,ms%ix,ms%iy,ms%iz,ms%fx,ms%fy,ms%fz)

       ! deallocate field data

       deallocate(data)

    end do

  end subroutine output_mesh


  subroutine output_mesh_data(unit,data,ix,iy,iz,fx,fy,fz)
    ! OUTPUT_MESH_DATA interpolates data (if needed) and calls output routines
    ! 
    ! Modified: 26 May 2006

    use utilities, only : interp_linear

    implicit none

    integer(pin),intent(in) :: unit
    logical,intent(in) :: ix,iy,iz
    real(pr),dimension(:,:,:),intent(in) :: data
    real(pr),intent(in) :: fx,fy,fz

    integer :: i,j,k,n1,n2,n3
    real(pr) :: data_int3d
    real(pr),dimension(:),allocatable :: data_int2d
    real(pr),dimension(:,:),allocatable :: data_int1d
    real(pr),dimension(2) :: c

    c = real((/0,1/),pr)

    n1 = size(data,1)
    n2 = size(data,2)
    n3 = size(data,3)

    if (ix) then ! int - x

       if (iy) then ! int - y

          if (iz) then ! int - z
             ! interpolate in x,y,z
             data_int3d = interp_linear(c,c,c,data,fx,fy,fz)
             call output_data(unit,data_int3d)
          else ! no - z
             ! interpolate in x,y
             allocate(data_int2d(n3))
             do k = 1,n3
                data_int2d(k) = interp_linear(c,c,data(1:2,1:2,k),fx,fy)
             end do
             call output_data(unit,data_int2d)
             deallocate(data_int2d)
          end if

       else ! no - y

          if (iz) then ! int - z
             ! interpolate in x,z
             allocate(data_int2d(n2))
             do j = 1,n2
                data_int2d(j) = interp_linear(c,c,data(1:2,j,1:2),fx,fz)
             end do
             call output_data(unit,data_int2d)
             deallocate(data_int2d)
          else ! no - z
             ! interpolate in x
             allocate(data_int1d(n2,n3))
             do k = 1,n3
                do j = 1,n2
                   data_int1d(j,k) = interp_linear(c,data(1:2,j,k),fx)
                end do
             end do
             call output_data(unit,data_int1d)
             deallocate(data_int1d)
          end if

       end if

    else ! no - x

       if (iy) then ! int - y

          if (iz) then ! int - z
             ! interpolate in y,z
             allocate(data_int2d(n1))
             do i = 1,n1
                data_int2d(i) = interp_linear(c,c,data(i,1:2,1:2),fy,fz)
             end do
             call output_data(unit,data_int2d)
             deallocate(data_int2d)
          else ! no - z
             ! interpolate in y
             allocate(data_int1d(n1,n3))
             do k = 1,n3
                do i = 1,n1
                   data_int1d(i,k) = interp_linear(c,data(i,1:2,k),fy)
                end do
             end do
             call output_data(unit,data_int1d)
             deallocate(data_int1d)
          end if

       else ! no - y

          if (iz) then ! int - z
             ! interpolate in z
             allocate(data_int1d(n1,n2))
             do j = 1,n2
                do i = 1,n1
                   data_int1d(i,j) = interp_linear(c,data(i,j,1:2),fz)
                end do
             end do
             call output_data(unit,data_int1d)
             deallocate(data_int1d)
          else ! no - z
             ! no interpolation
             call output_data(unit,data)
          end if

       end if

    end if


  end subroutine output_mesh_data


  subroutine output_data0d(unit,data)

    use constants, only : psav
    use io, only : error,flush_io

    implicit none

    integer(pin),intent(in) :: unit
    real(pr),intent(in) :: data

    integer(pin) :: stat
    character(64) :: name

    write(unit,iostat=stat) real(data,psav)
    
    if (stat/=0) then
       inquire(unit,name=name)
       call error("Error writing field to file '" // trim(adjustl(name))  // "'",'output_data0d')    
    end if

    call flush_io(unit)

  end subroutine output_data0d


  subroutine output_data1d(unit,data)

    use constants, only : psav
    use io, only : error,flush_io

    implicit none

    integer(pin),intent(in) :: unit
    real(pr),dimension(:),intent(in) :: data

    integer(pin) :: i,n,stat
    character(64) :: name

    n = size(data)

    write(unit,iostat=stat) &
         (real(data(i),psav),i=1,n)
    
    if (stat/=0) then
       inquire(unit,name=name)
       call error("Error writing field to file '" // trim(adjustl(name))  // "'",'output_data1d')    
    end if

    call flush_io(unit)

  end subroutine output_data1d


  subroutine output_data2d(unit,data)

    use constants, only : psav
    use io, only : error,flush_io

    implicit none

    integer(pin),intent(in) :: unit
    real(pr),dimension(:,:),intent(in) :: data

    integer(pin) :: i,j,n1,n2,stat
    character(64) :: name

    n1 = size(data,1)
    n2 = size(data,2)

    write(unit,iostat=stat) &
         ((real(data(i,j),psav),i=1,n1),j=1,n2)
    
    if (stat/=0) then
       inquire(unit,name=name)
       call error("Error writing field to file '" // trim(adjustl(name))  // "'",'output_data2d')    
    end if

    call flush_io(unit)

  end subroutine output_data2d


  subroutine output_data3d(unit,data)

    use constants, only : psav
    use io, only : error,flush_io

    implicit none

    integer(pin),intent(in) :: unit
    real(pr),dimension(:,:,:),intent(in) :: data

    integer(pin) :: i,j,k,n1,n2,n3,stat
    character(64) :: name

    n1 = size(data,1)
    n2 = size(data,2)
    n3 = size(data,3)

    write(unit,iostat=stat) &
         (((real(data(i,j,k),psav),i=1,n1),j=1,n2),k=1,n3)
    
    if (stat/=0) then
       inquire(unit,name=name)
       call error("Error writing field to file '" // trim(adjustl(name))  // "'",'output_data3d')    
    end if

    call flush_io(unit)

  end subroutine output_data3d


  subroutine destroy_mesh(msh)

    implicit none

    type(mesh_type),intent(inout) :: msh

    call destroy_mesh_list(msh%list)

    close(msh%unit)

  end subroutine destroy_mesh


  subroutine create_mesh_node(list,ms)

    implicit none

    type(mesh_list),intent(inout) :: list
    type(mesh_item),intent(in) :: ms

    ! Internal Parameters:
    ! CURRENT = pointer to current node in list

    type(mesh_node),pointer :: current=>null()

    ! move to end of list, create new node, point current to new 
    ! node, nullify current%next

    if (.not.associated(list%head)) then
       ! allocate memory for first node in list
       allocate(list%head)
       ! nullify pointer to next node
       nullify(list%head%next)
       ! point current to head
       current => list%head
    else
       ! start at head of list
       current => list%head
       ! move down list until empty node is reached
       do while(associated(current%next))
          ! end of list not found, move down one node
          current => current%next
       end do
       ! end of list found, add new node
       ! allocate memory for new node
       allocate(current%next)
       ! point current to new node
       current => current%next
       ! nullify pointer to next node
       nullify(current%next)
    end if

    ! store data in new node (pointed to by current)

    current%ms = ms

  end subroutine create_mesh_node
  

  subroutine get_mesh_node(list,i_ms,ms)
    ! GET_MESH_NODE gets mesh data from node
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! LIST = mesh list
    ! I_MS = index of node from which data is extracted
    ! MS = mesh variables

    type(mesh_list),intent(in) :: list
    integer(pin),intent(in) :: i_ms
    type(mesh_item),intent(out) :: ms

    ! Internal Parameters:
    ! CURRENT = pointer to current node in list
    ! I = index of node

    type(mesh_node),pointer :: current=>null()
    integer(pin) :: i

    ! start at beginning of list, move down it until i=i_ms

    current => list%head
    i = 1
    do while(i/=i_ms)
       i = i+1
       current => current%next
    end do

    ! move data from this node to output variables

    ms = current%ms

  end subroutine get_mesh_node


  subroutine overwrite_mesh_node(list,i_ms,ms)
    ! OVERWRITE_MESH_NODE overwrites mesh data at a particular node
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! LIST = mesh list
    ! I_MS = index of node which will be overwritten
    ! MS = mesh variables

    type(mesh_list),intent(inout) :: list
    integer(pin),intent(in) :: i_ms
    type(mesh_item),intent(in) :: ms

    ! Internal Parameters:
    ! CURRENT = pointer to current node in list
    ! I = index of node

    type(mesh_node),pointer :: current=>null()
    integer(pin) :: i

    ! start at beginning of list, move down it until i=i_ms

    current => list%head
    i = 1
    do while(i/=i_ms)
       i = i+1
       current => current%next
    end do

    ! overwrite data at this node

    current%ms = ms

  end subroutine overwrite_mesh_node


  function present_mesh_list(list)
    ! PRESENT_MESH_LIST checks if list exists
    ! 
    ! Modified: 19 July 2005

    implicit none

    ! I/O Parameters:
    ! LIST = mesh list
    ! PRESENT_MESH_LIST = whether or not list exists
    !      T = list exists
    !      F = list does not exist

    type(mesh_list),intent(in) :: list
    logical :: present_mesh_list

    present_mesh_list = associated(list%head)

  end function present_mesh_list


  function get_num_mesh_nodes(list) result(n)
    ! GET_NUM_MESH_NODES returns number of nodes in list
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! LIST = mesh list
    ! N = number of nodes in list

    type(mesh_list),intent(in) :: list
    integer(pin) :: n

    ! Internal Parameters:
    ! CURRENT = pointer to current node in list

    type(mesh_node),pointer :: current=>null()

    ! start at beginning of list, point current to head
    current => list%head

    ! move down list one node at a time, counting number of nodes
    if (associated(current)) then
       n = 1
       do while(associated(current%next))
          current => current%next
          n = n+1
       end do
    else
       n = 0
    end if

  end function get_num_mesh_nodes


  subroutine destroy_mesh_list(list)
    !  DESTROY_MESH_LIST destroys all nodes in a list
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! LIST = mesh list

    type(mesh_list),intent(inout) :: list

    ! Internal Parameters:
    ! CURRENT = pointer to current node in list
    ! PREVIOUS = pointer to previous node in list

    type(mesh_node),pointer :: current=>null(),previous=>null()

    ! start at beginning, point previous to first node in list, point current to next one down

    previous => list%head
    current => list%head%next

    ! move down the list, removing second node and replacing it with the third node

    do while(associated(previous%next))
       ! point previous%next to third node
       previous%next => current%next
       ! nullify pointer from second to third node
       nullify(current%next)
       ! destroy objects in second node
       call destroy_mesh_node(current%ms)
       ! deallocate second node
       deallocate(current)
       ! point current to what was the third and is now the second node
       current => previous%next
    end do

    ! now remove the head node

    deallocate(list%head)

  end subroutine destroy_mesh_list


  subroutine destroy_mesh_node(ms)
    !  DESTROY_MESH_NODE destroys one node
    ! 
    ! Modified: 12 November 2006

    implicit none

    type(mesh_item),intent(inout) :: ms

    close(ms%unit)

  end subroutine destroy_mesh_node


end module mesh

