! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module problem_routines

  ! PROBLEM_ROUTINES contains routines to solve a problem
  ! 
  ! Modified: 9 February 2007

  use constants, only : pr,pin

  implicit none

contains


  subroutine solve_problem(pb)
    ! SOLVE_PROBLEM solves a problem
    ! 
    ! Modified: 26 January 2007

    use constants, only : zero
    use problem, only : problem_type
    use init, only : initialize
    use mesh, only : output_mesh
    use front, only : output_front,set_front
    use io, only : write_status,error
    use rates, only : init_rates
    use static, only : iterate_static
    use mpi_routines, only : is_master,master,MPI_REAL_PR
    use mpi

    implicit none

    ! I/O Parameters:
    ! PB = problem variables

    type(problem_type),intent(inout) :: pb

    ! Internal Parameters:
    ! DT = time step
    ! DTS = substep length
    ! STR = character string used for writing message to status file
    ! ACCEPT_STEP = accept integration step
    ! DONE = flag indicating when all time steps are complete
    ! IERR = MPI error flag

    real(pr) :: dt,dts
    character(128) :: str
    logical :: accept_step,done
    integer(pin) :: ierr

    ! initialize everything at t=0
    
    call initialize(pb)

    ! solve static or dynamic problem

    select case(pb%mdl%method)
     
    case default
          
       ! set initial rates (at t=0)
       
       call init_rates(pb,accept_step)
          
       if (is_master(pb%mp)) then
          
          if (.not.accept_step) call error('Initial conditions rejected','solve_problem')
          
          ! print initialized statement
          
          call write_status('initialized',time=.true.)
          
          ! output fields on mesh
          
          call output_mesh(pb%fld,pb%msh,pb%mdl%t)
          
          ! update position of rupture front
          
          call set_front(pb%frt,pb%fld,pb%mdl%t,dt)
          
          ! output rupture front position
          
          call output_front(pb%name,pb%frt)
          
          ! set initial time step (to elastodynamic time step)
       
          dt = pb%mdl%dt
          dts = pb%mdl%dt
                    
       end if

       call MPI_Bcast(dt ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
       call MPI_Bcast(dts,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)

       ! loop over time steps

       done = .false.

       do
          
          ! advance one time step (all processes participate)
          
          call advance_time_step(pb,dt,dts)
          
          if (is_master(pb%mp)) then

             ! output fields on mesh
             
             call output_mesh(pb%fld,pb%msh,pb%mdl%t)
             
             ! update position of rupture front
             
             call set_front(pb%frt,pb%fld,pb%mdl%t,dt)
             
             ! output rupture front position
             
             call output_front(pb%name,pb%frt)
             
             ! print current time step and total time steps to status file
             
             if (pb%mdl%method=='adaptive') then
                write(str,'("simulation time: ",e20.10," out of ",e20.10," at ",i12," time steps")') pb%mdl%t,pb%mdl%tmax,pb%mdl%n
             else
                write(str,'("time steps: ",i12," out of ",i12)') pb%mdl%n,pb%mdl%nt
             end if
             call write_status(str,overwrite=.true.,time=.true.)
             
             ! done if time reaches/exceeds maximum time
             
             if (pb%mdl%t>=pb%mdl%tmax.and.pb%mdl%tmax/=zero) done = .true.
             
             ! done if maximum number of time steps is exceeded
             
             if (pb%mdl%n>=pb%mdl%nt) done = .true.
             
          end if

          call MPI_Bcast(done,1,MPI_LOGICAL,master,MPI_COMM_WORLD,ierr)

          if (done) exit

       end do

    case('static')
       
       if (is_master(pb%mp)) then

          ! print initialized statement
          
          call write_status('initialized',time=.true.)
          
          ! solve static problem
          
          call iterate_static(pb)
          
          ! output fields on mesh
          
          call output_mesh(pb%fld,pb%msh,pb%mdl%t)
          
          ! update position of rupture front
          
          call set_front(pb%frt,pb%fld,pb%mdl%t,dt)
          
          ! output rupture front position
          
          call output_front(pb%name,pb%frt)
     
       end if

    end select

  end subroutine solve_problem


  subroutine advance_time_step(pb,dt,dts)
    ! ADVANCE_TIME_STEP advances problem through one time step
    ! 
    ! Modified: 9 February 2007

    use constants, only : zero
    use problem, only : problem_type
    use integration, only : finalize_fields,move_error
    use time_step, only : step_explicit,step_trapezoid,step_rk
    use io, only : error
    use mpi_routines, only : is_master

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DT = time step
    ! DTS = substep length

    type(problem_type),intent(inout) :: pb
    real(pr),intent(inout) :: dt,dts

    ! Internal Parameters:
    ! ITER = iteration of elastodynamic time step
    ! T0 = initial time
    ! ATTEMPTS = number of attempts to take acceptable time step
    ! ACCEPT_STEP = accept integration step
    ! S0 = stage at beginning of time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values

    integer(pin) :: iter,attempts
    real(pr) :: t0
    logical :: accept_step
    integer(pin),parameter :: s0 = 1
    character(64),parameter :: stress_transfer = 'convolution'

    ! update time step index
    
    pb%mdl%n = pb%mdl%n+1
    
    ! store initial time
       
    t0 = pb%mdl%t
    
    ! iterate over step sizes until acceptable

    attempts = 0

    do

       ! update step size attempt counter

       attempts = attempts+1

       ! start time from beginning of time step
       
       pb%mdl%t = t0

       ! prevent overshooting tmax by reducing time step to stop at exactly at tmax if dt too large
       
       if (pb%mdl%tmax/=zero.and.pb%mdl%t+dt>pb%mdl%tmax) dt = pb%mdl%tmax-pb%mdl%t
       
       ! take time step

       select case(pb%mdl%method)

       case ('iterative')

          ! take explicit step

          call step_explicit(pb,dt,s0,stress_transfer,accept_step,update_history=.true.)

          ! repeat step using trapezoidal rule some number of times

          do iter = 1,pb%mdl%niter
             
             pb%mdl%iter = iter

             call step_trapezoid(pb,dt,s0,stress_transfer,accept_step,update_history=.false.)

             if (.not.accept_step) exit

          end do
             
          ! exit loop searching for time step if step is acceptable
          
          if (accept_step) exit
          
          ! time step failed

          if (is_master(pb%mp)) call error('time step failed','time_step')

       case ('substep')

          ! take explicit step to predict stress transfer f(t+dt)
          
          call step_explicit(pb,dt,s0,stress_transfer,accept_step,update_history=.true.)

          ! integrate over substeps, iterating as desired to improve stress transfer estimate
          ! (note that first time, iter=0, must always be done; iter>=1 are corrections)

          do iter = 0,pb%mdl%niter
             
             pb%mdl%iter = iter
             
             ! revert current time back to t0
             
             pb%mdl%t = t0
             
             ! integrate equations over substeps (length dts) with interpolation
             ! of stress transfer between t and t+dt
             
            call step_substep(pb,dt,dts,accept_step)
             
             if (.not.accept_step) exit
             
          end do

          ! exit loop searching for time step if step is acceptable
             
          if (accept_step) exit

          if (is_master(pb%mp)) call error('time step failed','time_step')
          
       case('adaptive')

          ! advance one time step, estimating error to predict new time step
       
          call step_rk(pb,dt,s0,stress_transfer,accept_step,attempts,t0)

          ! exit loop searching for time step if step is acceptable
          
          if (accept_step) exit

       end select

    end do

    ! accept step

    ! move error estimate
    
    if (is_master(pb%mp).and.pb%mdl%method=='adaptive') call move_error(pb%fld%flt)

    ! overwrite fields at time t with fields at time t+dt
       
    call finalize_fields(pb%fld%flt,pb%mdl,pb%mp)
    
  end subroutine advance_time_step


  subroutine step_substep(pb,dt,dts,accept_step)
    ! STEP_SUBSTEP advances problem through one elastodynamic time step with
    ! substeps for friction law and diffusion equations with interpolation of stress transfer
    ! 
    ! Modified: 9 February 2007

    use problem, only : problem_type
    use rates, only : set_rates
    use substep, only : integrate_substep

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DT = time step
    ! DTS = substep length
    ! ACCEPT_STEP = accept integration step
    
    type(problem_type),intent(inout) :: pb
    real(pr),intent(in) :: dt
    real(pr),intent(inout) :: dts
    logical,intent(out) :: accept_step

    ! Internal Parameters:
    ! STAGE = stage holding values at end of elastodynamic time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values

    integer(pin),parameter :: stage = 2
    character(64),parameter :: stress_transfer = 'convolution'

    ! go through smaller time steps to integrate elasticity+friction law
          
    call integrate_substep(pb,dt,dts,accept_step)

    ! set rates at t+dt

    call set_rates(pb,stage,stress_transfer,accept_step,update_history=.false.)

  end subroutine step_substep


end module problem_routines

