! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module rates

  ! RATES contains routines to set the rates (velocity, etc.) in the system of
  ! functional differential equations to be solved
  ! 
  ! Modified: 22 February 2007

  use constants, only : pr,pin

  implicit none

contains


  subroutine init_rates(pb,accept_step)
    ! INIT_RATES calculates initial rates (velocity, state-rate, rates of 
    ! temperature and pressure change, etc.) and stores initial values of
    ! convolution history
    ! 
    ! Modified: 22 February 2007

    use problem, only : problem_type
    use friction_routines, only : normal_friction,solve_friction,state_rate
    use fourier, only : fft_current_U,fft_history,store_history
    use load, only : set_load
    use thermpres, only : thermpres_rate
    use mpi_routines, only : is_master,master
    use mpi

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! ACCEPT_STEP = accept integration step

    type(problem_type),intent(inout) :: pb
    logical,intent(out) :: accept_step

    ! Internal Parameters:
    ! IERR = MPI error flag    

    integer(pin) :: ierr

    if (is_master(pb%mp)) then

       ! set time-dependent loads s0(t)
       
       call set_load(pb%mdl,pb%fld%ld,pb%fld%flt)
       
       ! FFT displacement/slip U(t)
       
       call fft_current_U(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage=1)
       
    end if

    ! if solving static elasticity problem, set all rates to zero and return
       
    if (pb%mdl%method=='static') then
          
       call zero_rates(pb%mdl,pb%fld%flt,pb%mp)
          
       accept_step = .true.
       
       return
       
    end if

    if (is_master(pb%mp)) then
       
       ! set fault-normal velocity vz(t) and effective normal stress sz(t)
       
       call normal_friction(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage=1)
       
       ! solve elasticity equation coupled with friction for
       ! V(t) and s(t) at each point using
       ! U(t), f(t), Q(t), p(t), T(t)
       
       call solve_friction(pb%fld%flt,pb%mdl,pb%fri,stage=1,accept=accept_step)
       
       ! if unacceptable, reject this step and exit routine
       ! (usually to try again with smaller step size)
       
       if (.not.accept_step) return
       
       ! calculate state-rate dQ(t)
       
       call state_rate(pb%fld%flt,pb%mdl,pb%fri,stage=1)
       
       ! forward FFT displacement/slip U(t) or velocity V(t) for use in convolution routines
    
       call fft_history(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage=1)

    end if

    ! calculate rates dT(t) and dp(t) in thermal pressurization equations
    
    call thermpres_rate(pb%mdl,pb%fld%flt,pb%fld%tp,pb%mp,stage=1)

    ! store current step in history vector for use in convolution routines
    
    call store_history(pb%mdl,pb%cnv,pb%mp,update_history=.true.)

    call MPI_Bcast(accept_step,1,MPI_LOGICAL,master,MPI_COMM_WORLD,ierr)

  end subroutine init_rates


  subroutine set_rates(pb,stage,stress_transfer,accept_step,update_history,t1,t2)
    ! SET_RATES calculates rates (velocity, state-rate, rates of 
    ! temperature and pressure change, etc.) at some stage.  Stress
    ! transfer is calculated by either 1. FFT, convolution, inverse FFT, or 2. interpolation.
    ! Note that storage of slip/slip velocity history is done outside this routine.
    ! 
    ! Modified: 22 February 2007

    use problem, only : problem_type
    use friction_routines, only : normal_friction,solve_friction,state_rate
    use convolution_routines, only : do_convolution
    use fourier, only : fft_current_U,fft_history,store_history,inverse_fft_F,set_normal_displacement
    use fault, only : set_stress_transfer,interp_stress_transfer
    use load, only : set_load
    use thermpres, only : thermpres_rate
    use io, only : error
    use mpi_routines, only : is_master,master
    use mpi

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! STAGE = integration stage
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values
    ! ACCEPT_STEP = accept integration step
    ! UPDATE_HISTORY = update history for convolution or overwrite
    ! T1 = time at beginning of elastodynamic step
    ! T2 = time at end of elastodynamic step

    type(problem_type),intent(inout) :: pb
    integer(pin),intent(in) :: stage
    character(*),intent(in) :: stress_transfer
    logical,intent(out) :: accept_step
    logical,intent(in),optional :: update_history
    real(pr),intent(in),optional :: t1,t2

    ! Internal Parameters:
    ! METHOD = interpolation method
    ! STAGE_F = stage index of stress transfer
    ! IERR = MPI error flag    

    character(64) :: method
    integer(pin) :: stage_f,ierr

    if (is_master(pb%mp)) then

       ! set time-dependent loads s0(t)
       
       call set_load(pb%mdl,pb%fld%ld,pb%fld%flt)
       
       ! FFT displacement/slip U(t)
       
       call fft_current_U(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage)
       
       ! set fault-normal displacements (when opening is not permitted)
       
       if (pb%mdl%method=='static') &
            call set_normal_displacement(pb%mdl,pb%krn,pb%fld%flt,pb%cnv,pb%fft,stage)
       
    end if

    select case(stress_transfer)

    case('convolution')

       if (is_master(pb%mp)) then

          if (.not.(present(update_history))) &
               call error('update_history must be present for convolution','set_rates')

          ! forward FFT displacement/slip U(t) or velocity V(t) for use in convolution routines
              
          call fft_history(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage)

       end if

       ! store current step in history vector for use in convolution routines
    
       call store_history(pb%mdl,pb%cnv,pb%mp,update_history)

       ! perform time convolution to get F(t)

       call do_convolution(pb%mdl,pb%krn,pb%cnv,pb%mp,update_history)

       ! inverse FFT F(t) to f(t)

       if (is_master(pb%mp)) then

          select case(pb%mdl%method)
          case default
             stage_f = 0
          case('substep')
             stage_f = stage-1
          end select

          call inverse_fft_F(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage_f)
          call set_stress_transfer(pb%mdl,pb%fld%flt,stage_f)

       end if

    case ('interpolate')

       if (is_master(pb%mp)) then

          if (.not.(present(t1).and.present(t2))) &
               call error('t1 and t2 must be present for interpolation','set_rates')
          
          ! interpolate stress transfer f(t)

          ! set method

          if (pb%mdl%iter==0) then

             ! explicit or predictor

             select case(pb%mdl%interp_method)
             case('forward')
                if(pb%mdl%n==1) then
                   method = 'linear_forward'
                else
                   method = 'quadratic'
                end if
             case('backward')
                if(pb%mdl%n==1) then
                   method = 'constant'
                else
                   method = 'linear_backward'
                end if
             case('safe1','safe2')
                method = 'constant'
             end select

          else

             ! corrector

             select case(pb%mdl%interp_method)
             case('forward')
                if(pb%mdl%n==1) then
                   method = 'linear_forward'
                else
                   method = 'quadratic'
                end if
             case('backward','safe2')
                if(pb%mdl%n==1) then
                   method = 'linear_forward'
                else
                   method = 'quadratic'
                end if
             case('safe1')
                method = 'linear_forward'
             end select

          end if
          
          call interp_stress_transfer(pb%fld%flt,pb%mdl,t1,t2,pb%mdl%t,method)

       end if

    end select

    ! if solving static elasticity problem, set all rates to zero and return
    
    if (pb%mdl%method=='static') then
       
       call zero_rates(pb%mdl,pb%fld%flt,pb%mp)
       
       accept_step = .true.
       
       return
       
    end if

    if (is_master(pb%mp)) then

       ! set fault-normal velocity vz(t) and effective normal stress sz(t)
       
       call normal_friction(pb%mdl,pb%fld%flt,pb%cnv,pb%fft,stage)
       
       ! solve elasticity equation coupled with friction for
       ! V(t) and s(t) at each point using
       ! U(t), f(t), Q(t), p(t), T(t)
       
       call solve_friction(pb%fld%flt,pb%mdl,pb%fri,stage,accept=accept_step)

    end if
       
    call MPI_Bcast(accept_step,1,MPI_LOGICAL,master,MPI_COMM_WORLD,ierr)

    ! if unacceptable, reject this step and exit routine
    ! (usually to try again with smaller step size)
    
    if (.not.accept_step) return
       
    ! calculate state-rate dQ(t)
       
    if (is_master(pb%mp)) call state_rate(pb%fld%flt,pb%mdl,pb%fri,stage)

    ! calculate rates dT(t) and dp(t) in thermal pressurization equations
    
    call thermpres_rate(pb%mdl,pb%fld%flt,pb%fld%tp,pb%mp,stage)
      
  end subroutine set_rates


  subroutine zero_rates(mdl,flt,mp)
    ! ZERO_RATES sets all rates (velocity, state-rate, rates of 
    ! temperature and pressure change, etc.) to zero
    ! 
    ! Modified: 14 February 2007

    use constants, only : zero
    use model, only : model_type
    use fault, only : fault_type
    use mpi_routines, only : mpi_type,is_master

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! MP = MPI variables

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(mpi_type),intent(in) :: mp

    ! BEGIN DEBUG (FIXED RATE EQUATION TO TEST ADAPTIVE INTEGRATION)
    !real(pr),parameter :: a=10._pr
    !real(pr) :: t,R
    !t = pb%mdl%t
    !R = exp(-t/a)*(sin(t/a)+(t/a)*(cos(t/a)-sin(t/a)))
    !if (pb%mdl%bm) then
    !   pb%fld%flt%vxp(:,:,stage) = R
    !   pb%fld%flt%vxm(:,:,stage) = R
    !else
    !   pb%fld%flt%Vx(:,:,stage) = R
    !end if
    !pb%fld%flt%V( :,:,stage) = R
    !if (associated(pb%fld%flt%dQ)) pb%fld%flt%dQ(:,:,stage) = R
    !if (associated(pb%fld%flt%dT)) pb%fld%flt%dT(:,:,:,stage) = R
    !if (associated(pb%fld%flt%dp)) pb%fld%flt%dp(:,:,:,stage) = R
    !accept_step = .true.
    !return
    ! END DEBUG

    if (is_master(mp)) then

       if (mdl%bm) then
          flt%vxp = zero
          flt%vxm = zero
          flt%vyp = zero
          flt%vym = zero
          flt%vzp = zero
          flt%vzm = zero
       else
          flt%Vx = zero
          flt%Vy = zero
       end if
       
       if (allocated(flt%dQ)) flt%dQ = zero
       
    end if

    if (allocated(flt%dT)) flt%dT = zero
    if (allocated(flt%dp)) flt%dp = zero

  end subroutine zero_rates


end module rates
