! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module ratestate

  ! RATESTATE contains variables and routines related to the
  ! rate-and-state friction laws
  ! 
  ! Modified: 20 April 2007

  use constants, only : pr,pin
  use asperity, only : asperity_list

  implicit none

  ! RATESTATE_TYPE is a derived type containing variables related to 
  ! rate-and-state friction law
  !
  ! Parameters:
  ! ASPERITY_FILE = add heterogeneity to initial rate-and-state fields
  !      T = use asperity array data file
  !      F = do not use asperity array data file
  ! FILENAME = asperity file name
  ! ASCII = asperity file format
  !      T = ascii
  !      F = binary
  ! LIST = singly-linked list of asperities
  ! FORM = which form of rate-and-state to use, options are:
  !      DR = Dieterich-Ruina
  !      LN = linearized Dieterich-Ruina
  !      RG = regularized Dieterich-Ruina
  !      EW = Dieterich-Ruina with enhanced weakening
  !      SL = slip law with low-speed weakening
  !      FL = slip law with flash-heating
  !      AF = aging law with flash-heating
  !      CF = composite law (slip law with aging at V<Vc) with flash-heating
  ! F0 = reference (or unweakened) friction coefficient
  ! V0 = reference sliding velocity
  ! A = direct effect parameter
  ! B = healing parameter b
  ! L = state evolution distance
  ! VW = weakening velocity
  ! FW = weakened friction coefficient
  ! TC = temperature scale, (shear contact strength)/(specific heat)
  ! TW = weakening temperature
  ! VTH = thermal velocity (thermal diffusivity)/(contact dimension)
  ! VC = cut-off velocity for composite law

  type ratestate_type
     logical :: asperity_file,ascii
     character(64) :: filename,form
     type(asperity_list) :: list
     real(pr),dimension(:,:),allocatable :: f0,V0,a,b,L,Vw,fw,Tc,Tw,Vth,Vc
  end type ratestate_type

contains


  subroutine read_ratestate(ninput,rs,temperature)
    ! READ_RATESTATE reads in rate-and-state variables from file
    ! 
    ! Modified: 20 April 2007

    use io, only : error
    use asperity, only : read_asperity_list

    implicit none
    
    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! RS = rate-and-state variables
    ! TEMPERATURE = whether or not friction law requires temperature field

    integer(pin),intent(in) :: ninput
    logical,intent(out) :: temperature
    type(ratestate_type),intent(out) :: rs

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! ASPERITY_LIST = add heterogeneity to initial rs fields with list
    ! ASPERITY_FILE = add heterogeneity to initial rs fields from file
    ! FILENAME = asperity file name
    ! ASCII = asperity file format
    ! NDATA = number of field perturbations
    ! FORM = which form of rate-and-state to use

    integer(pin) :: stat,ndata
    logical :: asperity_list,asperity_file,ascii
    character(64) :: filename,form

    ! make namelist of user input variables

    namelist /ratestate_list/ asperity_list,asperity_file,filename,ascii,form

    ! default values

    asperity_list = .false.
    asperity_file = .false.
    ascii = .false.

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=ratestate_list,iostat=stat)
    if (stat/=0) call error("Error reading namelist 'ratestate_list' in .in file",'read_ratestate')

    ! assign input variables to components of derived type

    rs%asperity_file = asperity_file
    rs%filename = 'problems/' // adjustl(filename)
    rs%ascii = ascii
    rs%form = adjustl(form)

    ! set flag indicating if temperature will be required for particular friction law

    select case(rs%form)
    case('FL','AF','CF')
       temperature = .true.
    case default
       temperature = .false.
    end select

    ! asperity list, set ndata based on form of friction law

    select case(rs%form)
    case('DR','LN','RG','SL')
       ndata = 5 ! for f0,V0,a,b,L
    case('EW')
       ndata = 6 ! for f0,V0,a,b,L,Vw
    case('FL','AF')
       ndata = 9 ! for f0,V0,a,b,L,fw,Tc,Tw,Vth
    case('CF')
       ndata = 10 ! for f0,V0,a,b,L,fw,Tc,Tw,Vth,Vc
    case default
       call error('Invalid form of rate-and-state friction','read_ratestate')
    end select

    if (asperity_list) &
         rs%list = read_asperity_list(ninput,'ratestate',ndata)

  end subroutine read_ratestate


  subroutine init_ratestate(necho,mdl,rs)
    ! INIT_RATESTATE initializes rate-and-state variables
    ! 
    ! Modified: 20 April 2007

    use constants, only : zero
    use io, only : write_matlab,open_array_data,close_array_data,read_array_data
    use model, only : model_type
    use asperity, only : assign_list_data,destroy_asperity_list

    implicit none
    
    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! MDL = model variables
    ! RS = rate-and-state variables

    integer(pin),intent(in) :: necho
    type(model_type),intent(in) :: mdl
    type(ratestate_type),intent(inout) :: rs

    ! Internal Parameters:
    ! FILEUNIT = asperity file unit

    integer(pin) :: fileunit

    select case(rs%form)

    case('DR','LN','RG','SL')

       ! allocate memory for rate-and-state parameter arrays
       
       allocate(rs%f0(mdl%nx,mdl%ny))
       allocate(rs%V0(mdl%nx,mdl%ny))
       allocate(rs%a( mdl%nx,mdl%ny))
       allocate(rs%b( mdl%nx,mdl%ny))
       allocate(rs%L( mdl%nx,mdl%ny))
       
       ! initialize to zero

       rs%f0 = zero
       rs%V0 = zero
       rs%a  = zero
       rs%b  = zero
       rs%L  = zero
       
       ! read asperity arrays from file
       
       if (rs%asperity_file) then
          call open_array_data(rs%filename,fileunit,rs%ascii)
          call read_array_data(fileunit,rs%f0,rs%ascii,'f0')
          call read_array_data(fileunit,rs%V0,rs%ascii,'V0')
          call read_array_data(fileunit,rs%a ,rs%ascii,'a')
          call read_array_data(fileunit,rs%b ,rs%ascii,'b')
          call read_array_data(fileunit,rs%L ,rs%ascii,'L')
          call close_array_data(fileunit)
       end if
       
       ! use list data to perturb fields
       
       call assign_list_data(rs%f0,rs%list,1,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%V0,rs%list,2,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%a ,rs%list,3,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%b ,rs%list,4,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%L ,rs%list,5,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       
    case('EW')

       ! allocate memory for rate-and-state parameter arrays
       
       allocate(rs%f0(mdl%nx,mdl%ny))
       allocate(rs%V0(mdl%nx,mdl%ny))
       allocate(rs%a( mdl%nx,mdl%ny))
       allocate(rs%b( mdl%nx,mdl%ny))
       allocate(rs%L( mdl%nx,mdl%ny))
       allocate(rs%Vw(mdl%nx,mdl%ny))
       
       ! initialize to zero

       rs%f0 = zero
       rs%V0 = zero
       rs%a  = zero
       rs%b  = zero
       rs%L  = zero
       rs%Vw = zero
       
       ! read asperity arrays from file
       
       if (rs%asperity_file) then
          call open_array_data(rs%filename,fileunit,rs%ascii)
          call read_array_data(fileunit,rs%f0,rs%ascii,'f0')
          call read_array_data(fileunit,rs%V0,rs%ascii,'V0')
          call read_array_data(fileunit,rs%a ,rs%ascii,'a')
          call read_array_data(fileunit,rs%b ,rs%ascii,'b')
          call read_array_data(fileunit,rs%L ,rs%ascii,'L')
          call read_array_data(fileunit,rs%Vw,rs%ascii,'Vw')
          call close_array_data(fileunit)
       end if
       
       ! use list data to perturb fields
       
       call assign_list_data(rs%f0,rs%list,1,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%V0,rs%list,2,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%a ,rs%list,3,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%b ,rs%list,4,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%L ,rs%list,5,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Vw,rs%list,6,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       
    case('FL','AF')

       ! allocate memory for rate-and-state parameter arrays
       
       allocate(rs%f0( mdl%nx,mdl%ny))
       allocate(rs%V0( mdl%nx,mdl%ny))
       allocate(rs%a(  mdl%nx,mdl%ny))
       allocate(rs%b(  mdl%nx,mdl%ny))
       allocate(rs%L(  mdl%nx,mdl%ny))
       allocate(rs%fw( mdl%nx,mdl%ny))
       allocate(rs%Tc( mdl%nx,mdl%ny))
       allocate(rs%Tw( mdl%nx,mdl%ny))
       allocate(rs%Vth(mdl%nx,mdl%ny))

       ! initialize to zero

       rs%f0  = zero
       rs%V0  = zero
       rs%a   = zero
       rs%b   = zero
       rs%L   = zero
       rs%fw  = zero
       rs%Tc  = zero
       rs%Tw  = zero
       rs%Vth = zero
       
       ! read asperity arrays from file
       
       if (rs%asperity_file) then
          call open_array_data(rs%filename,fileunit,rs%ascii)
          call read_array_data(fileunit,rs%f0 ,rs%ascii,'f0')
          call read_array_data(fileunit,rs%V0 ,rs%ascii,'V0')
          call read_array_data(fileunit,rs%a  ,rs%ascii,'a')
          call read_array_data(fileunit,rs%b  ,rs%ascii,'b')
          call read_array_data(fileunit,rs%L  ,rs%ascii,'L')
          call read_array_data(fileunit,rs%fw ,rs%ascii,'fw')
          call read_array_data(fileunit,rs%Tc ,rs%ascii,'Tc')
          call read_array_data(fileunit,rs%Tw ,rs%ascii,'Tw')
          call read_array_data(fileunit,rs%Vth,rs%ascii,'Vth')
          call close_array_data(fileunit)
       end if
       
       ! use list data to perturb fields
       
       call assign_list_data(rs%f0 ,rs%list,1,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%V0 ,rs%list,2,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%a  ,rs%list,3,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%b  ,rs%list,4,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%L  ,rs%list,5,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%fw ,rs%list,6,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Tc ,rs%list,7,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Tw ,rs%list,8,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Vth,rs%list,9,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       
    case('CF')

       ! allocate memory for rate-and-state parameter arrays
       
       allocate(rs%f0( mdl%nx,mdl%ny))
       allocate(rs%V0( mdl%nx,mdl%ny))
       allocate(rs%a(  mdl%nx,mdl%ny))
       allocate(rs%b(  mdl%nx,mdl%ny))
       allocate(rs%L(  mdl%nx,mdl%ny))
       allocate(rs%fw( mdl%nx,mdl%ny))
       allocate(rs%Tc( mdl%nx,mdl%ny))
       allocate(rs%Tw( mdl%nx,mdl%ny))
       allocate(rs%Vth(mdl%nx,mdl%ny))
       allocate(rs%Vc( mdl%nx,mdl%ny))

       ! initialize to zero

       rs%f0  = zero
       rs%V0  = zero
       rs%a   = zero
       rs%b   = zero
       rs%L   = zero
       rs%fw  = zero
       rs%Tc  = zero
       rs%Tw  = zero
       rs%Vth = zero
       rs%Vc  = zero
       
       ! read asperity arrays from file
       
       if (rs%asperity_file) then
          call open_array_data(rs%filename,fileunit,rs%ascii)
          call read_array_data(fileunit,rs%f0 ,rs%ascii,'f0')
          call read_array_data(fileunit,rs%V0 ,rs%ascii,'V0')
          call read_array_data(fileunit,rs%a  ,rs%ascii,'a')
          call read_array_data(fileunit,rs%b  ,rs%ascii,'b')
          call read_array_data(fileunit,rs%L  ,rs%ascii,'L')
          call read_array_data(fileunit,rs%fw ,rs%ascii,'fw')
          call read_array_data(fileunit,rs%Tc ,rs%ascii,'Tc')
          call read_array_data(fileunit,rs%Tw ,rs%ascii,'Tw')
          call read_array_data(fileunit,rs%Vth,rs%ascii,'Vth')
          call read_array_data(fileunit,rs%Vc ,rs%ascii,'Vc')
          call close_array_data(fileunit)
       end if
       
       ! use list data to perturb fields
       
       call assign_list_data(rs%f0 ,rs%list,1 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%V0 ,rs%list,2 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%a  ,rs%list,3 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%b  ,rs%list,4 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%L  ,rs%list,5 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%fw ,rs%list,6 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Tc ,rs%list,7 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Tw ,rs%list,8 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Vth,rs%list,9 ,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)
       call assign_list_data(rs%Vc ,rs%list,10,mdl%x,mdl%y,mdl%dx,mdl%dy,mdl%dim)

    end select

    ! deallocate memory for asperity list

    call destroy_asperity_list(rs%list)

    ! output variable values into matlab file

    call write_matlab(necho,'form',rs%form,'rs')
    
  end subroutine init_ratestate


  subroutine destroy_ratestate(rs)
    ! DESTROY_RATESTATE destroys derived type rs
    ! 
    ! Modified: 26 November 2006

    implicit none

    ! I/O Parameters:
    ! RS = rate-and-state variables

    type(ratestate_type),intent(inout) :: rs

    ! deallocate memory assigned to allocatable arrays

    if (allocated(rs%f0)) deallocate(rs%f0)
    if (allocated(rs%V0)) deallocate(rs%V0)
    if (allocated(rs%a)) deallocate(rs%a)
    if (allocated(rs%b)) deallocate(rs%b)
    if (allocated(rs%L)) deallocate(rs%L)
    if (allocated(rs%Vw)) deallocate(rs%Vw)
    if (allocated(rs%fw)) deallocate(rs%fw)
    if (allocated(rs%Tc)) deallocate(rs%Tc)
    if (allocated(rs%Tw)) deallocate(rs%Tw)
    if (allocated(rs%Vth)) deallocate(rs%Vth)
    if (allocated(rs%Vc)) deallocate(rs%Vc)

  end subroutine destroy_ratestate


  subroutine fric_rs(Q,V,i,j,f,dfdV,err,rs)
    ! FRIC_RS returns the coefficient of friction for a rate-and-state friction law
    ! 
    ! Modified: 20 April 2007

    use constants, only : zero,one,two
    use utilities, only : arcsinh

    implicit none

    ! I/O Parameters:
    ! Q = state variable
    ! V = slip velocity
    ! I = index in x direction
    ! J = index in y direction
    ! F = coefficient of friction
    ! DFDV = derivative of friction with respect to velocity
    ! ERR = error flag (non-zero if friction isn't defined at particular V)
    ! RS = rate-and-state variables

    real(pr),intent(in) :: Q,V
    integer(pin),intent(in) :: i,j
    real(pr),intent(out) :: f,dfdV
    integer(pin),intent(out) :: err
    type(ratestate_type),intent(in) :: rs

    ! Internal  Parameters:
    ! WK = parameter combination in EW law
    ! O = parameter combination in RG law

    real(pr) :: wk,O

    err = 0

    select case(rs%form)
    case('LN')
       f = rs%f0(i,j)+rs%a(i,j)*V/rs%V0(i,j)+rs%b(i,j)*rs%V0(i,j)*Q/rs%L(i,j)
       dfdV = rs%a(i,j)/rs%V0(i,j)
    case('DR','CF','AF')
       if (V<=zero) then
          err = 1
          return
       end if
       f = rs%f0(i,j)+rs%a(i,j)*log(V/rs%V0(i,j))+rs%b(i,j)*log(rs%V0(i,j)*Q/rs%L(i,j))
       dfdV = rs%a(i,j)/V
    case('RG')
       O = exp((rs%f0(i,j)+rs%b(i,j)*log(rs%V0(i,j)*Q/rs%L(i,j)))/rs%a(i,j))/(two*rs%V0(i,j))
       f = rs%a(i,j)*arcsinh(V*O)
       dfdV = rs%a(i,j)*O/sqrt(one+(V*O)**2)
    case('EW')
       if (V<=zero) then
          err = 1
          return
       end if
       wk = one+rs%L(i,j)/(Q*rs%Vw(i,j))
       f = (rs%f0(i,j)+rs%a(i,j)*log(V/rs%V0(i,j))+rs%b(i,j)*log(rs%V0(i,j)*Q/rs%L(i,j)))/wk
       dfdV = rs%a(i,j)/(V*wk)
    case('FL','SL')
       if (V<=zero) then
          err = 1
          return
       end if
       f = rs%a(i,j)*log(V/rs%V0(i,j))+Q
       dfdV = rs%a(i,j)/V
    end select

  end subroutine fric_rs


  subroutine strength_rs(Q,sz,V,i,j,stV,dstdV,err,rs)
    ! STRENGTH_RS returns the strength for rate-and-state law
    ! 
    ! Modified: 20 September 2006

    use constants, only : zero

    implicit none

    ! I/O Parameters:
    ! Q = state variable
    ! SZ = stress in z direction
    ! V = slip velocity
    ! I = index in x direction at which friction is applied
    ! J = index in y direction at which friction is applied  
    ! STV = velocity-dependent strength
    ! DSTDV = derivative of strength with respect to slip velocity
    ! ERR = error flag (non-zero if strength isn't defined at particular V)
    ! RS = rate-and-state variables

    real(pr),intent(in) :: Q,sz,V
    integer(pin),intent(in) :: i,j
    real(pr),intent(out) :: stV,dstdV
    integer(pin),intent(out) :: err
    type(ratestate_type),intent(in) :: rs

    ! Internal  Parameters:
    ! F = coefficient of friction
    ! DFDV = derivative of f with respect to V
    ! SN = normal stress, positive in compression

    real(pr) :: f,dfdV,sn

    ! calculate normal stress

    sn = max(-sz,zero)

    ! calculate coefficient of friction and derivative with respect to V

    call fric_rs(Q,V,i,j,f,dfdV,err,rs)

    ! set strength to product of coefficient of friction and normal stress

    stV = f*sn
    dstdV = dfdV*sn
    
  end subroutine strength_rs


  function evolution_rs(V,Q,i,j,rs,T) result(dQ)
    ! EVOLUTION_RS returns the state-rate
    ! 
    ! Modified: 20 April 2007

    use constants, only : one,pi

    implicit none

    ! I/O Parameters:
    ! V = slip velocity
    ! Q = state variable
    ! I = index in x direction at which friction is applied
    ! J = index in y direction at which friction is applied  
    ! RS = rate-and-state variables
    ! T = temperature
    ! DQ = state-rate

    real(pr),intent(in) :: V,Q
    integer(pin),intent(in) :: i,j
    type(ratestate_type),intent(in) :: rs
    real(pr),intent(in),optional :: T
    real(pr) :: dQ

    ! Internal  Parameters:
    ! F = friction
    ! FSS = steady-state friction
    ! FSSL = low-speed steady-state friction
    ! VW = weakening velocity
    ! QSS = steady-state value of state

    real(pr) :: f,fss,fssl,Vw,Qss

    select case(rs%form)
    case('LN') ! Q is perturbation in theta
       dQ = -V/rs%V0(i,j)-rs%V0(i,j)*Q/rs%L(i,j)
    case('DR','EW','RG') ! Q is theta
       Qss = rs%L(i,j)/V
       dQ = one-Q/Qss
    case('SL')           ! Q is Psi
       fss = rs%f0(i,j)+(rs%a(i,j)-rs%b(i,j))*log(V/rs%V0(i,j))
       f = rs%a(i,j)*log(V/rs%V0(i,j))+Q
       dQ = -V/rs%L(i,j)*(f-fss)
    case('FL')           ! Q is Psi
       Vw = pi*rs%Vth(i,j)*((rs%Tw(i,j)-T)/rs%Tc(i,j))**2
       fssl = rs%f0(i,j)+(rs%a(i,j)-rs%b(i,j))*log(V/rs%V0(i,j))
       if (V<=Vw) then
          fss = fssl
       else
          fss = rs%fw(i,j)+(fssl-rs%fw(i,j))*Vw/V
       end if
       f = rs%a(i,j)*log(V/rs%V0(i,j))+Q
       dQ = -V/rs%L(i,j)*(f-fss)
    case('AF')           ! Q is theta
       Vw = pi*rs%Vth(i,j)*((rs%Tw(i,j)-T)/rs%Tc(i,j))**2
        if (V<=Vw) then
          Qss = rs%L(i,j)/V
       else
          Qss = rs%L(i,j)/V*(V/rs%V0(i,j))**((one-rs%a(i,j)/rs%b(i,j))*(one-Vw/V))* &
               exp(-(rs%f0(i,j)-rs%fw(i,j))/rs%b(i,j)*(one-Vw/V))
       end if
       dQ = -V/rs%L(i,j)*(Q-Qss)
    case('CF')          ! Q is theta
       Vw = pi*rs%Vth(i,j)*((rs%Tw(i,j)-T)/rs%Tc(i,j))**2
       if (V<=Vw) then
          dQ = exp(-V/rs%Vc(i,j))-V*Q/rs%L(i,j)*log(V*Q/rs%L(i,j))
       else
          fss = rs%fw(i,j)+(rs%f0(i,j)+(rs%a(i,j)-rs%b(i,j))*log(V/rs%V0(i,j))-rs%fw(i,j))*Vw/V
          f = rs%f0(i,j)+rs%a(i,j)*log(V/rs%V0(i,j))+rs%b(i,j)*log(rs%V0(i,j)*Q/rs%L(i,j))
          dQ = -V*Q/rs%L(i,j)*(f-fss)/rs%b(i,j)
       end if
    end select

  end function evolution_rs


  function check_state_rs(Q,rs) result(accept)
    ! CHECK_STATE_RS checks if state variable value is acceptable
    ! for the particular rate-and-state law being used (some do not
    ! permit negative state values)
    ! 
    ! Modified: 14 February 2007

    use constants, only : zero

    implicit none

    ! I/O Parameters:
    ! Q = state variable
    ! RS = rate-and-state variables

    real(pr),intent(in) :: Q
    type(ratestate_type),intent(in) :: rs
    logical :: accept

    accept = .true.

    select case(rs%form)
    case('DR','RG','EW','CF','AF')
       if (Q<=zero) then
          accept = .false.
       end if
    end select

    !if (.not.accept) print *, 'reject negative Q=',Q

  end function check_state_rs
  

end module ratestate
