! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module substep
  ! SUBSTEP contains routines for integrating over several small substeps within
  ! an elastodynamic time step using an interpolated value of stress transfer
  ! 
  ! Modified: 10 February 2007

  use constants, only : pr,pin

  implicit none


contains


  subroutine integrate_substep(pb,dte,dt,accept_step)
    ! INTEGRATE_SUBSTEP integrates diffusion and state-evolution equations over
    ! multiple substeps within one elastodynamic time step
    ! 
    ! Modified: 10 February 2007

    use problem, only : problem_type
    use integration, only : move_fields

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DTE = elastodynamic time step
    ! DT = substep length
    ! ACCEPT_STEP = accept integration step

    type(problem_type),intent(inout) :: pb
    real(pr),intent(in) :: dte
    real(pr),intent(inout) :: dt
    logical,intent(out) :: accept_step

    select case(pb%mdl%substep_method)

    case('iterative')

       call substep_iterative(pb,dte,dt,accept_step)

    case('adaptive')

       call substep_adaptive(pb,dte,dt,accept_step)

    end select

  end subroutine integrate_substep


  subroutine substep_iterative(pb,dte,dt,accept_step)
    ! SUBSTEP_ITERATIVE integrates diffusion and state-evolution equations over constant substeps
    ! within one elastodynamic time step using iterative predictor-corrector method
    ! 
    ! Modified: 9 February 2007

    use problem, only : problem_type
    use integration, only : move_fields
    use time_step, only : step_explicit,step_trapezoid
    use mpi_routines, only : is_master

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DTE = elastodynamic time step
    ! DT = substep time step
    ! ACCEPT_STEP = accept integration step

    type(problem_type),intent(inout) :: pb
    real(pr),intent(in) :: dte
    real(pr),intent(inout) :: dt
    logical,intent(out) :: accept_step

    ! Internal variables:
    ! T1 = time at beginning of elastodynamic time step
    ! T2 = time at end of elastodynamic time step
    ! ITER = iteration index
    ! N = index of substep
    ! NT = number of substeps in elastodynamic time step
    ! S0 = stage at beginning of time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values

    real(pr) :: t1,t2
    integer(pin) :: iter,n,nt
    integer(pin),parameter :: s0 = 2
    character(64),parameter :: stress_transfer = 'interpolate'

    ! fix number of substeps

    nt = pb%mdl%nsubs

    ! calculate substep length
    
    dt = dte/real(nt,pr)
    
    ! store initial and final time

    t1 = pb%mdl%t
    t2 = pb%mdl%t+dte
       
    ! transfer values from stage=1 to stage=2 (to start integration)
       
    call move_fields(pb%fld%flt,pb%mdl,pb%mp,n1=s0,n2=s0-1)
        
    ! loop over substeps
    
    do n = 1,nt
       
       ! take explicit step

       call step_explicit(pb,dt,s0,stress_transfer,accept_step,t1=t1,t2=t2)

       ! if unacceptable, reject this step and exit routine
    
       if (.not.accept_step) return

       ! iterate to improve solution

       do iter = 1,pb%mdl%substep_niter

          ! repeat step using trapezoidal rule
          
          call step_trapezoid(pb,dt,s0,stress_transfer,accept_step,t1=t1,t2=t2)
          
          ! if unacceptable, reject this step and exit routine
          
          if (.not.accept_step) return

       end do

       ! move field values from end of current time step [stage=3] to start of next [stage=2]
       
       call move_fields(pb%fld%flt,pb%mdl,pb%mp,n1=s0,n2=s0+1)
          
    end do

  end subroutine substep_iterative


  subroutine substep_adaptive(pb,dte,dt,accept_step)
    ! SUBSTEP_ADAPTIVE integrates diffusion and state-evolution equations over
    ! adaptively chosen substeps within one elastodynamic time step
    ! 
    ! Modified: 9 February 2007

    use constants, only : zero
    use problem, only : problem_type
    use integration, only : move_fields,move_error
    use time_step, only : step_rk
    use mpi_routines, only : is_master

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DTE = elastodynamic time step
    ! DT = substep length
    ! ACCEPT_STEP = accept integration step

    type(problem_type),intent(inout) :: pb
    real(pr),intent(in) :: dte
    real(pr),intent(inout) :: dt
    logical,intent(out) :: accept_step

    ! Internal variables:
    ! T0 = initial time (at beginning of substep time step)
    ! T1 = time at beginning of elastodynamic time step
    ! T2 = time at end of elastodynamic time step
    ! NT = number of substeps
    ! ATTEMPTS = number of attempts to take acceptable time step
    ! S0 = stage at beginning of time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values

    real(pr) :: t0,t1,t2
    integer(pin) :: nt,attempts
    integer(pin),parameter :: s0 = 2
    character(64),parameter :: stress_transfer = 'interpolate'

    ! store initial and final time

    t1 = pb%mdl%t
    t2 = pb%mdl%t+dte

    ! set substep counter to zero

    nt = 0
       
    ! transfer values from stage=1 to stage=2 (to start integration)
    
    call move_fields(pb%fld%flt,pb%mdl,pb%mp,n1=s0,n2=s0-1)
       
    ! resent error estimate
    
    if (is_master(pb%mp)) then
       pb%fld%flt%aerr = zero
       pb%fld%flt%rerr = zero
    end if

    ! take multiple substeps
    
    do
       
       ! update substep counter

       nt = nt+1

       ! store initial time
       
       t0 = pb%mdl%t
    
       ! iterate over step sizes until acceptable

       attempts = 0

       do 

          ! update step size attempt counter
          
          attempts = attempts+1
                 
          ! start time from beginning of substep
          
          pb%mdl%t = t0

          ! prevent overshooting t2 by reducing time step to stop at exactly at t2 if dt too large
          
          if (pb%mdl%t+dt>t2) dt = t2-pb%mdl%t

          ! advance one time step, estimating error to predict new time step
       
          !print *, pb%mp%myid,'attempt',attempts

          call step_rk(pb,dt,s0,stress_transfer,accept_step,attempts,t0,t1,t2)

          !print *, pb%mp%myid,'substep accepted? ',accept_step

          ! exit loop searching for time step if step is acceptable
          
          if (accept_step) exit

       end do

       ! accept step

       ! move error estimate
       
       if (is_master(pb%mp)) call move_error(pb%fld%flt)
       
       ! move field values from end of current time step [stage=ns=rk%s+1] to start of next [stage=s0=2]
       
       call move_fields(pb%fld%flt,pb%mdl,pb%mp,n1=s0,n2=pb%mdl%ns)
          
       ! exit when end of elastodynamic time step is reached
       
       if (pb%mdl%t>=t2) exit

    end do

    !if (is_master(pb%mp)) &
    !     write(6,'(a,i6,a,i6)') 'n = ',pb%mdl%n,' adaptive substeps = ',nt

  end subroutine substep_adaptive


end module substep

