! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module thermpres

  ! THERMPRES contains variables and routines related to the
  ! thermal pressurization friction law
  ! 
  ! Modified: 2 March 2007

  use constants, only : pr,pin

  implicit none


  ! THERMPRES_TYPE is a derived type containing variables related to 
  ! thermal pressurization friction law
  !
  ! Parameters:
  ! K = thermal conductivity
  ! ALPHATH = thermal diffusivity
  ! ALPHAHY = hydraulic diffusivity
  ! LAMBDA = pore pressure rise per unit temprature rise
  ! SHAPE = shape of strain distribution
  ! W = width of deformation zone
  ! DTDZ = initial gradient in temperature, used to initialize temperture profile
  !      and used as the outer boundary condition on the diffusion equation

  type thermpres_type
     real(pr) :: K,alphath,alphahy,Lambda,dTdz,W
     character(64) :: shape
  end type thermpres_type


contains


  subroutine read_thermpres(ninput,tp)
    ! READ_THERMPRES reads in thermpres variables from file
    ! 
    ! Modified: 12 November 2006

    use constants, only : zero,half
    use io, only : error

    implicit none
    
    ! I/O Parameters:
    ! NINPUT = unit number for *.in input file
    ! TP = thermal pressurization variables

    integer(pin),intent(in) :: ninput
    type(thermpres_type),intent(out) :: tp

    ! Internal Parameters:
    ! STAT = I/O error flag
    ! K = thermal conductivity
    ! ALPHATH = theamal diffusivity
    ! ALPHAHY = hydraulic diffusivity
    ! LAMBDA = pore pressure rise per unit temerature rise
    ! DTDZ = initial gradient in temperature
    ! SHAPE = shape of strain distribution
    ! W = width of deformation zone

    integer(pin) :: stat
    real(pr) :: K,alphath,alphahy,Lambda,dTdz,W
    character(64) :: shape

    ! make namelist of user input variables

    namelist /thermpres_list/ K,alphath,alphahy,Lambda,dTdz,shape,W

    ! assign default values

    K = 2.07d-3
    alphath = 0.675_pr
    alphahy = 0.61_pr
    Lambda = 2.8_pr
    dTdz = zero
    shape = 'plane'
    W = half

    ! read namelist from input file, call error routine if needed

    rewind(ninput)
    read(ninput,nml=thermpres_list,iostat=stat)
    if (stat>0) call error("Error reading namelist 'thermpres_list' in .in file",'read_thermpres')

    ! assign input variables to components of derived type

    tp%K = K
    tp%alphath = alphath
    tp%alphahy = alphahy
    tp%Lambda = Lambda
    tp%dTdz = dTdz
    tp%shape = shape
    tp%W = W

  end subroutine read_thermpres


  subroutine init_thermpres_master(necho,flt,tp)
    ! INIT_THERMPRES_MASTER initializes thermpres variables (master)
    ! 
    ! Modified: 1 February 2007

    use io, only : write_matlab
    use fault, only : fault_type

    implicit none
    
    ! I/O Parameters:
    ! NECHO = unit number for output file
    ! FLT = fault variables
    ! TP = thermal pressurization variables
    ! T = temperature

    integer(pin),intent(in) :: necho
    type(thermpres_type),intent(in) :: tp
    type(fault_type),intent(in) :: flt

    ! return if not including this effect

    if (.not.flt%thermpres) return

    ! output variable values into matlab file

    call write_matlab(necho,'K',tp%K,'tp')
    call write_matlab(necho,'alphath',tp%alphath,'tp')
    call write_matlab(necho,'alphahy',tp%alphahy,'tp')
    call write_matlab(necho,'Lambda',tp%Lambda,'tp')
    call write_matlab(necho,'dTdz',tp%dTdz,'tp')
    call write_matlab(necho,'shape',tp%shape,'tp')
    call write_matlab(necho,'W',tp%W,'tp')

  end subroutine init_thermpres_master


  subroutine init_thermpres_slave(mdl,flt,tp,mp)
    ! INIT_THERMPRES_SLAVE initializes thermpres variables (slave)
    ! 
    ! Modified: 2 February 2007

    use constants, only : zero
    use io, only : write_matlab
    use model, only : model_type
    use fault, only : fault_type
    use mpi_routines, only : mpi_type,master,is_master,MPI_REAL_PR,scatterv
    use mpi

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! TP = thermal pressurization variables
    ! MP = MPI variables

    type(model_type),intent(in) :: mdl
    type(thermpres_type),intent(in) :: tp
    type(fault_type),intent(inout) :: flt
    type(mpi_type),intent(in) :: mp

    ! Internal Parameters:
    ! J = index in y direction
    ! K = index in z direction
    ! S = index of stage
    ! TO = initial profile of temperature
    ! IERR = MPI error flag
    ! COUNTS = MPI message sizes
    ! DISPLS = MPI message displacements

    integer(pin) :: j,k,s,ierr
    real(pr),dimension(:),allocatable :: To
    integer,dimension(0:mp%np-1) :: counts,displs

    ! return if not including this effect

    if (.not.flt%thermpres) return

    ! distribute variables from master to slaves

    call MPI_Bcast(tp%K      ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%alphath,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%alphahy,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%Lambda ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%dTdz   ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
    call MPI_Bcast(tp%W      ,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)

    call MPI_Bcast(tp%shape,64,MPI_CHARACTER,master,MPI_COMM_WORLD,ierr)

    ! allocate and initialize diffusion grids

    if (.not.is_master(mp)) then
       allocate(flt%T0(mdl%li:mdl%ui,mdl%ny,mdl%ns))
       allocate(flt%p0(mdl%li:mdl%ui,mdl%ny,mdl%ns))
    end if

    allocate(flt%T( mdl%nz,mdl%li:mdl%ui,mdl%ny,mdl%ns))
    allocate(flt%dT(mdl%nz,mdl%li:mdl%ui,mdl%ny,mdl%ns))
    allocate(flt%p( mdl%nz,mdl%li:mdl%ui,mdl%ny,mdl%ns))
    allocate(flt%dp(mdl%nz,mdl%li:mdl%ui,mdl%ny,mdl%ns))
    
    ! initialize temperature with linear profile and pressure as zero

    flt%T = zero
    flt%dT = zero
    flt%p = zero
    flt%dp = zero

    ! prepare for scatter

    call scatterv(mdl%nx,mp%np,counts,displs)

    ! scatter fault temperature field to slaves

    do s = 1,mdl%ns
       do j = 1,mdl%ny

          call MPI_Scatterv(flt%T0(:,j,s),counts,displs,MPI_REAL_PR, &
               flt%T0(:,j,s),counts(mp%myid),MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)

       end do
    end do

    ! add linear profile

    allocate(To(mdl%nz))

    if (is_master(mp)) To = tp%dTdz*mdl%z

    call MPI_Bcast(To,mdl%nz,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)

    forall (k=1:mdl%nz)
       flt%T(k,mdl%li:mdl%ui,:,:) = flt%T0(mdl%li:mdl%ui,:,:)+To(k)
    end forall

    deallocate(To)

  end subroutine init_thermpres_slave


  subroutine destroy_thermpres(tp,flt)
    ! DESTROY_THERMPRES destroys derived type tp
    ! 
    ! Modified: 1 February 2007

    use fault, only : fault_type

    implicit none

    ! I/O Parameters:
    ! TP = thermal pressurization variables
    ! FLT = fault variables

    type(thermpres_type),intent(inout) :: tp
    type(fault_type),intent(inout) :: flt

    if (allocated(flt%T)) deallocate(flt%T)
    if (allocated(flt%dT)) deallocate(flt%dT)
    if (allocated(flt%p)) deallocate(flt%p)
    if (allocated(flt%dp)) deallocate(flt%dp)

  end subroutine destroy_thermpres


  subroutine thermpres_rate(mdl,flt,tp,mp,stage)
    ! THERMPRES_RATE solves the 1D diffusion equations for thermal pressurization
    ! using explicit finite differences
    ! 
    ! Modified: 1 March 2007

    use model, only : model_type
    use fault, only : fault_type
    use mpi_routines, only : mpi_type,is_master,MPI_REAL_PR,master,scatterv
    use mpi

    implicit none
    
    ! I/O Parameters:
    ! MDL = model variables
    ! FLT = fault variables
    ! TP = thermal pressurization variables
    ! MP = MPI variables
    ! STAGE = integration step

    type(model_type),intent(in) :: mdl
    type(fault_type),intent(inout) :: flt
    type(thermpres_type),intent(in) :: tp
    type(mpi_type),intent(in) :: mp
    integer(pin),intent(in) :: stage

    ! Internal Parameters:
    ! I = index in x direction
    ! J = index in y direction
    ! Q = shear heating rate
    ! IERR = MPI error flag
    ! COUNTS = MPI message sizes
    ! DISPLS = MPI message displacements

    integer(pin) :: i,j,ierr
    real(pr),dimension(:,:),allocatable :: Q
    integer,dimension(0:mp%np-1) :: counts,displs

    ! return if not including thermal pressurization model
    
    if (.not.flt%thermpres) return
       
    ! calculate shear heat (master)

    if (is_master(mp)) then

       allocate(Q(mdl%nx,mdl%ny))

       !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(i,j)

       !$OMP DO

       do j = 1,mdl%ny
          do i = 1,mdl%nx
             
             Q(i,j) = shear_heat(flt,i,j,stage)
             
          end do
       end do
       
       !$OMP END DO

       !$OMP END PARALLEL
    
    else
       
       allocate(Q(mdl%li:mdl%ui,mdl%ny))

    end if

    ! prepare for scatter/gather

    call scatterv(mdl%nx,mp%np,counts,displs)

    ! scatter shear heat array to slaves

    do j = 1,mdl%ny

       call MPI_Scatterv(Q(:,j),counts,displs,MPI_REAL_PR, &
            Q(:,j),counts(mp%myid),MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)
 
    end do

    ! loop over the fault plane and solve diffusion equations (slaves)
    
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(i,j)

    !$OMP DO

    do j = 1,mdl%ny
       do i = mdl%li,mdl%ui
                    
          call thermpres_rate_point(mdl,flt%dT(:,i,j,stage),flt%dp(:,i,j,stage), &
               flt%T(:,i,j,stage),flt%p(:,i,j,stage),Q(i,j),tp)
          
       end do
    end do

    !$OMP END DO

    !$OMP END PARALLEL
    
    ! deallocate memory

    deallocate(Q)

  end subroutine thermpres_rate


  subroutine thermpres_rate_point(mdl,dT,dp,T,p,Q,tp)
    ! THERMPRES_RATE_POINT solves the 1D diffusion equations for thermal pressurization
    ! using explicit finite differences at one point on fault
    ! 
    ! Modified: 2 March 2007

    use model, only : model_type,w_of_z
    use constants, only : zero,half,one,two,twopi
    use asperity, only : asperity_type,ampl
    use io, only : error

    implicit none

    ! I/O Parameters:
    ! MDL = model variables
    ! DT = rate of change of temperature
    ! DP = rate of change of pore pressure
    ! T = temperature
    ! P = pore pressure
    ! Q = shear heating rate (slip velocity times shear stress)
    ! TP = thermal pressurization variables

    type(model_type),intent(in) :: mdl
    real(pr),dimension(:),intent(out) :: dT,dp
    real(pr),dimension(:),intent(in) :: T,p
    real(pr),intent(in) :: Q
    type(thermpres_type),intent(in) :: tp

    ! Internal Parameters:
    ! M = length of T,p
    ! K = index in z direction
    ! SN = effective normal stress, positive in compression
    ! RT = alpha_th/dz**2 = (alpha_th/dw**2)*dwdz**2
    ! RP = alpht_hy/dz**2 = (alpha_hy/dw**2)*dwdz**2
    ! ST = (alpha_th/2dw) * d2wdz2
    ! SP = (alpht_hy/2dw) * d2wdz2
    ! TP1 = T(i+1), used in finite difference equations
    ! TM1 = T(i-1), used in finite difference equations
    ! PP1 = p(i+1), used in finite difference equations
    ! PM1 = p(i-1), used in finite difference equations
    ! DTDZM = derivative of T with respect to z at left boundary
    ! DTDZP = derivative of T with respect to z at right boundary
    ! DPDZM = derivative of p with respect to z at left boundary
    ! DPDZP = derivative of p with respect to z at right boundary
    ! DTDWM = derivative of T with respect to w at left boundary
    ! DTDWP = derivative of T with respect to w at right boundary
    ! DPDWM = derivative of p with respect to w at left boundary
    ! DPDWP = derivative of p with respect to w at right boundary
    ! G = nondimensional shear strain rate
    ! A = normalization for strain rate distribution
    ! Z = distance normal to fault
    ! ASP = asperity, used to evaluate strain rate distribution
    ! C2 = coefficients of second-order approximation of Laplacian

    integer(pin) :: M,k
    real(pr) :: dTdzm,dTdzp,dpdzm,dpdzp,dTdwm,dTdwp,dpdwm,dpdwp,A
    real(pr),dimension(size(T)) :: rt,rp,st,sp,Tp1,Tm1,pp1,pm1
    real(pr),dimension(:),allocatable :: G
    type(asperity_type),pointer :: asp=>null()

    M = size(T)

    ! temperature gradient at fault (no heat flux by symmetry unless 
    ! shear heating enters as boundary condition--only for slip-on-plane model)

    if (tp%shape=='plane') then
       dTdzm = -half*Q/tp%K
    else
       dTdzm = zero
    end if

    ! temperature gradient at right boundary

    dTdzp = tp%dTdz

    ! no fluid flow across fault (by symmetry) or out of modeled region

    dpdzm = zero
    dpdzp = zero

    ! transform to w coordinate

    dTdwm = dTdzm/mdl%dwdz(1)
    dTdwp = dTdzp/mdl%dwdz(M)
    dpdwm = dpdzm/mdl%dwdz(1)
    dpdwp = dpdzp/mdl%dwdz(M)

    ! parameter combinations

    rt = tp%alphath/mdl%dw**2 * mdl%dwdz**2
    rp = tp%alphahy/mdl%dw**2 * mdl%dwdz**2
    st = tp%alphath/(two*mdl%dw)* mdl%d2wdz2
    sp = tp%alphahy/(two*mdl%dw)* mdl%d2wdz2

    ! shift T and p for use in difference formulas

    Tp1 = eoshift(T, 1,T(M-1)+two*mdl%dw*dTdwp)
    Tm1 = eoshift(T,-1,T(2  )-two*mdl%dw*dTdwm)

    pp1 = eoshift(p, 1,p(M-1)+two*mdl%dw*dpdwp)
    pm1 = eoshift(p,-1,p(2  )-two*mdl%dw*dpdwm)

    ! temperature rate

    dT = (rt+st)*Tp1 - two*rt*T + (rt-st)*Tm1

    ! add volumetric heat source if needed

    if (tp%shape/='plane') then

       allocate(G(M))
       allocate(asp)

       asp%type = tp%shape
       asp%inside = .true.
       asp%x0 = zero

       ! calculate nondimensionalized strain rate, G=W*(strain rate)/V,
       ! note that integral over z of G(z) must equal W, in order that
       ! integral over z of strain rate equals V

       ! calculate coefficient to properly normalize strain rate distribution

       select case(tp%shape)
       case default
          call error('Invalid shear zone shape','thermpres_rate')
       case('box')
          ! W is half-width in z-coordinate for consistency with 'box' in asperity routines
          A = half
          ! change the measure in z direction
          asp%length_x = w_of_z(mdl,tp%W)
          do k = 1,M
             G(k) = A*ampl(mdl%w(k),zero,mdl%dw,one,asp,dim=2)
          end do
       case('gaussian')
          ! W is standard deviation of gaussian (see asperity.f90 for exact formula)
          A = one/sqrt(twopi)
          asp%length_x = tp%W
          do k = 1,M
             G(k) = A*ampl(mdl%z(k),zero,mdl%dw,one,asp,dim=2) ! dw is not used in ampl
          end do
       end select

       dT = dT+tp%alphath/tp%K*Q/tp%W*G
       
       deallocate(G)
       deallocate(asp)
       asp => null()
       
    end if

    ! pressure rate

    dp = (rp+sp)*pp1 - two*rp*p + (rp-sp)*pm1 + tp%Lambda*dT

  end subroutine thermpres_rate_point


  function shear_heat(flt,i,j,stage) result(Q)
    ! SHEAR_HEAT calculates shear heating rate
    ! 
    ! Modified: 6 November 2006

    use fault, only : fault_type

    implicit none

    ! I/O Parameters:
    ! FLT = fault variables   
    ! I = index in x direction
    ! J = index in y direction
    ! STAGE = integration sub-step

    type(fault_type),intent(inout) :: flt
    integer(pin),intent(in) :: i,j,stage
    real(pr) :: Q

    ! Internal Parameters:
    ! S = amplitude of shear stress

    real(pr) :: s

    s = sqrt(flt%sx(i,j,stage)**2+flt%sy(i,j,stage)**2)

    Q = s*flt%V(i,j,stage)

  end function shear_heat


end module thermpres

