! MDSBI version 3.9.10 -- Multi-Dimensional Spectral Boundary Integral Code
module time_step

  ! TIME_STEP contains routines to advance problem through one time step
  ! 
  ! Modified: 10 February 2007

  use constants, only : pr,pin

  implicit none


contains


  subroutine step_explicit(pb,dt,s0,stress_transfer,accept_step,update_history,t1,t2)
    ! STEP_EXPLICIT advances problem through one time step by explicit integration
    ! 
    ! Modified: 9 February 2007

    use problem, only : problem_type
    use integration, only : integrate_rate_eqns,move_rates
    use rates, only : set_rates
    use io, only : error
    use mpi_routines, only : is_master

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DT = time step
    ! S0 = stage at beginning of time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values
    ! ACCEPT_STEP = accept integration step
    ! UPDATE_HISTORY = update history for convolution or overwrite
    ! T1 = time at beginning of elastodynamic step
    ! T2 = time at end of elastodynamic step

    type(problem_type),intent(inout) :: pb
    real(pr),intent(in) :: dt
    integer(pin),intent(in) :: s0
    character(*),intent(in) :: stress_transfer
    logical,intent(out) :: accept_step
    logical,intent(in),optional :: update_history
    real(pr),intent(in),optional :: t1,t2

    ! Internal Parameters:
    ! STEP = type of integration step

    character(64),parameter :: step = 'explicit'

    ! advance current time from t to t+dt
    
    pb%mdl%t = pb%mdl%t+dt
    
    ! predict rates at t+dt are those at t
    
    call move_rates(pb%fld%flt,pb%mdl,pb%mp,s0+1,s0)
    
    ! integrate rate equations

    call integrate_rate_eqns(pb%fld%flt,pb%mdl,pb%mp,dt,step,s0)
    
    ! set rates at t+dt

    select case(stress_transfer)

    case('convolution')

       if (is_master(pb%mp).and.(.not.(present(update_history)))) &
            call error('update_history must be present for convolution','step_explicit')

       call set_rates(pb,s0+1,stress_transfer,accept_step,update_history)

    case('interpolate')

       if (is_master(pb%mp).and.(.not.(present(t1).and.present(t2)))) &
            call error('t1 and t2 must be present for interpolation','step_explicit')

       call set_rates(pb,s0+1,stress_transfer,accept_step,t1=t1,t2=t2)

    end select

  end subroutine step_explicit


  subroutine step_trapezoid(pb,dt,s0,stress_transfer,accept_step,update_history,t1,t2)
    ! STEP_TRAPEZOID advances problem through one time step by trapezoidal rule integration
    ! 
    ! Modified: 9 February 2007

    use problem, only : problem_type
    use integration, only : integrate_rate_eqns
    use rates, only : set_rates
    use io, only : error
    use mpi_routines, only : is_master

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DT = time step
    ! S0 = stage at beginning of time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values
    ! ACCEPT_STEP = accept integration step
    ! UPDATE_HISTORY = update history for convolution or overwrite
    ! T1 = time at beginning of elastodynamic step
    ! T2 = time at end of elastodynamic step

    type(problem_type),intent(inout) :: pb
    real(pr),intent(in) :: dt
    integer(pin),intent(in) :: s0
    character(*),intent(in) :: stress_transfer
    logical,intent(out) :: accept_step
    logical,intent(in),optional :: update_history
    real(pr),intent(in),optional :: t1,t2

    ! Internal Parameters:
    ! STEP = type of integration step

    character(64),parameter :: step = 'trapezoid'

    ! integrate rate equations
    
    call integrate_rate_eqns(pb%fld%flt,pb%mdl,pb%mp,dt,step,s0)
    
    ! set rates at t+dt
    
    select case(stress_transfer)

    case('convolution')

       if (is_master(pb%mp).and.(.not.(present(update_history)))) &
            call error('update_history must be present for convolution','step_trapezoid')

       call set_rates(pb,s0+1,stress_transfer,accept_step,update_history=.false.)

    case('interpolate')

       if (is_master(pb%mp).and.(.not.(present(t1).and.present(t2)))) &
            call error('t1 and t2 must be present for interpolation','step_trapezoid')

       call set_rates(pb,s0+1,stress_transfer,accept_step,t1=t1,t2=t2)

    end select


  end subroutine step_trapezoid


  subroutine step_rk(pb,dt,s0,stress_transfer,accept_step,attempts,t0,t1,t2)
    ! STEP_RK advances problem through one time step using embedded Runge-Kutta
    ! time stepping
    ! 
    ! Modified: 14 February 2007
    
    use problem, only : problem_type
    use integration, only : integrate_rate_eqns
    use rates, only : set_rates
    use io, only : error
    use mpi_routines, only : is_master,master,MPI_REAL_PR
    use mpi

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! DT = time step
    ! S0 = stage at beginning of time step
    ! STRESS_TRANSFER = method used to calculate stress transfer
    !      convolution = perform convolution
    !      interpolation = interpolate between known values
    ! ACCEPT_STEP = accept integration step
    ! ATTEMPTS = number of attempts to take acceptable time step
    ! T0 = initial time
    ! T1 = time at beginning of elastodynamic time step
    ! T2 = time at end of elastodynamic time step

    type(problem_type),intent(inout) :: pb
    real(pr),intent(inout) :: dt
    integer(pin),intent(in) :: s0,attempts
    character(*),intent(in) :: stress_transfer
    logical,intent(out) :: accept_step
    real(pr),intent(in) :: t0
    real(pr),intent(in),optional :: t1,t2

    ! Internal Parameters:
    ! STEP = type of integration step
    ! M = Runge-Kutta stage
    ! STAGE = current integration stage (for storage)
    ! IERR = MPI error flag
    ! FAILED = flag indicating if error exceeds tolerance so time step failed

    character(64) :: step
    integer(pin) :: m,stage,ierr
    logical :: failed

    ! set number of stages

    ! ---stage 1---

    ! keep values from end of previous time step

    ! ---stages m=2,...,s---

    step = 'rk_stage'

    do m = 2,pb%mdl%rk%s

       ! set storage stage

       stage = s0-1+m

       ! advance current time
          
       pb%mdl%t = t0+pb%mdl%rk%c(m)*dt

       ! integrate rate equations
       
       call integrate_rate_eqns(pb%fld%flt,pb%mdl,pb%mp,dt,step,s0,m)
       
       ! set rates at new time

       select case(stress_transfer)

       case('convolution')

          call set_rates(pb,stage,stress_transfer,accept_step,update_history=.true.)

       case('interpolate')

          if (is_master(pb%mp).and.(.not.(present(t1).and.present(t2)))) &
               call error('t1 and t2 must be present for interpolation','step_rk')

          call set_rates(pb,stage,stress_transfer,accept_step,t1=t1,t2=t2)

       end select

       ! if unacceptable, reject this step and exit routine
       
       if (.not.accept_step) then

          call set_time_step(pb,controller='min',dt=dt)

          return

       end if

    end do

    ! ---error estimate---
    
    step = 'rk_error'

    call integrate_rate_eqns(pb%fld%flt,pb%mdl,pb%mp,dt,step,s0)

    ! reject step if error greater than tolerance

    if (is_master(pb%mp)) then

       if (pb%fld%flt%rerr(1)>pb%mdl%rtol) then
          failed = .true.
       else
          failed = .false.
       end if
          
    end if

    call MPI_Bcast(failed,1,MPI_LOGICAL,master,MPI_COMM_WORLD,ierr)

    if (failed) then

       ! decrease time step
          
       if (attempts>=10) then
             
          ! after 10 failed attempts, decrease by smallest permissible ratio
          ! (for a total reduction factor of step_min**10) and automatically accept step
             
          call set_time_step(pb,controller='min',dt=dt)

          if (is_master(pb%mp)) then
             print *, 'accepting unacceptable time step'
             print *, 'relative error = ',pb%fld%flt%rerr(1)
             print *, '     tolerance = ',pb%mdl%rtol
          end if

          accept_step = .true.

       else

          ! use error estimate to predict dt
             
          call set_time_step(pb,controller='I',dt=dt)

          accept_step = .false.

          return

       end if

    end if

    ! ---use higher-order formula for final values---

    step = 'rk_final'

    ! set storage stage

    stage = pb%mdl%ns

    ! advance current time
       
    pb%mdl%t = t0+dt

    call integrate_rate_eqns(pb%fld%flt,pb%mdl,pb%mp,dt,step,s0)

    ! set final rates

    select case(stress_transfer)
       
    case('convolution')
       
       call set_rates(pb,stage,stress_transfer,accept_step,update_history=.true.)

    case('interpolate')
       
       if (is_master(pb%mp).and.(.not.(present(t1).and.present(t2)))) &
            call error('t1 and t2 must be present for interpolation','step_rk')
       
       call set_rates(pb,stage,stress_transfer,accept_step,t1=t1,t2=t2)

    end select

    ! if unacceptable, reject this step and exit routine

    if (.not.accept_step) then
       
       call set_time_step(pb,controller='min',dt=dt)

       return

    end if
    
    ! step successful, predict next time step
       
    call set_time_step(pb,controller='PID',dt=dt)
    
  end subroutine step_rk


  subroutine set_time_step(pb,controller,dt)
    ! SET_TIME_STEP sets adaptive time step
    ! 
    ! Modified: 14 February 2007

    use constants, only : zero,one
    use problem, only : problem_type
    use io, only : error
    use mpi_routines, only : is_master,master,MPI_REAL_PR
    use mpi

    implicit none

    ! I/O Parameters:
    ! PB = problem variables
    ! CONTROLLER = step-size controller
    ! DT = time step

    type(problem_type),intent(in) :: pb
    character(*),intent(in) :: controller
    real(pr),intent(inout) :: dt

    ! Internal Parameters:
    ! CONTROL = actual step-size controller used
    ! ERR = error
    ! TOL = tolerance
    ! KAPPA = safety factor
    ! ALPHA = exponent in step-size formula
    ! BETA = exponent in step-size formula
    ! GAMMA = exponent in step-size formula
    ! STEP_RATIO = step-size ratio (new step-size/current step-size)
    ! STEP_RATIO_MIN = minimum step-size ratio
    ! STEP_RATIO_MAX = maximum step-size ratio
    ! ERROR_TYPE = which error (relative or absolute) to control
    ! IERR = MPI error flag

    character(64) :: control
    integer(pin) :: ierr
    real(pr) :: err(3),tol,step_ratio,alpha,beta,gamma
    real(pr),parameter :: step_ratio_min=0.1_pr, step_ratio_max=10._pr, kappa=0.9_pr
    character(64),parameter :: error_type = 'relative'

    ! index of error corresponds to following times:
    ! time:  t-2*dt t-dt t
    ! index:   3     2   1

    if (is_master(pb%mp)) then
    
       ! set error and tolerance
       
       select case(error_type)
       case('relative')
          err = pb%fld%flt%rerr
          tol = pb%mdl%rtol
       case('absolute')
          err = pb%fld%flt%aerr
          tol = pb%mdl%atol
       end select
              
       ! set default controller to input value

       control = controller

       ! switch controller for special cases that might cause problems

       if (controller/='min') then
          if (err(1)==zero) control = 'max'
          if (err(2)==zero.and.(control=='PID'.or.control=='PI')) control = 'I'
          if (err(3)==zero.and.(control=='PID')) control = 'PI'
       end if

       ! set time step with controller
       
       select case(control)
          
       case default
          
          call error('invalid step-size controller','set_time_step')
          
       case ('min') ! minimum step-size ratio
          
          step_ratio = step_ratio_min
          
       case ('max') ! maximum step-size ratio
          
          step_ratio = step_ratio_max

       case('I') ! integral feedback (I) controller
          
          alpha = one/(one+pb%mdl%rk%p)
          step_ratio = kappa*(tol/err(1))**alpha

       case('PI') ! proportional+integral feedback (PI) controller
          
          alpha = 0.7_pr/pb%mdl%rk%p
          beta = 0.4_pr/pb%mdl%rk%p
          step_ratio = kappa*(tol/err(1))**alpha*(err(2)/tol)**beta

       case('PID') ! proportional+integral+derivative feedback (PID) controller
          
          alpha = 0.49_pr/pb%mdl%rk%p
          beta = 0.34_pr/pb%mdl%rk%p
          gamma = 0.1_pr/pb%mdl%rk%p
          step_ratio = kappa*(tol/err(1))**alpha*(err(2)/tol)**beta*(tol/err(3))**gamma
          
       end select
       
       ! bound step-size ratio (prevent overly small or large changes)
       
       dt = dt*min(step_ratio_max,max(step_ratio_min,step_ratio))
       
       ! bound step size (if lower limit is specified)
       
       if (pb%mdl%dtmin/=zero) dt = max(dt,pb%mdl%dtmin)
       
    end if

    call MPI_Bcast(dt,1,MPI_REAL_PR,master,MPI_COMM_WORLD,ierr)

  end subroutine set_time_step


end module time_step
