% This file contains examples of using matlab commands to
% analyze data.  It will show you how to exploit the output
% opportunities from MDSBI.

% Set "datadir" to the directory where your data is stored.  If
% you don't do this, it will use the default ../data (which is most
% likely what you will want).

datadir = '~/esag/mdsbi/';

% Set "endian" to the endian of the system on which you ran the
% code, which might be different from the one on which you are
% running matlab to analyze the data.  If you used the same machine
% and don't know the endian, set this to 'n' (for 'native').

endian = 'l';

% Now we will analyze a few specific problems.

% Problem: slipweak_identical_300m

% Store the name of the problem in "name".

name = 'slipweak_identical_300m';


% Call initialization function "init.m".  This only needs to be run once per
% problem, unless you have rerun the program with different parameters or if
% you are examining the data while the program is running, and further time steps
% have occurred.  This loads information about the run in the structure A.

A = init(name,datadir,endian);

% Now we will load the data using the function "loadmesh.m".  It loads a particular
% field output into a data structure.  To see which field outputs are
% available for loading, type "A":

A

% This displays all of matlab variables.  You see several
% structures containing fields that were output (Vxh, Uxf, Vxf, etc.).
% We will step through these.

% First we will look at the slip velocity at the hypocenter (a
% single point) as a function of time, which is named "Vxh".
% Let's see what is stored in this structure:

A.Vxh

% We see a number of components.  There's "name", which is the same
% as the name of the structure, but stored as a string.  "field" is
% the field that was output.  There are the field dimensions, "nx",
% "ny", "nz", "nt". There are the minimum and maximum values of the
% position coordinates ("xmin", "xmax", etc.).  The program is
% already set up to use the proper coordinates, so you never have to
% figure out how to translate the index to a coordinate value.  If
% any of the dimensions has only one point, then the field may have
% been interpolated (say, to look at seismograms at a particular
% point that is not exactly located on a grid point).  The data can
% also be output every few points, using the "stride_x", etc. variables.

% Now load the data:

Vxh = loadmesh(A,'Vxh');

% The name on the left is the new structure (if you want to rename
% it, you can).  Let's examine its contents:

Vxh

% There are a few additions to what we saw previously: "x", "y",
% "z", "t", and "Vx".  These new variables are the position
% and time coordinates, and an array containing the field values.

% Plot slip velocity in the x direction versus time:

plot(Vxh.t,Vxh.Vx,'-o')
xlabel('Time (s)')
ylabel('Slip velocity (m/s)')

% This gives the slip rate at the hypocenter.  The field here has
% zero spatial dimensions, but the code gives us the option to have
% 1D, 2D, or 3D fields as well.

% Now look at Vxf, which is the slip velocity on the fault.  Load it:

Vxf = loadmesh(A,'Vxf');

% This field has two spatial dimensions.  You can see this by
% examining its contents:

Vxf

% Now let's loop over each time step, and plot snapshots of slip
% velocity on the fault.  The way matlab's pcolor command works is
% to use the coordinate (x,y) as the left and bottom corner of 
% the colored rectangle.  To properly center the rectangle, offset it by
% half of a grid spacing.

for it=1:Vxf.nt
  pcolor(Vxf.x-Vxf.dx/2,Vxf.y-Vxf.dy/2,squeeze(Vxf.Vx(it,:,:))')
  shading flat
  caxis([0 10])
  colorbar
  axis image
  xlabel('x, Distance along strike (km)')
  ylabel('y, Distance down dip (km)')
  title('Slip velocity on fault (m/s)')
  pause(0.1)
end

% This shows the rupture process.  You can see the influence of the
% high stress asperity on the right, and the barrier of high
% static friction on the left.

% You can do something similar to look at the shear stress
% component parallel to (most of the) slip on the fault.

sxf = loadmesh(A,'sxf');

for it=1:sxf.nt
  pcolor(sxf.x-sxf.dx/2,sxf.y-sxf.dy/2,squeeze(sxf.sx(it,:,:))')
  shading flat
  caxis([62 106])
  colorbar
  axis image
  xlabel('x, Distance along strike (km)')
  ylabel('y, Distance down dip (km)')
  title('Horizontal shear stress on fault (MPa)')
  pause(0.1)
end

% And likewise with the transverse stress component on the fault.

syf = loadmesh(A,'syf');

for it=1:syf.nt
  pcolor(syf.x-syf.dx/2,syf.y-syf.dy/2,squeeze(syf.sy(it,:,:))')
  shading flat
  caxis([-6 6])
  colorbar
  axis image
  xlabel('x, Distance along strike (km)')
  ylabel('y, Distance down dip (km)')
  title('Vertical shear stress on fault (MPa)')
  pause(0.1)
end


% In addition to the "mesh" output option, you can also output the
% time at which the rupture front passes (defined in the input file
% as the time when slip velocity exceeds 1e-3 m/s.  This is stored in
% the structure "frt".

% Take a look at its contents:

A.frt

% Next, load in the timing data (stored as a text file):

F = loadfront(A);

% Again, look at its contents:

F

% You see the addition of three arrays: "x" and "y" contain the
% spatial coordinates and "t" contains the time at which each
% point in the array satisfied the rupture front condition.  Now
% make a contour plot of the rupture front data:

t_contour = [0.5:0.5:2.5];
contour(F.x,F.y,F.t,t_contour,'k')
axis image
xlabel('Distance along strike (km)')
ylabel('Depth (km)')
title('Rupture front contours (every 0.5 s)')


% -----------------------------------------------------------------------
% Bimaterial problem - compare with identical materials version
%
% This problem is similar to the one above, except that the materials
% on either side of the fault are different.  Since it is so similar, I'll
% just list a few commands without explaining them in detail.

name = 'slipweak_bimaterial_100m';

B = init(name,datadir,endian);

% In this case, we can output data on either side of the fault.  "p" stands
% for the top (or "plus") side and "m" stands for the bottom (or "minus")
% side.

% First, look at what happens at the hypocenter.

vxmh = loadmesh(B,'vxmh');
vxph = loadmesh(B,'vxph');
Vxh = loadmesh(B,'Vxh');

plot(vxph.t,vxph.vxp,'-o',vxmh.t,vxmh.vxm,'-o')
xlabel('Time (s)')
ylabel('Particle velocity (m/s)')
legend('upper side','lower side')

plot(Vxh.t,Vxh.Vx,'-o')
xlabel('Time (s)')
ylabel('Slip velocity (m/s)')

% Here's the slip velocity history.

Vxf = loadmesh(B,'Vxf');

for it=1:Vxf.nt
  pcolor(Vxf.x-Vxf.dx/2,Vxf.y-Vxf.dy/2,squeeze(Vxf.Vx(it,:,:))')
  shading flat
  caxis([0 10])
  colorbar
  axis image
  xlabel('x, Distance along strike (km)')
  ylabel('y, Distance down dip (km)')
  title('Slip velocity on fault (m/s)')
  pause(0.1)
end

% There are other fields to explore as well.  Take a look in the structure "B"
% to find out which ones are availible.  If you set the material parameters on
% either side to identical values (rcs=1, rrho=1) and turn off the poroelastic effect,
% you can re-run this problem and confirm that this method gives almost the same
% results as the identical materials version (although it's slower and more memory intensive!).


% -----------------------------------------------------------------------
% Point force test problem - compare with analytical solution
%
% This is a good problem to see the effects of varying the resolution,
% integration options, kernel interpolation method, etc.  In the
% plots, the magnitude of the force is S.

name = 'pointforce_identical';
C = init(name,datadir,endian);

% Load data
Ux = loadmesh(C,'Ux');
Uy = loadmesh(C,'Uy');

% observation point (try different locations - points further from
% the center show better separation of the wavefronts)

ix = 10; iy = 10; % numbers in terms of grid points
x0 = Ux.x(ix); y0 = Ux.y(iy);

% r0 is the distance from the force to the observation point
r0 = sqrt(x0^2+y0^2); 

% i is for identical materials

Uxi = squeeze(Ux.Ux(:,ix,iy));
Uyi = squeeze(Uy.Uy(:,ix,iy));

% calculate the analytical solution

Uxa = zeros(1,Ux.nt);
Uya = zeros(1,Uy.nt);

for it=1:Ux.nt
  Uxa(it) = pointforce_displacement(Ux.t(it),x0,y0,C.mdl.cp,'x');
end

for it=1:Uy.nt
  Uya(it) = pointforce_displacement(Uy.t(it),x0,y0,C.mdl.cp,'y');
end

% plot both components of slip as a function of time, compare with
% analytical solution in non-dimensionalized coordinates

% slip in direction of force (note factor of 2 to get slip from displacement)
plot(Ux.t*C.mdl.cs./r0,2*pi*r0*Uxa,'bo',Ux.t*C.mdl.cs./r0,C.mdl.mu*pi*r0*Uxi,'r.')
set(gca,'fontsize',16)
xlabel('Non-dimensional time (c_st/r)')
ylabel('Non-dimensional slip (\mu\pir\Deltau_x/S)')

% slip perpendicular to force (note factor of 2 to get slip from displacement)
plot(Uy.t*C.mdl.cs./r0,2*pi*r0*Uya,'bo',Uy.t*C.mdl.cs./r0,C.mdl.mu*pi*r0*Uyi,'r.')
set(gca,'fontsize',16)
xlabel('Non-dimensional time (c_st/r)')
ylabel('Non-dimensional slip (\mu\pir\Deltau_y/S)')

% As you can see, the solution needs refinement.  Try running this problem again
% at a higher resolution.

% -----------------------------------------------------------------------
% Thermal pressurization and flash heating example
%
% This problem demonstrates rupture on a fault with flash heating
% (in a rate-and-state framework) and thermal pressurization.

name='thermpres';
E = init(name,datadir,endian);

% plot slip velocity on fault plane as a function of time

Vxf = loadmesh(E,'Vxf');

pcolor(Vxf.x*1e3,Vxf.t*1e6,Vxf.Vx)
shading flat
colorbar
set(gca,'fontsize',16)
xlabel('Distance along strike (m)','fontsize',16)
ylabel('Time (\mus)','fontsize',16)
title('Slip velocity (m/s)')

% plot effective stress on fault plane as a function of time

szf = loadmesh(E,'szf');

pcolor(szf.x*1e3,szf.t*1e6,szf.sz)
shading flat
colorbar
set(gca,'fontsize',16)
xlabel('Distance along strike (m)','fontsize',16)
ylabel('Time (\mus)','fontsize',16)
title('Effective normal stress (MPa)')

% plot pore pressure on fault plane as a function of time

pf = loadmesh(E,'pf');

pcolor(pf.x*1e3,pf.t*1e6,pf.p0)
shading flat
colorbar
set(gca,'fontsize',16)
xlabel('Distance along strike (m)','fontsize',16)
ylabel('Time (\mus)','fontsize',16)
title('Pore pressure (MPa)')

% plot temperature on fault plane as a function of time

Tf = loadmesh(E,'Tf');

pcolor(Tf.x*1e3,Tf.t*1e6,Tf.T0)
shading flat
colorbar
set(gca,'fontsize',16)
xlabel('Distance along strike (m)','fontsize',16)
ylabel('Time (\mus)','fontsize',16)
title('Pore pressure (MPa)')

% -----------------------------------------------------------------------
% Heterogeneous fields example
%
% This problem is similar to slipweak_identical_300m.  The
% additional feature is that the initial stress
% and the static friction fields are given a complicated
% perturbation.  The perturbations to the fields are determined in
% two user input files, one in ASCII format and the other in binary
% format.  This problem consequently demonstrates how to use these 
% input files to set the initial conditions.  The input files were generated
% in the matlab script input_heterogeneous.m in the problems directory.

name='heterogeneous';
D = init(name,datadir,endian);

% Plot slip velocity at hypocenter versus time:

Vxh = loadmesh(D,'Vxh');

plot(Vxh.t,Vxh.Vx,'-o')
xlabel('Time (s)')
ylabel('Slip velocity (m/s)')

% plot slip velocity on fault plane as a function of time

Vxf = loadmesh(D,'Vxf');

for it=1:Vxf.nt
  pcolor(Vxf.x,Vxf.y,squeeze(Vxf.Vx(it,:,:))')
  shading flat
  colorbar
  axis image
  set(gca,'fontsize',16)
  xlabel('Distance along strike (km)','fontsize',16)
  ylabel('Depth (km)','fontsize',16)
  title(['Slip velocity (m/s) at t=',num2str(Vxf.t(it),4),'s'])
  pause(0.1)
end

% rupture front contours

F = loadfront(D);

t_contour = [0.5:0.5:2.5];
contour(F.x,F.y,F.t,t_contour,'k')
axis image
xlabel('Distance along strike (km)')
ylabel('Depth (km)')
title('Rupture front contours (every 0.5 s)')

